package be.iminds.ilabt.jfed.ssh_terminal_tool.putty;

import be.iminds.ilabt.jfed.lowlevel.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnection;
import be.iminds.ilabt.jfed.lowlevel.ssh_key_info.SshKeyInfo;
import be.iminds.ilabt.jfed.preferences.GuiPreferenceKey;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.util.IOUtils;
import be.iminds.ilabt.jfed.util.JFedUtils;
import be.iminds.ilabt.jfed.util.KeyUtil;
import be.iminds.ilabt.jfed.util.tmp_file_helpers.TmpFile;
import org.apache.commons.codec.binary.Base64;
import org.slf4j.LoggerFactory;

import javax.inject.Inject;
import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.URL;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.interfaces.RSAPrivateCrtKey;
import java.security.interfaces.RSAPublicKey;
import java.util.*;
import java.util.regex.Pattern;

/**
 * PuttySession helper to store a putty session in registry
 * <p/>
 * Note:  in r888 and before, this used awt.Robot to call puttgen.exe and convert the key.
 * since then, PuTTYPrivateKeyFile has been written to natively read/write PuTTY files
 */
public class PuttyHelper {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(PuttyHelper.class);
    private final static Pattern ipv6pat = Pattern.compile("([0-9a-f]{1,4}:){7}([0-9a-f]){1,4}");

    private final JFedPreferences jFedPreferences;
    private final PageantHelper pageantHelper;

    @Inject
    public PuttyHelper(JFedPreferences jFedPreferences, PageantHelper pageantHelper) {
        this.jFedPreferences = jFedPreferences;
        this.pageantHelper = pageantHelper;
    }

    /*
         * makePpkIfNeeded and getUserEncryptedPPK are used to create PPK files with the same private key and password as in the PEM private key + certificate file.
         * */
    private static Map<String, String> userUrnToEncryptedPPK = new HashMap<>();

    private static File tmpFile(String baseName, String ext) {
        File tmpDir = new File(JFedUtils.getUserDataDirectoryFile(), "tmp");
        if (!tmpDir.exists()) tmpDir.mkdir();
        long nr = new Date().getTime();
        return new File(tmpDir, baseName + nr + "." + ext);
    }

    /**
     * make sure the file has windows line termination! \r\n
     */
    public static void runReg(File regFile) throws IOException {
        assert regFile.exists();

        LOG.debug("Starting reg for " + regFile.getPath());
        ArrayList<String> command = new ArrayList<>();
        command.add("reg");
        command.add("IMPORT");
        command.add("\"" + regFile.getPath() + "\"");
        Process p = Runtime.getRuntime().exec(command.toArray(new String[command.size()]));

        BufferedReader input = new BufferedReader(new InputStreamReader(p.getInputStream()));
        String line = null;
        while ((line = input.readLine()) != null) {
            LOG.debug("reg output: " + line);
        }
        input.close();

        try {
            p.waitFor();
        } catch (InterruptedException e) {
            //ignore
        }
        LOG.debug("reg stopped");
    }

    /**
     * make sure the file has windows line termination! \r\n
     */
    public static List<String> getAllPuttySessionNames() throws IOException {
        List<String> res = new ArrayList<>();

        LOG.debug("Starting reg query");
        ArrayList<String> command = new ArrayList<>();
        command.add("reg");
        command.add("query");
        command.add("\"hkcu\\Software\\SimonTatham\\PuTTY\\Sessions\"");
        Process p = Runtime.getRuntime().exec(command.toArray(new String[command.size()]));

        BufferedReader input = new BufferedReader(new InputStreamReader(p.getInputStream()));
        String line = null;
        while ((line = input.readLine()) != null) {
            //TODO parse and add to res
            res.add(line);
            LOG.debug("reg output: " + line);
        }
        input.close();

        try {
            p.waitFor();
        } catch (InterruptedException e) {
            //ignore
        }
        LOG.debug("reg stopped");

        return res;
    }

    /**
     * make sure the file has windows line termination! \r\n
     */
    public static void deletePuttySession(String sessionName) throws IOException {
        LOG.debug("Starting reg delete");
        ArrayList<String> command = new ArrayList<>();
        command.add("reg");
        command.add("delete");
        command.add("\"hkcu\\Software\\SimonTatham\\PuTTY\\Sessions\\" + sessionName + "\"");
        command.add("/f");
        Process p = Runtime.getRuntime().exec(command.toArray(new String[command.size()]));

        BufferedReader input = new BufferedReader(new InputStreamReader(p.getInputStream()));
        String line = null;
        while ((line = input.readLine()) != null) {
            LOG.debug("reg output: " + line);
        }
        input.close();

        try {
            p.waitFor();
        } catch (InterruptedException e) {
            //ignore
        }
        LOG.debug("reg stopped");
    }

    /* inspired by http://svn.tartarus.org/sgt/putty/contrib/kh2reg.py?view=markup */
    public static String convertSshHostKeyToPuttyFormat(String openSshKnownHost) {
        StringBuilder res = new StringBuilder();

        String[] parts = openSshKnownHost.split(" ");
        if (parts.length < 3)
            throw new IllegalArgumentException("The provided known host line is not in the correct format: " + openSshKnownHost);
        String hostname = parts[0];
        String sshkeytype = parts[1];
        String encodedKey = parts[2];

        String puttySshKeyType;
        if (sshkeytype.equals("ssh-rsa"))
            puttySshKeyType = "rsa2";
        else {
            if (sshkeytype.equals("dss-rsa"))
                puttySshKeyType = "dss";
            else
                throw new IllegalArgumentException("The provided known host line has an unknown key type: " + sshkeytype);
        }

        byte[] all_restored = Base64.decodeBase64(encodedKey);
        ByteBuffer bb = ByteBuffer.wrap(all_restored).order(ByteOrder.BIG_ENDIAN);
        List<String> fields = new ArrayList<>();
        while (bb.remaining() > 0) {
            int len = bb.getInt();

            if (len > bb.remaining())
                throw new IllegalArgumentException("The provided known host line is not in the correct format. field " + fields.size() + " has len=" + len + " but only " + bb.remaining() + " bytes remain. key=" + encodedKey);

            StringBuilder field = new StringBuilder();
            for (int i = 0; i < len; i++) {
                byte b = bb.get();
                String hex = String.format("%02x", b);

                //skip leading zeroes
                if (field.length() == 0) {
                    if (b == 0) {
                        //do not append
                    } else {
                        if (hex.startsWith("0")) {
                            //remove leading zero
                            hex = hex.substring(1);
                            field.append(hex);
                        } else
                            field.append(hex);
                    }
                } else
                    field.append(hex);


            }
            fields.add(field.toString());
        }

        if (fields.size() < 2)
            throw new IllegalArgumentException("The provided known host line is not in the correct format. field count: " + fields.size());

        for (int i = 1; i < fields.size(); i++) {
            if (i > 1)
                res.append(",");
            res.append("0x" + fields.get(i));
        }

        return res.toString();
    }

    public static String convertSshHostKeyToRegistryEditLine(String openSshKnownHost, int port) {
        String key = convertSshHostKeyToPuttyFormat(openSshKnownHost);

        String[] parts = openSshKnownHost.split(" ");
        if (parts.length < 3)
            throw new IllegalArgumentException("The provided known host line is not in the correct format: " + openSshKnownHost);
        String hostname = parts[0];
        String sshkeytype = parts[1];

        String puttySshKeyType;
        if (sshkeytype.equals("ssh-rsa"))
            puttySshKeyType = "rsa2";
        else {
            if (sshkeytype.equals("dss-rsa"))
                puttySshKeyType = "dss";
            else
                throw new IllegalArgumentException("The provided known host line has an unknown key type: " + sshkeytype);
        }

        return "\"" + puttySshKeyType + "@" + port + ":" + hostname + "\"=\"" + key + "\"";
    }

    public static void storeOpenSshHostKeyToPuttyRegistry(String openSshKnownHost, int port) throws IOException {
        String regEditLine = convertSshHostKeyToRegistryEditLine(openSshKnownHost, port);
        String regEditContent = "Windows Registry Editor Version 5.00\n" +
                "\n" +
                "[HKEY_CURRENT_USER\\Software\\SimonTatham\\PuTTY\\SshHostKeys]\n" +
                regEditLine + "\n";

        File regFile = tmpFile("hostKey", "reg");
        IOUtils.stringToFile(regFile, regEditContent.replace("\n", "\r\n"));

        runReg(regFile);
    }

    public static File convertPrivateKeyToPuttyFormat(PrivateKey pk, File targetPpkFile, String passphrase) {
        RSAPrivateCrtKey privKey = (RSAPrivateCrtKey) pk;
        RSAPublicKey publicKey = KeyUtil.rsaPrivateCrtKeyToPublicKey(privKey);
        PuTTYPrivateKeyFile.createFile("jFed-generated-ppk", publicKey, privKey, targetPpkFile, passphrase);
        return targetPpkFile;
    }

    /**
     * This assumes:
     * - the private keys are stored in files in the puttygen ppk format.
     * - the private key for the proxy is unlocked
     * - the proxy host key has been registered to putty
     *
     * @param username is not needed for putty session, it may be null
     */
    public static void createPuttySession(String sessionName,
                                          String hostname, int port, File puttyPrivateKeyFile, String username,
                                          String proxyHostname, int proxyPort, File proxyPuttyPrivateKeyFile, String proxyUsername,
                                          boolean enableX11Forward) throws IOException {
        StringBuilder regContent = new StringBuilder();
        regContent.append("Windows Registry Editor Version 5.00\n" +
                "\n" +
                "[HKEY_CURRENT_USER\\Software\\SimonTatham\\PuTTY\\Sessions]\n" +
                "\n" +
                "[HKEY_CURRENT_USER\\Software\\SimonTatham\\PuTTY\\Sessions\\" + sessionName + "]\n");

        URL sessionTemplateUrl = PuttyHelper.class.getResource("PuttySessionRegistryTemplate");
        String sessionTemplate = IOUtils.streamToString(sessionTemplateUrl.openStream(), "UTF-8");
        regContent.append(sessionTemplate);
        if (!sessionTemplate.endsWith("\n"))
            regContent.append("\n");

        regContent.append("\"X11Forward\"=dword:0000000"+(enableX11Forward?"1":"0") + "\n");

        if (hostname != null)
            regContent.append("\"HostName\"=\"" + hostname + "\"" + "\n");
        else
            regContent.append("\"HostName\"=\"\"" + "\n");
        if (port >= 0)
            regContent.append("\"PortNumber\"=dword:" + Integer.toHexString(port) + "\n");
        if (puttyPrivateKeyFile != null)
            regContent.append("\"PublicKeyFile\"=\"" + puttyPrivateKeyFile.getPath().replace("\\", "\\\\") + "\"" + "\n");
        else
            regContent.append("\"PublicKeyFile\"=\"\"" + "\n");

        if (proxyHostname != null) {
            regContent.append("\"ProxyMethod\"=dword:00000005" + "\n");
            regContent.append("\"ProxyHost\"=\"" + proxyHostname + "\"" + "\n");
            regContent.append("\"ProxyPort\"=dword:" + Integer.toHexString(proxyPort) + "\n");
//            regContent.append("\"ProxyTelnetCommand\"=\"plink -i \\\""+proxyPuttyPrivateKeyFile.getPath().replace("\\", "\\\\")+"\\\" -l "+proxyUsername+" %proxyhost -nc %host:%port\""+"\n");
            if (proxyPuttyPrivateKeyFile != null)
                regContent.append("\"ProxyTelnetCommand\"=\"plink -i \\\"" + proxyPuttyPrivateKeyFile.getPath().replace("\\", "\\\\\\\\") + "\\\" " +
                        "-l " + proxyUsername + " -P %proxyport %proxyhost -nc %host:%port\"" + "\n");
            else
                regContent.append("\"ProxyTelnetCommand\"=\"plink -l " + proxyUsername + " -P %proxyport %proxyhost -nc %host:%port\"" + "\n");
        } else {
            regContent.append("\"ProxyMethod\"=dword:00000000" + "\n");
            regContent.append("\"ProxyHost\"=\"\"" + "\n");
            regContent.append("\"ProxyPort\"=dword:00000000" + "\n");
            regContent.append("\"ProxyTelnetCommand\"=\"\"" + "\n");
        }

        if (username != null)
            regContent.append("\"UserName\"=\"" + username + "\"" + "\n");
        else
            regContent.append("\"UserName\"=\"\"" + "\n");

        //Allow agent forwarding (cfr. FEDIBBTDEV-280)
        regContent.append("\"AgentFwd\"=dword:00000001" + "\n");

        File regFile = tmpFile("session", "reg");
        IOUtils.stringToFile(regFile, regContent.toString().replace("\n", "\r\n"));

        runReg(regFile);

        //no reason to keep registry file
        regFile.delete();
    }

    public void runPutty(final PuttyRunParameters parameters) {
        if (parameters.hasProxy())
            runPuttyWithPartialSession(parameters);
        else {
//            if (parameters.isEnableX11Forward())
//                runPuttyWithPartialSession(parameters);
//            else
            runPuttyWithoutSession(parameters);
        }
    }

    /**
     * supports both with and without proxy, new session, which includes destination
     * <p/>
     * Specifies target details in session.
     * Specifies proxy details in session.
     */
    public void runPuttyWithFullSession(final PuttyRunParameters parameters) {
        LOG.debug("Creating PuTTY session and starting PuTTY with it.");
        final boolean usePageant = jFedPreferences.getBoolean(GuiPreferenceKey.PREF_PUTTY_PAGEANT, true);

        List<SshKeyInfo> pageantKeys = new ArrayList<>();
        final TmpFile puttyPrivateKeyFile;
        if (usePageant) {
            puttyPrivateKeyFile = null;
            //PageantHelper.registerKey(parameters.getSshKeyInfo());
            pageantKeys.add(parameters.getSshKeyInfo());
        } else
            puttyPrivateKeyFile = parameters.getSshKeyInfo().getPuttyKeyFile();

        if (puttyPrivateKeyFile != null)
            puttyPrivateKeyFile.store();

        TmpFile proxyPuttyPrivateKeyFile = null;
        try {
            if (parameters.hasProxy() &&
                    (puttyPrivateKeyFile == null || !parameters.samePrivateKeyForProxy() || PuTTYPrivateKeyFile.read(puttyPrivateKeyFile.getFile()).isEncrypted())) {
                if (usePageant) {
                    //PageantHelper.registerKey(parameters.getProxySshKeyInfo());
                    pageantKeys.add(parameters.getProxySshKeyInfo());

                    //                    //use locked key file now
                    //                    proxyPuttyPrivateKeyFileHelper= parameters.getProxySshKeyInfo().getPuttyKeyFile();

                    proxyPuttyPrivateKeyFile = null; //use no key file
                } else {
                    proxyPuttyPrivateKeyFile = parameters.getProxySshKeyInfo().getUnlockedPuttyKeyFile();
                }
            } else
                proxyPuttyPrivateKeyFile = null;
            final TmpFile finalProxyPuttyPrivateKeyFile = proxyPuttyPrivateKeyFile;

            if (finalProxyPuttyPrivateKeyFile != null)
                finalProxyPuttyPrivateKeyFile.store();

            //register all keys at once
            if (!pageantKeys.isEmpty())
                pageantHelper.registerKeys(pageantKeys);

            //store proxyOpenSshKey
            if (parameters.hasProxy() && parameters.getProxyOpenSshKeyHostKey() != null)
                storeOpenSshHostKeyToPuttyRegistry(parameters.getProxyOpenSshKeyHostKey(), parameters.getProxyPort());

            //create session
            final String sessionName = "jFed" + (new Date().getTime());
            createPuttySession(sessionName, parameters.getHostname(), parameters.getPort(), puttyPrivateKeyFile.getFile(), parameters.getUsername(),
                    parameters.getProxyHostname(), parameters.getProxyPort(),
                    finalProxyPuttyPrivateKeyFile == null ? (puttyPrivateKeyFile == null ? null : puttyPrivateKeyFile.getFile()) : finalProxyPuttyPrivateKeyFile.getFile(),
                    parameters.getProxyUsername(),
                    parameters.isEnableX11Forward());

            //run session
            Runnable processOutputMonitor = () -> {
                Process p = null;
                try {
                    ArrayList<String> command = new ArrayList<>();
                    command.add(jFedPreferences.getFile(GuiPreferenceKey.PREF_PUTTY_DIRECTORY).getPath() + File.separator + "putty.exe");

                    command.add("-load");
                    command.add(sessionName);

                    LOG.debug("Starting PuTTY and waiting for it to stop");
                    LOG.info("PuTTY command: " + command);
                    p = Runtime.getRuntime().exec(command.toArray(new String[command.size()]));
                    BufferedReader input = new BufferedReader(new InputStreamReader(p.getInputStream()));
                    String line = null;
                    while ((line = input.readLine()) != null) {
                        LOG.debug("putty.exe output: " + line);
                    }
                    input.close();
                    try {
                        p.waitFor();
                    } catch (InterruptedException e) {
                        //ignore
                    }
                    LOG.debug("PuTTY stopped");
                } catch (IOException e) {
                    LOG.error("IOException while trying to launch external terminal: " + e.getMessage(), e);
                } finally {
                    LOG.debug("removing unencrypted putty key files");
                    if (puttyPrivateKeyFile != null) puttyPrivateKeyFile.delete();
                    if (finalProxyPuttyPrivateKeyFile != null) finalProxyPuttyPrivateKeyFile.delete();
                }
            };
            Thread t = new Thread(processOutputMonitor);
            t.setName("PuTTY-Thread");
            t.setDaemon(true); //=> this application may exit before putty stops
            t.start();
            LOG.debug("Started PuTTY thread.");
        } catch (IOException e) {
            //remove private key files
            if (puttyPrivateKeyFile != null) puttyPrivateKeyFile.delete();
            if (proxyPuttyPrivateKeyFile != null) proxyPuttyPrivateKeyFile.delete();
        }
    }

    /**
     * supports both with and without proxy, reuses same session for same proxy.
     * <p/>
     * Specifies target details on command line.
     * Specifies proxy details in session.
     */
    public void runPuttyWithPartialSession(final PuttyRunParameters parameters) {
        LOG.debug("Creating PuTTY session and starting PuTTY with it.");

        List<SshKeyInfo> pageantKeys = new ArrayList<>();

        final boolean usePageant = jFedPreferences.getBoolean(GuiPreferenceKey.PREF_PUTTY_PAGEANT, true);
        final TmpFile puttyPrivateKeyFile;
        if (usePageant) {
            puttyPrivateKeyFile = null;
            //PageantHelper.registerKey(parameters.getSshKeyInfo());
            pageantKeys.add(parameters.getSshKeyInfo());
        } else
            puttyPrivateKeyFile = parameters.getSshKeyInfo().getPuttyKeyFile();

        if (puttyPrivateKeyFile != null)
            puttyPrivateKeyFile.store();

        TmpFile proxyPuttyPrivateKeyFile = null;
        try {
            if (parameters.hasProxy() &&
                    (puttyPrivateKeyFile == null || !parameters.samePrivateKeyForProxy() || PuTTYPrivateKeyFile.read(puttyPrivateKeyFile.getFile()).isEncrypted())) {
                if (usePageant) {
                    //PageantHelper.registerKey(parameters.getProxySshKeyInfo());
                    pageantKeys.add(parameters.getProxySshKeyInfo());

                    //                    //use locked key file now
                    //                    proxyPuttyPrivateKeyFileHelper= parameters.getProxySshKeyInfo().getPuttyKeyFile();

                    proxyPuttyPrivateKeyFile = null; //use no key file
                } else {
                    proxyPuttyPrivateKeyFile = parameters.getProxySshKeyInfo().getUnlockedPuttyKeyFile();
                }
            } else
                proxyPuttyPrivateKeyFile = null;
            final TmpFile finalProxyPuttyPrivateKeyFile = proxyPuttyPrivateKeyFile;

            if (finalProxyPuttyPrivateKeyFile != null)
                finalProxyPuttyPrivateKeyFile.store();

            //register all keys at once
            if (!pageantKeys.isEmpty())
                pageantHelper.registerKeys(pageantKeys);

            //store proxyOpenSshKey
            if (parameters.hasProxy() && parameters.getProxyOpenSshKeyHostKey() != null)
                storeOpenSshHostKeyToPuttyRegistry(parameters.getProxyOpenSshKeyHostKey(), parameters.getProxyPort());

            //create session, without target details
            String sessionNameHelper;
            if (parameters.hasProxy())
                sessionNameHelper = "jFed-proxy-" + parameters.getProxyHostname().replace(".", "_");
            else
                sessionNameHelper = "jFed-no-proxy"; //won't be usefull to use a session in this case
            final String sessionName = sessionNameHelper;

            createPuttySession(sessionName,
                    null/*hostname*/, -1/*port*/, null/*puttyPrivateKeyFile*/, null/*username*/,
                    parameters.getProxyHostname(), parameters.getProxyPort(),
                    finalProxyPuttyPrivateKeyFile == null ? (puttyPrivateKeyFile == null ? null : puttyPrivateKeyFile.getFile()) : finalProxyPuttyPrivateKeyFile.getFile(),
                    parameters.getProxyUsername(),
                    parameters.isEnableX11Forward());

            //run session
            Runnable processOutputMonitor = () -> {
                Process p = null;
                try {
                    ArrayList<String> command = new ArrayList<>();
                    command.add(jFedPreferences.getFile(GuiPreferenceKey.PREF_PUTTY_DIRECTORY).getPath() + File.separator + "putty.exe");

                    command.add("-load");
                    command.add(sessionName);

                    command.add(parameters.getUsername() + "@" + parameters.getHostname());
                    command.add("-P");
                    command.add(Integer.toString(parameters.getPort()));

                    SshKeyInfo sshKeyInfo = parameters.getSshKeyInfo();
                    if (!usePageant && sshKeyInfo != null && puttyPrivateKeyFile != null) {
                        command.add("-i");
                        command.add(puttyPrivateKeyFile.getFilename());
                    }

                    LOG.debug("Starting PuTTY and waiting for it to stop");
                    LOG.info("PuTTY command: " + command);
                    p = Runtime.getRuntime().exec(command.toArray(new String[command.size()]));
                    BufferedReader input = new BufferedReader(new InputStreamReader(p.getInputStream()));
                    String line = null;
                    while ((line = input.readLine()) != null) {
                        LOG.debug("putty.exe output: " + line);
                    }
                    input.close();
                    try {
                        p.waitFor();
                    } catch (InterruptedException e) {
                        //ignore
                    }
                    LOG.debug("PuTTY stopped");
                } catch (IOException e) {
                    LOG.error("IOException while trying to launch external terminal: " + e.getMessage(), e);
                } finally {
                    LOG.debug("removing unencrypted putty key files");
                    if (puttyPrivateKeyFile != null) puttyPrivateKeyFile.delete();
                    if (finalProxyPuttyPrivateKeyFile != null) finalProxyPuttyPrivateKeyFile.delete();
                }
            };
            Thread t = new Thread(processOutputMonitor);
            t.setName("PuTTY-Thread");
            t.setDaemon(true); //=> this application may exit before putty stops
            t.start();
            LOG.debug("Started PuTTY thread.");
        } catch (IOException e) {
            if (puttyPrivateKeyFile != null) puttyPrivateKeyFile.delete();
            if (proxyPuttyPrivateKeyFile != null) proxyPuttyPrivateKeyFile.delete();
        }
    }

    /**
     * supports only without proxy
     * <p/>
     * Specifies target details on command line.
     * Specifies NO proxy details
     */
    public void runPuttyWithoutSession(final PuttyRunParameters parameters) {
        LOG.debug("Starting PuTTY without session.");
        final boolean usePageant = jFedPreferences.getBoolean(GuiPreferenceKey.PREF_PUTTY_PAGEANT, true);
        assert !parameters.hasProxy();

        if (usePageant)
            pageantHelper.registerKey(parameters.getSshKeyInfo());


        Runnable processOutputMonitor = () -> {
            Process p = null;
            TmpFile puttyPrivateKeyFile = null;
            try {
                ArrayList<String> command = new ArrayList<>();
                command.add(jFedPreferences.getFile(GuiPreferenceKey.PREF_PUTTY_DIRECTORY).getPath() + File.separator + "putty.exe");

                command.add("-ssh");
                command.add(parameters.getUsername() + "@" + parameters.getHostname());
                command.add("-P");
                command.add(Integer.toString(parameters.getPort()));
                if (parameters.isEnableX11Forward())
                    command.add("-X");

                if (!usePageant) {
                    puttyPrivateKeyFile = parameters.getSshKeyInfo() == null ? null : parameters.getSshKeyInfo().getPuttyKeyFile();
                    if (puttyPrivateKeyFile != null) {
                        puttyPrivateKeyFile.store();
                        command.add("-i");
                        command.add(puttyPrivateKeyFile.getFilename());
                    }
                }

                LOG.debug("Starting PuTTY and waiting for it to stop");
                LOG.info("PuTTY command: " + command);
                p = Runtime.getRuntime().exec(command.toArray(new String[command.size()]));
                BufferedReader input = new BufferedReader(new InputStreamReader(p.getInputStream()));
                String line = null;
                while ((line = input.readLine()) != null) {
                    LOG.debug("putty.exe output: " + line);
                }
                input.close();
                try {
                    p.waitFor();
                } catch (InterruptedException e) {
                    //ignore
                }
                LOG.debug("PuTTY stopped");
            } catch (IOException e) {
                LOG.error("IOException while trying to launch external terminal: " + e.getMessage(), e);
            } finally {
                LOG.debug("removing unencrypted putty key files");
                if (puttyPrivateKeyFile != null) puttyPrivateKeyFile.delete();
            }
        };
        Thread t = new Thread(processOutputMonitor);
        t.setName("PuTTY-Thread");
        t.setDaemon(true); //=> this application may exit before putty stops
        t.start();
        LOG.debug("Started PuTTY thread.");
    }

    /**
     * forgets password on return
     */
    public static void makePpkIfNeeded(GeniUser user, String pass) {
        if (pass == null) {
            LOG.debug("No pass, so not storing encrypted PPK for user " + user.getUserUrnString());
            return;
        }
        try {
            if (!userUrnToEncryptedPPK.containsKey(user.getUserUrnString())) {
                String comment = "PPK-generated-from-PEM-by-jFed";
                RSAPublicKey publicKey = (RSAPublicKey) user.getPublicKey();
                RSAPrivateCrtKey privateKey = (RSAPrivateCrtKey) user.getPrivateKey();
                String ppkContent = PuTTYPrivateKeyFile.createContent(comment, publicKey, privateKey, pass);
                assert ppkContent.contains("aes256-cbc");
                userUrnToEncryptedPPK.put(user.getUserUrnString(), ppkContent);
                LOG.debug("Now storing encrypted PPK for user " + user.getUserUrnString());
            }
        } catch (Exception e) {
            LOG.error("Error in makePpkIfNeeded. Can safely be ignored. message: " + e.getMessage(), e);
        }
    }

    public static String getUserEncryptedPPK(GeniUser user) {
        return userUrnToEncryptedPPK.get(user.getUserUrnString());
    }

    public static class TmpFileInfo {
        public final String tmpFileNameWithEnvVar;
        public final File tmpFile;

        private TmpFileInfo(String tmpFileNameWithEnvVar, File tmpFile) {
            this.tmpFileNameWithEnvVar = tmpFileNameWithEnvVar;
            this.tmpFile = tmpFile;
        }

        public static TmpFileInfo tmpFilePair(String baseName, String ext) {
            File tmpDir = new File(JFedUtils.getUserDataDirectoryFile(), "tmp");
            if (!tmpDir.exists()) tmpDir.mkdir();
            long nr = new Date().getTime();
            return new TmpFileInfo(
                    JFedUtils.getWindowsUserDataDirectoryString() + "tmp" + File.separator + baseName + nr + "." + ext,
                    new File(tmpDir, baseName + nr + "." + ext)
            );
        }
    }

    public static class PuttyRunParameters {
        private String hostname;
        private int port;
        private SshKeyInfo sshKeyInfo;
        private String username;

        private String proxyHostname;
        private int proxyPort;
        private SshKeyInfo proxySshKeyInfo;
        private String proxyUsername;

        private String proxyOpenSshKeyHostKey;

        public boolean enableX11Forward = false;

        public String getHostname() {
            return hostname;
        }

        public int getPort() {
            return port;
        }

        public SshKeyInfo getSshKeyInfo() {
            return sshKeyInfo;
        }

        public PrivateKey getPrivateKey() {
            return sshKeyInfo.getPrivateKey();
        }

        public PublicKey getPublicKey() {
            return sshKeyInfo.getPublicKey();
        }

        public String getUsername() {
            return username;
        }

        public String getProxyHostname() {
            return proxyHostname;
        }

        public int getProxyPort() {
            return proxyPort;
        }

        public SshKeyInfo getProxySshKeyInfo() {
            return proxySshKeyInfo;
        }

        public String getProxyUsername() {
            return proxyUsername;
        }

        public String getProxyOpenSshKeyHostKey() {
            return proxyOpenSshKeyHostKey;
        }


        public boolean hasProxy() {
            return proxyHostname != null;
        }

        public boolean samePrivateKeyForProxy() {
            if (sshKeyInfo == null) return false;
            if (sshKeyInfo.getPrivateKey() == null) return false;
            if (proxySshKeyInfo == null) return false;
            return sshKeyInfo.getPrivateKey().equals(proxySshKeyInfo.getPrivateKey());
        }


        public void setTarget(String hostname, int port, SshKeyInfo sshKeyInfo, String username) {
            if (hostname != null && ipv6pat.matcher(hostname).matches()) {
                //enclose IPv6 literal with square brackets to make it work
                this.hostname = "[" + hostname + "]";
            } else
                this.hostname = hostname;
            this.port = port;
            this.sshKeyInfo = sshKeyInfo;
            this.username = username;
        }

        public void setProxy(JFedConnection.ProxyInfo proxyInfo) {
            if (proxyInfo == null) {
                this.proxyHostname = null;
                this.proxyPort = -1;
                this.proxySshKeyInfo = null;
                this.proxyUsername = null;
                this.proxyOpenSshKeyHostKey = null;
            } else {
                assert proxyInfo instanceof JFedConnection.SshProxyInfo;
                JFedConnection.SshProxyInfo sshProxyInfo = (JFedConnection.SshProxyInfo) proxyInfo;
                if (sshProxyInfo.getHostname() != null && ipv6pat.matcher(sshProxyInfo.getHostname()).matches()) {
                    //enclose IPv6 literal with square brackets to make it work
                    this.proxyHostname = "[" + sshProxyInfo.getHostname() + "]";
                } else
                    this.proxyHostname = sshProxyInfo.getHostname();
                this.proxyPort = sshProxyInfo.getPort();
                this.proxySshKeyInfo = sshProxyInfo.getSshKeyInfo();
                this.proxyUsername = sshProxyInfo.getUsername();
                this.proxyOpenSshKeyHostKey = sshProxyInfo.getHostKey();
            }
        }

        public void setEnableX11Forward(boolean enableX11Forward) {
            this.enableX11Forward = enableX11Forward;
        }
        public boolean isEnableX11Forward() {
            return enableX11Forward;
        }
    }
}
