package be.iminds.ilabt.jfed.ui.javafx.dialogs;

import javafx.application.Platform;
import javafx.scene.control.Alert;
import javafx.scene.control.ButtonType;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Arrays;
import java.util.Collection;
import java.util.Optional;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.FutureTask;

/**
 * User: twalcari
 * Date: 8/21/2014
 * Time: 10:51
 */
public class ThreadingAwareDialogsUtil {
    private static final Logger LOG = LoggerFactory.getLogger(ThreadingAwareDialogsUtil.class);

    private ThreadingAwareDialogsUtil() {

    }

    /**
     * Shows information dialog.
     */
    public static void showInformation(final JFDialogs JFDialogs) {
        if (!Platform.isFxApplicationThread()) {
            Platform.runLater(JFDialogs::showInformation);
        } else {
            JFDialogs.showInformation();
        }
    }

    /**
     * Shows confirmation dialog.
     *
     * @return action used to close dialog.
     */
    public static Optional<ButtonType> showConfirm(final JFDialogs JFDialogs) {
        if (!Platform.isFxApplicationThread()) {
            FutureTask<Optional<ButtonType>> futureTask = new FutureTask<>(JFDialogs::showConfirm);
            Platform.runLater(futureTask);

            try {
                return futureTask.get();
            } catch (InterruptedException | ExecutionException e) {
                LOG.error("Error while showing confirm dialog", e);
                return null;
            }
        } else {
            return JFDialogs.showConfirm();
        }
    }

    /**
     * Shows warning dialog
     *
     * @return action used to close dialog
     */
    public static Optional<ButtonType> showWarning(final JFDialogs JFDialogs) {
        if (!Platform.isFxApplicationThread()) {
            FutureTask<Optional<ButtonType>> futureTask = new FutureTask<>(JFDialogs::showWarning);
            Platform.runLater(futureTask);

            try {
                return futureTask.get();
            } catch (InterruptedException | ExecutionException e) {
                LOG.error("Error while showing confirm dialog", e);
                return null;
            }
        } else {
            return JFDialogs.showWarning();
        }
    }

    /**
     * Show error dialog
     *
     * @return action used to close dialog
     */
    public static Optional<ButtonType> showError(final JFDialogs JFDialogs) {
        if (!Platform.isFxApplicationThread()) {
            FutureTask<Optional<ButtonType>> futureTask = new FutureTask<>(JFDialogs::showError);

            Platform.runLater(futureTask);

            try {
                return futureTask.get();
            } catch (InterruptedException | ExecutionException e) {
                LOG.error("Error while showing confirm dialog", e);
                return null;
            }
        } else {
            return JFDialogs.showError();
        }
    }

    /**
     * Shows exception dialog with expandable stack trace.
     *
     * @param exception exception to present
     * @return action used to close dialog
     */
    public static Optional<ButtonType> showException(final JFDialogs JFDialogs, final Throwable exception) {
        if (!Platform.isFxApplicationThread()) {
            FutureTask<Optional<ButtonType>> futureTask = new FutureTask<>(() -> JFDialogs.showException(exception));

            Platform.runLater(futureTask);

            try {
                return futureTask.get();
            } catch (InterruptedException | ExecutionException e) {
                LOG.error("Error while showing confirm dialog", e);
                return null;
            }
        } else {
            return JFDialogs.showException(exception);
        }
    }

    /**
     * Shows dialog with one text field
     *
     * @param defaultValue text field default value
     * @return Optional of text from input field if OK action is used otherwise Optional.EMPTY
     */
    public static Optional<String> showTextInput(final JFDialogs JFDialogs, final String defaultValue) {
        if (!Platform.isFxApplicationThread()) {
            FutureTask<Optional<String>> futureTask = new FutureTask<>(() -> JFDialogs.showTextInput(defaultValue));
            Platform.runLater(futureTask);

            try {
                return futureTask.get();
            } catch (InterruptedException | ExecutionException e) {
                LOG.error("Error while showing confirm dialog", e);
                return null;
            }
        } else {
            return JFDialogs.showTextInput(defaultValue);
        }
    }

    /**
     * Shows dialog with one text field
     *
     * @return Optional of text from input field or Optional.EMPTY if dialog is cancelled
     */
    public static Optional<String> showTextInput(JFDialogs JFDialogs) {
        return showTextInput(JFDialogs, "");
    }

    /**
     * Show a dialog with one combobox filled with provided choices. The combobox selection
     * will be set to a default value if one is provided.
     *
     * @param defaultChoice default combobox selection
     * @param choices       dialog choices
     * @return Optional of selected choice or Optional.EMPTY if dialog is cancelled.
     */
    public static <T> Optional<T> showChoices(final JFDialogs JFDialogs, final T defaultChoice, final Collection<T> choices) {
        if (!Platform.isFxApplicationThread()) {
            FutureTask<Optional<T>> futureTask = new FutureTask<>(() -> JFDialogs.showChoices(defaultChoice, choices));

            Platform.runLater(futureTask);

            try {
                return futureTask.get();
            } catch (InterruptedException | ExecutionException e) {
                LOG.error("Error while showing confirm dialog", e);
                return null;
            }
        } else {
            return JFDialogs.showChoices(defaultChoice, choices);
        }
    }

    /**
     * Show a dialog with one combobox filled with provided choices
     *
     * @param choices dialog choices
     * @return Optional of selected choice or Optinal.EMPTY if dialog is cancelled
     */
    public static <T> Optional<T> showChoices(JFDialogs JFDialogs, Collection<T> choices) {
        return showChoices(JFDialogs, null, choices);
    }

    /**
     * Show a dialog with one combobox filled with provided choices
     *
     * @param choices dialog choices
     * @return Optional of selected choice or Optional.EMPTY if dialog is cancelled
     */
    public static <T> Optional<T> showChoices(JFDialogs JFDialogs, @SuppressWarnings("unchecked") T... choices) {
        return showChoices(JFDialogs, Arrays.asList(choices));
    }

    public static Optional<ButtonType> showAlert(Alert alert) {
        if (!Platform.isFxApplicationThread()) {
            FutureTask<Optional<ButtonType>> futureTask =
                    new FutureTask<>(alert::showAndWait);

            Platform.runLater(futureTask);

            try {
                return futureTask.get();
            } catch (InterruptedException | ExecutionException e) {
                LOG.error("Error while showing confirm dialog", e);
                return null;
            }
        } else {
            return alert.showAndWait();
        }
    }

}
