package be.iminds.ilabt.jfed.call_log_output;

import be.iminds.ilabt.jfed.util.IOUtils;
import org.slf4j.LoggerFactory;

import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlRootElement;
import javax.xml.bind.annotation.XmlType;
import java.io.*;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * LogOutput
 */
public class LogOutput {
    protected static final org.slf4j.Logger LOG = LoggerFactory.getLogger(LogOutput.class);
    protected static boolean disableRemoveSecurityRisks = false;

    public static List<SecurityRiskDescription> getAllSecurityRiskDescriptions() {
        List<SecurityRiskDescription> res = new ArrayList<SecurityRiskDescription>();
        res.add(new SecurityRiskDescription("<signed-credential", "</signed-credential>"));
        res.add(new SecurityRiskDescription("<owner_gid", "</owner_gid>"));
        res.add(new SecurityRiskDescription("<target_gid", "</target_gid>"));
        res.add(new SecurityRiskDescription("<X509Certificate", "<X509Certificate>"));

        res.add(new SecurityRiskDescription("&lt;signed-credential", "&lt;/signed-credential&gt;"));
        res.add(new SecurityRiskDescription("&lt;owner_gid", "&lt;/owner_gid&gt;"));
        res.add(new SecurityRiskDescription("&lt;target_gid", "&lt;/target_gid&gt;"));
        res.add(new SecurityRiskDescription("&lt;X509Certificate", "&lt;X509Certificate&gt;"));

        res.add(new SecurityRiskDescription("-----BEGIN RSA PRIVATE KEY-----", "-----END RSA PRIVATE KEY-----"));

        res.add(new SecurityRiskDescription("MIID", "[^A-Za-z0-9+/=._: \n-]")); //PEM Base64 certificate block detection (starts with MIID due to header)

//        res.add(new SecurityRiskDescription("", ""));
        return res;
    }

    public void copyToFile(URL in, File fileOut) throws IOException {
        assert in != null;
        assert fileOut != null;
        OutputStream out = new FileOutputStream(fileOut);
        InputStream is = in.openStream();
        byte[] buffer = new byte[1024];
        int len;
        while ((len = is.read(buffer)) != -1) {
            out.write(buffer, 0, len);
        }
        out.close();
    }

    /* Scans the input string for credentials, and removes them. */
    public String removeSecurityRisks(String text) {
        if (disableRemoveSecurityRisks) return text;
        if (text == null) return null;

        StringBuilder sb = new StringBuilder();
        List<SecurityRiskDescription> secRisks = getAllSecurityRiskDescriptions();

//        System.out.println("Scanning string of size "+text.length()+" for "+secRisks.size()+" security risks.");

        while (text.length() > 0) {
            int firstStringIndex = -1;
            int secRiskIndex = -1;
            int i = 0;
            for (SecurityRiskDescription secRisk : secRisks) {
                int stringIndex = text.indexOf(secRisk.start);
                if (stringIndex >= 0 && (stringIndex < firstStringIndex || firstStringIndex == -1)) {
                    firstStringIndex = stringIndex;
                    secRiskIndex = i;
                }
                i++;
            }

            if (secRiskIndex != -1) {
                //add part before security risk
                String safePart = text.substring(0, firstStringIndex);
                sb.append(safePart);
                text = text.substring(firstStringIndex);

                //cut out security risk
                SecurityRiskDescription secRisk = secRisks.get(secRiskIndex);
//                System.out.println("   ********* Found security risk: "+secRisk.start+" @ position "+firstStringIndex);
                Pattern pattern = Pattern.compile(secRisk.end);
                Matcher matcher = pattern.matcher(text);
                if (matcher.find()) {
                    //int startIndex = matcher.start();
                    int endIndex = matcher.end();
                    String unsafe = text.substring(0, endIndex);
                    text = text.substring(endIndex);
                    sb.append("**** SECURITY: removed text that looked like a credential. (" + unsafe.length() + " characters) ****");
                } else {
                    //problem: out of paranoia, lose the rest of the string
                    sb.append("**** SECURITY: removed text that looked like a credential. WARNING: could not find end, so removed rest of text (" + text.length() + " characters) ****");
                    text = "";
                }

            } else {
//                System.out.println("   No (more) security risks found.");
                //add entire string
                sb.append(text);
                text = "";
            }
        }

        return sb.toString();
    }

    public enum TestResultState {SUCCESS, WARN, SKIPPED, FAILED}

    public enum LogLineType {DEBUG, NOTE, WARN, ERROR, FATAL_ERROR, EXCEPTION}

    @XmlRootElement(name = "logentry")
    @XmlType(name="logentry")
    public static class LogLine {
        @XmlElement(name = "type")
        final LogLineType type;

        @XmlElement(name = "text")
        final String text;

        @XmlElement(name = "timeInMs")
        final long timeMs;

//        final Throwable exception;
        @XmlElement(name = "exception")
        final String exception;

        public LogLine() {
         /* only to be used by JAXB */
            this.type = null;
            this.text = null;
            this.exception = null;
            this.timeMs = -1;
        }

        public LogLine(LogLineType type, String text, long timeMs) {
            this(type, text, timeMs, (Throwable) null);
        }

        public LogLine(LogLineType type, String text, long timeMs, String exception) {
            assert type != null;
            assert text != null;
            this.type = type;
            this.text = text;
            this.timeMs = timeMs;
            this.exception = exception;
        }

        public LogLine(LogLineType type, String text, long timeMs, Throwable exception) {
            assert type != null;
            assert text != null;
            this.type = type;
            this.text = text;
            this.timeMs = timeMs;
            this.exception = exception == null ? null : IOUtils.exceptionToStacktraceString(exception);
        }

        public LogLine(LogLineType type, String text) {
            this(type, text, null);
        }
        public LogLine(LogLineType type, String text, Throwable exception) {
            this(type, text, System.currentTimeMillis(), exception);
        }

        public LogLineType getType() {
            return type;
        }

        public String getText() {
            return text;
        }

        public long getTimeMs() {
            return timeMs;
        }

        public String getExceptionStackTrace() {
            return exception;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (!(o instanceof LogLine)) return false;

            LogLine logLine = (LogLine) o;

            if (timeMs != logLine.timeMs) return false;
            if (exception != null ? !exception.equals(logLine.exception) : logLine.exception != null) return false;
            if (text != null ? !text.equals(logLine.text) : logLine.text != null) return false;
            if (type != logLine.type) return false;

            return true;
        }

        @Override
        public int hashCode() {
            int result = type.hashCode();
            result = 31 * result + (text != null ? text.hashCode() : 0);
            result = 31 * result + (int) (timeMs ^ (timeMs >>> 32));
            result = 31 * result + (exception != null ? exception.hashCode() : 0);
            return result;
        }
    }

    private static class SecurityRiskDescription {
        public String start;
        public String end;

        private SecurityRiskDescription(String start, String end) {
            this.start = start;
            this.end = end;
        }
    }
}
