package be.iminds.ilabt.jfed.experimenter_gui.dialogs;

import be.iminds.ilabt.jfed.experimenter_gui.call_gui.TasksWindow;
import be.iminds.ilabt.jfed.experimenter_gui.slice.ExperimentController;
import be.iminds.ilabt.jfed.experimenter_gui.slice.tasks.ExperimentTasksFactory;
import be.iminds.ilabt.jfed.highlevel.controller.HighLevelController;
import be.iminds.ilabt.jfed.highlevel.model.Slice;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.ServerType;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.ui.javafx.GlyphUtils;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import be.iminds.ilabt.jfed.util.GeniUrn;
import javafx.event.ActionEvent;
import javafx.fxml.FXML;
import javafx.fxml.FXMLLoader;
import javafx.scene.control.CheckBox;
import javafx.scene.control.Label;
import javafx.scene.control.TextArea;
import javafx.scene.control.TextField;
import javafx.scene.layout.VBox;
import javafx.scene.paint.Color;
import javafx.stage.Window;
import org.controlsfx.dialog.DialogAction;
import org.controlsfx.glyphfont.FontAwesome;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.function.Consumer;

/**
 */
public class ShareSliceDialog extends VBox implements Consumer<ActionEvent> {
    private static final Logger LOG = LoggerFactory.getLogger(ShareSliceDialog.class);
    private static final String SIMPLE_PROPERTIES_FXML = "ShareSliceDialog.fxml";

    private final Slice slice;
    private final ExperimentController experimentController;
    private final ExperimentTasksFactory experimentTasksFactory;

    private final GeniUserProvider geniUserProvider;
    private final HighLevelController hlc;

    @FXML
    private TextArea userUrnsField;

    @FXML
    private TextField sliceUrnField;

    @FXML
    private Label helpLabel;

    @FXML
    private CheckBox addSshKeysCheckBox;
    private final TasksWindow tasksWindow;


    private final org.controlsfx.dialog.Dialog dialog;

    ShareSliceDialog(Window parentWindow, ExperimentController experimentController, ExperimentTasksFactory experimentTasksFactory,
                     GeniUserProvider geniUserProvider, HighLevelController hlc, TasksWindow tasksWindow) {
        this.experimentController = experimentController;
        this.experimentTasksFactory = experimentTasksFactory;
        this.geniUserProvider = geniUserProvider;
        this.hlc = hlc;
        this.tasksWindow = tasksWindow;
        this.slice = experimentController.getExperiment().getSlice();

        //initialize the Scene
        FXMLLoader fxmlLoader = new FXMLLoader(getClass().getResource(SIMPLE_PROPERTIES_FXML));

        fxmlLoader.setRoot(this);
        fxmlLoader.setController(this);

        try {
            fxmlLoader.load();
        } catch (IOException exception) {
            throw new RuntimeException(exception);
        }

        GeniUrn curUserUrn = geniUserProvider.getLoggedInGeniUser().getUserUrn();
        GeniUrn genericUserUrn = GeniUrn.createGeniUrnFromEncodedParts(curUserUrn.getEncodedTopLevelAuthority(), "user", "USERNAME");
        userUrnsField.setText(genericUserUrn.getValue() + "\n");
        sliceUrnField.setText(slice.getUrnString());

        addSshKeysCheckBox.managedProperty().bind(addSshKeysCheckBox.visibleProperty());

        SfaAuthority userAuth = geniUserProvider.getLoggedInGeniUser().getUserAuthority();
        boolean hasFedMAApiv2 = userAuth.getUrl(ServerType.GeniServerRole.GENI_CH_MA, 2) != null;
        boolean hasFedSAApiv2 = userAuth.getUrl(ServerType.GeniServerRole.GENI_CH_SA, 2) != null;

        //TODO: this is not even a guarantee, KEY needs to be supported by server as well.
        boolean supportsAutoAddUser = hasFedMAApiv2;

        addSshKeysCheckBox.setVisible(supportsAutoAddUser);
        addSshKeysCheckBox.setSelected(supportsAutoAddUser);

        boolean usersSeeSharedSlices = hasFedSAApiv2;
        String sliceProject = slice.getSubAuthorityName();
        String authName = userAuth.getName();

        String helptext =
                "Sharing an experiment has some limitations:\n" +
                        "- You can only share with users of the same authority (" + authName + ")\n" +
                        (sliceProject == null ? ""
                                : "- You can only share with users that are member of the \"" + sliceProject + "\" project in which this experiment was created.\n") +
                        (usersSeeSharedSlices ?
                                "- The experiment appear in the list of experiments seen by the other users.\n" +
                                        "  So they can just use the 'Recover' button and select the experiment.\n"
                                : "- The experiment will not appear in the list of experiments seen by the other users.\n" +
                                        "  These users will need to use the experiment URN \"" + slice.getUrnString() + "\" with the 'Open Shared' button.\n") +
                        (supportsAutoAddUser ?
                                "- The SSH of the new users can automatically be added to all nodes, if you enable the option.\n"
                                :
                                        "- The SSH keys of the new users cannot be automatically put on the nodes in the experiment.\n" +
                                                "  However, the user can add them themselves with the 'Edit ssh-keys' option\n");
//                        +"\n" +
//                        "(Note that jFed automatically changes the above list based on the account used to login)\n";
        helpLabel.setText(helptext);

        //initialize dialog
        dialog = new org.controlsfx.dialog.Dialog(parentWindow, "Share Experiment");
        dialog.setMasthead("Share your experiment");
        dialog.setGraphic(GlyphUtils.createDialogGlyph(FontAwesome.Glyph.AMBULANCE, Color.GREEN));
        dialog.setContent(this);
        dialog.getActions().setAll(new DialogAction("Share", org.controlsfx.control.ButtonBar.ButtonType.OK_DONE, false, false, true, this),
                org.controlsfx.dialog.Dialog.ACTION_CANCEL);

    }

    public void showDialog() {
        dialog.show();
    }

    @Override
    public void accept(ActionEvent actionEvent) {
        List<String> invalidLines = new ArrayList<>();

        String userUrnsText = userUrnsField.getText();
        List<GeniUrn> userUrns = new ArrayList<>();
        for (String line : userUrnsText.split("\n")) {
            if (line.trim().startsWith("urn:")) {
                GeniUrn userUrn = GeniUrn.parse(line.trim());
                if (userUrn != null)
                    userUrns.add(userUrn);
                else
                    invalidLines.add(line.trim());

            } else if (!line.trim().isEmpty())
                invalidLines.add(line.trim());
        }

        if (!invalidLines.isEmpty()) {
            String errorText = "Invalid URNs:\n";
            for (String line : invalidLines)
                errorText += line + "\n";
            JFDialogs.create()
                    .owner(this.getScene().getWindow())
                    .message(errorText)
                    .masthead("Field contains Invalid URNs.")
                    .showError();

            dialog.shake();
        } else if (userUrns.isEmpty()) {
            JFDialogs.create()
                    .owner(this.getScene().getWindow())
                    .message("No valid URNs have been provided.")
                    .masthead("Field contains no valid URNs")
                    .showError();
            dialog.shake();
        } else {
            experimentController.submitExperimentTaskGenerator(experimentTasksFactory.createShareSliceTasksFactory(experimentController, userUrns, addSshKeysCheckBox.isSelected()));
            dialog.hide();
        }
    }
}
