package be.iminds.ilabt.jfed.experimenter_gui.login;

import be.iminds.ilabt.jfed.experimenter_gui.connectivity_tester.ConnectivityTesterGUI;
import be.iminds.ilabt.jfed.experimenter_gui.ui.status.TaskStatusIndicator;
import be.iminds.ilabt.jfed.highlevel.model.AuthorityList;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.lowlevel.userloginmodel.InvalidLoginException;
import be.iminds.ilabt.jfed.lowlevel.userloginmodel.KeyCertUserLoginModel;
import be.iminds.ilabt.jfed.lowlevel.userloginmodel.PlanetlabUserLoginModel;
import be.iminds.ilabt.jfed.lowlevel.userloginmodel.UserLoginModelManager;
import be.iminds.ilabt.jfed.preferences.GuiPreferenceKey;
import be.iminds.ilabt.jfed.preferences.JFedGuiPreferences;
import be.iminds.ilabt.jfed.ssh_terminal_tool.putty.PuttyHelper;
import be.iminds.ilabt.jfed.ui.javafx.GlyphUtils;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import be.iminds.ilabt.jfed.ui.javafx.userlogin.UserLoginFactory;
import be.iminds.ilabt.jfed.util.GeniUrn;
import be.iminds.ilabt.jfed.util.IOUtils;
import be.iminds.ilabt.jfed.util.JFedCommonDialogs;
import be.iminds.ilabt.jfed.util.ProxyPreferencesManager;
import javafx.application.Platform;
import javafx.event.ActionEvent;
import javafx.fxml.FXML;
import javafx.scene.Node;
import javafx.scene.control.*;
import javafx.scene.layout.HBox;
import javafx.scene.paint.Color;
import javafx.stage.FileChooser;
import javafx.stage.Stage;
import org.controlsfx.glyphfont.FontAwesome;
import org.slf4j.LoggerFactory;

import javax.inject.Inject;
import javax.inject.Provider;
import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.security.cert.X509Certificate;
import java.text.SimpleDateFormat;
import java.time.Duration;
import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

/**
 * This logincontroller supports only one type of login: KEY_CERT_INTERNAL_INFO
 * <p/>
 * User: twalcari
 * Date: 10/31/13
 * Time: 5:27 PM
 */

public class UserLoginController {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(UserLoginController.class);
    private static final String CUSTOM_INPUT = "<CUSTOM INPUT>";
    private static final String USER_LOGIN_FXML = "UserLogin.fxml";
    private static final Duration EXPIRATION_WARNING_DURATION = Duration.ofDays(30);


    @FXML
    private Label usernameLabel;
    @FXML
    private Label authorityLabel;
    @FXML
    private Label expireLabel;
    @FXML
    private TextField fileInput;
    @FXML
    private Button fileBrowseButton;
    @FXML
    private PasswordField passwordField;
    @FXML
    private Label messageLabel;
    @FXML
    private Button advancedButton;
    @FXML
    private Button resetButton;
    @FXML
    private HBox bottomBar;
    @FXML
    private Button loginButton;
    @FXML
    private TaskStatusIndicator statusIndicator;

    private final UserLoginModelManager userLoginModelManager;

    private final UserLoginFactory advancedUserLoginFactory;

    private final Provider<ConnectivityTesterGUI> connectivityTesterGUIProvider;

    private final ProxyPreferencesManager proxyPreferencesManager;

    private final AuthorityList authorityList;

    private final JFedGuiPreferences jFedPreferences;

    @Inject
    public UserLoginController(UserLoginFactory advancedUserLoginFactory, UserLoginModelManager userLoginModelManager,
                               Provider<ConnectivityTesterGUI> connectivityTesterGUIProvider,
                               ProxyPreferencesManager proxyPreferencesManager, AuthorityList authorityList,
                               JFedGuiPreferences jFedPreferences) {
        this.advancedUserLoginFactory = advancedUserLoginFactory;
        this.userLoginModelManager = userLoginModelManager;
        this.connectivityTesterGUIProvider = connectivityTesterGUIProvider;
        this.proxyPreferencesManager = proxyPreferencesManager;
        this.authorityList = authorityList;
        this.jFedPreferences = jFedPreferences;
    }

    private boolean success = false;
    private static final SimpleDateFormat DATE_FORMAT = new SimpleDateFormat("yyyy-MM-dd");

    @FXML
    public void initialize() {

        //check if the current loginmodel in the UserLoginModelManager is a supported type, otherwise revert to one
        if (userLoginModelManager.getUserLoginModelType() != UserLoginModelManager.UserLoginModelType.KEY_CERT_INTERNAL_INFO
                && userLoginModelManager.getUserLoginModelType() != UserLoginModelManager.UserLoginModelType.PLANETLAB) {

            JFDialogs.create()
                    .title("Unsupported login-model detected")
                    .masthead("Unsupported login-model detected")
                    .message(String.format("Only KEY_CERT_INTERNAL_INFO and PLANETLAB(+sfahrn etc) are supported, but got %s.\n" +
                                    "jFed will now revert to the default login-method.",
                            userLoginModelManager.getUserLoginModelType().name())).showError();

            userLoginModelManager.setUserLoginModelType(UserLoginModelManager.UserLoginModelType.KEY_CERT_INTERNAL_INFO);
        }

        switch (userLoginModelManager.getUserLoginModelType()) {
            case KEY_CERT_INTERNAL_INFO: {
                KeyCertUserLoginModel m = userLoginModelManager.getKeyCertUserLoginModel();
                if (m.isPemGiven()) {
                    modelToGui();

                    if (m.isPasswordRequired()) {
                        setMessage("Enter the password associated with the certificate", MessageType.INFO);

                        Platform.runLater(passwordField::requestFocus);
                    } else {
                        setMessage("No password required: You can login", MessageType.INFO);
                        Platform.runLater(loginButton::requestFocus);
                    }
                }
                break;
            }
            case PLANETLAB: {
                PlanetlabUserLoginModel m = userLoginModelManager.getPlanetlabUserLoginModel();
                if (m.isPasswordRequired()) {
                    setMessage("Enter the password associated with the private key", MessageType.INFO);
                    Platform.runLater(passwordField::requestFocus);
                } else {
                    if (m.isReadyToFetchCertificate()) {
                        try {
                            m.fetchCertificate();
                            setMessage("No password required: You can login", MessageType.INFO);
                        } catch (PlanetlabUserLoginModel.CertificateFetchException ex) {
                            LOG.error("Failed to fetch certificate", ex);
                            setMessage("Could not fetch certificate. Contact your authority.", MessageType.ERROR);
                        }
                        modelToGui();

                        Platform.runLater(passwordField::requestFocus);
                    } else {
                        modelToGui();
                        setMessage("No password required, but missing internal info needed to login.", MessageType.INFO);

                        Platform.runLater(loginButton::requestFocus);
                    }
                }
                break;
            }
        }

        fileInput.focusedProperty().addListener((observableValue, oldValue, newValue) -> {

            //when the input-field loses focus, it should process it's content
            if (oldValue && !newValue) {
                onFileInputAction();
            }

        });
    }

    private void setMessage(String message, MessageType type) {
        messageLabel.setText(message);

        for (MessageType t : MessageType.values())
            messageLabel.getStyleClass().remove(t.getStyleName());

        messageLabel.getStyleClass().add(type.getStyleName());
    }

    private Date getEarliestExpire(List<X509Certificate> certs) {
        if (certs == null)
            return null;

        Date res = null;
        for (X509Certificate cert : certs) {
            Date expire = cert.getNotAfter();
            if (res == null || res.after(expire))
                res = expire;
        }
        return res;
    }

    private void setExpireLabel(Date certExpire) {
        if (certExpire == null) {
            expireLabel.setText("-");
            expireLabel.setGraphic(null);
            return;
        }

        String dateString = DATE_FORMAT.format(certExpire);
        expireLabel.setText(dateString);

        Duration durationToExpiration = Duration.between(Instant.now(), certExpire.toInstant());

        if (durationToExpiration.isNegative()) {
            Node errorGlyph = GlyphUtils.createSmallGlyph(FontAwesome.Glyph.TIMES_CIRCLE, Color.RED);
            Tooltip.install(errorGlyph, new Tooltip("Your certificate has expired! Please select a valid certificate."));
            expireLabel.setGraphic(errorGlyph);
        } else if (durationToExpiration.compareTo(EXPIRATION_WARNING_DURATION) < 0) {
            Node warningGlyph = GlyphUtils.createSmallGlyph(FontAwesome.Glyph.WARNING, Color.YELLOW);
            Tooltip.install(warningGlyph,
                    new Tooltip(
                            String.format("Your certificate will expire in %d days! " +
                                            "Please contact your authority to get a new certificate.",
                                    ChronoUnit.DAYS.between(Instant.now(), certExpire.toInstant()))));
            expireLabel.setGraphic(warningGlyph);
        } else {
            Node checkGlyph = GlyphUtils.createSmallGlyph(FontAwesome.Glyph.CHECK, Color.GREEN);
            Tooltip.install(checkGlyph, new Tooltip("The expiry date of your certificate is sufficiently far in the future."));
            expireLabel.setGraphic(checkGlyph);
        }
    }

    private void readCertificate() {
        assert userLoginModelManager.getUserLoginModelType() == UserLoginModelManager.UserLoginModelType.KEY_CERT_INTERNAL_INFO
                ||
                userLoginModelManager.getUserLoginModelType() == UserLoginModelManager.UserLoginModelType.PLANETLAB;

        switch (userLoginModelManager.getUserLoginModelType()) {
            case KEY_CERT_INTERNAL_INFO: {
                KeyCertUserLoginModel m = userLoginModelManager.getKeyCertUserLoginModel();
                if (m.isPemGiven() && m.getUserUrnString() != null) {
                    //derive username from userURN
                    GeniUrn userUrn = GeniUrn.parse(m.getUserUrnString());
                    usernameLabel.setText(userUrn.getEncodedResourceName());

                    //get user-friendly authority name
                    if (m.getUserAuthority() == null)
                        throw new RuntimeException("Could not fetch User Authority. Are you connected to the Internet?");

                    authorityLabel.setText(m.getUserAuthority().getName());

                    Date certExpire = getEarliestExpire(m.getClientCertificateChain());
                    setExpireLabel(certExpire);

                    // disable password field if no password is needed
                    if (!m.isPasswordRequired()) {
                        passwordField.setDisable(true);
                        passwordField.setText("");
                    } else {
                        passwordField.setDisable(false);
                    }

                } else {
                    //reset to empty
                    clear();
                }

                break;
            }
            case PLANETLAB: {
                PlanetlabUserLoginModel m = userLoginModelManager.getPlanetlabUserLoginModel();
                if (m.correctAuthority()) {
                    //get user-friendly authority name
                    if (m.getUserAuthority() == null)
                        throw new RuntimeException("Could not fetch User Authority. Are you connected to the Internet?");
                    authorityLabel.setText(m.getUserAuthority().getName());
                } else {
                    authorityLabel.setText("unknown");
                }

                Date certExpire = getEarliestExpire(m.getClientCertificateChain());
                setExpireLabel(certExpire);

                if (m.getUserUrnString() != null) {
                    //derive username from userURN
                    GeniUrn userUrn = GeniUrn.parse(m.getUserUrnString());
                    usernameLabel.setText(userUrn.getEncodedResourceName());
                } else {
                    //reset to empty
                    //clear();

                    //not downloaded automatically, as that would require a thread which does this in the background.
                    //and all extra complication that goes with that. And it's not that important to have the username...

//                    if (m.isPasswordRequired())
                    usernameLabel.setText("(known when login is pressed)");
//                    else
//                        usernameLabel.setText("fetching...");
                }

                // disable password field if no password is needed
                if (!m.isPasswordRequired()) {
                    passwordField.setDisable(true);
                    passwordField.setText("");
                } else {
                    passwordField.setDisable(false);
                }

                break;
            }
        }
    }

    /**
     * Resets the login-screen to the empty state.
     */
    private void clear() {
        fileInput.setText("");
        usernameLabel.setText("-");
        authorityLabel.setText("-");
        expireLabel.setText("-");
        setMessage("Please select your user certificate", MessageType.INFO);
    }

    public void unlockAndLogin() {
        final char[] password = passwordField.getText().toCharArray();

        boolean unlocked;
        SfaAuthority userAuthority;
        switch (userLoginModelManager.getUserLoginModelType()) {
            case KEY_CERT_INTERNAL_INFO: {
                KeyCertUserLoginModel m = userLoginModelManager.getKeyCertUserLoginModel();
                unlocked = m.unlock(password);
                userAuthority = m.getUserAuthority();
                break;
            }
            case PLANETLAB: {
                PlanetlabUserLoginModel m = userLoginModelManager.getPlanetlabUserLoginModel();
                unlocked = m.unlock(password);

                if (unlocked) {
                    if (m.isReadyToFetchCertificate()) {
                        setMessage("Password correct, trying to fetch user certificate from server...\n", MessageType.INFO);

                        try {
                            m.fetchCertificate();
                        } catch (PlanetlabUserLoginModel.CertificateFetchException ex) {
                            unlocked = false;
                            LOG.error("Fetching user certificate failed.", ex);
                            setMessage("Password correct, but fetching user certificate from server failed.\n", MessageType.ERROR);

                            JFDialogs.create()
                                    .owner(null)
                                    .message("Your password is correct. However, jFed could not fetch the user certificate from the authority. Contact support for more information.")
                                    .masthead("Login Failed")
                                    .title("Login Failed")
                                    .showException(ex);

                            return;
                        }

                        setMessage("Password correct and user certificate successfully fetched from server.\n", MessageType.INFO);

                    } else {
                        LOG.warn("No password required, but missing internal info needed to login:" + "\n" +
                                        " keyfile=" + m.getPrivateKeyFile() + "\n" +
                                        " sfa hrn=" + m.getSfaHrn() + "\n" +
                                        " user auth hrn=" + m.getUserAuthorityUrn() + "\n" +
                                        " isPasswordRequired=" + m.isPasswordRequired() + "\n" +
                                        " isReadyToFetchCertificate=" + m.isReadyToFetchCertificate() + "\n" +
                                        " correctAuthority=" + m.correctAuthority() + "\n" +
                                        " correctPlanetlabSfaUrn=" + m.correctPlanetlabSfaUrn()
                        );
                        setMessage("No password required, but missing internal info needed to login.", MessageType.ERROR);
                        return;
                    }
                }
                userAuthority = m.getUserAuthority();
                break;
            }
            default:
                throw new RuntimeException("Unsupported type: " + userLoginModelManager.getUserLoginModelType());
        }

        if (unlocked) {
            LOG.debug("Successfully decrypted private keys");

            setMessage("Password OK!\nLogging in...", MessageType.INFO);

            boolean success = userLoginModelManager.login();
            LOG.debug("Login success=" + success);
            if (success) {
                try {
                    if (!userLoginModelManager.checkLoginValidity())
                        throw new InvalidLoginException("Unspecified Invalid Login");
                } catch (InvalidLoginException e) {
                    LOG.warn("userLoginModelManager.checkLoginValidity() failed with InvalidLoginException: " + e.getMessage(), e);
                    setMessage("Invalid Login cert/key: " + e.getMessage(), MessageType.ERROR);


                    if (JFDialogs.create()
                            .owner(this)
                            .message(
                                    "jFed detected a problem with your login PEM file:\n\n" +
                                            "   " + e.getMessage() + "\n\n" +
                                            "Continuing will likely result in failed calls.\n" +
                                            "Do you want to continue anyway?")
                            .title("Continue login?")
                            .showConfirm().filter(r -> r == ButtonType.YES).isPresent())
                        return;
                }

                //create a PPK encrypted with the same password, and store that PPK. Then forget the password.
                if (!passwordField.getText().isEmpty())
                    PuttyHelper.makePpkIfNeeded(userLoginModelManager.getLoggedInGeniUser(), passwordField.getText());

                //always save latest successful login details
                if (userLoginModelManager.hasChanged()) {
                    userLoginModelManager.save();
                }
            } else {
                LOG.info("userLoginModelManager.login() failed." +
                        " userLoginModelType=" + userLoginModelManager.getUserLoginModelType() +
                        " userLoginModelManager.getCurrentUserLoginModel().isUserLoggedIn()=" + userLoginModelManager.getCurrentUserLoginModel().isUserLoggedIn());
                setMessage("Password is correct, but login failed. (this should not happen, please send a bugreport).\n", MessageType.ERROR);

                return;
            }

            setMessage("Password OK!\nChecking connection with server...", MessageType.INFO);

            statusIndicator.setStatus(TaskStatusIndicator.Status.BUSY);
            statusIndicator.setManaged(true);
            statusIndicator.setVisible(true);

            //now test connectivity before continuing
            final LoginConnectivityTest loginConnectivityTest =
                    new LoginConnectivityTest(userAuthority, proxyPreferencesManager, userLoginModelManager, authorityList, jFedPreferences) {
                        @Override
                        protected void busyMessage(final String message) {
                            Platform.runLater(() -> {
                                setMessage(message, MessageType.INFO);
                                statusIndicator.setStatus(TaskStatusIndicator.Status.BUSY);
                                statusIndicator.setManaged(true);
                                statusIndicator.setVisible(true);
                            });
                        }

                        @Override
                        protected void okMessage(final String message) {
                            Platform.runLater(() -> {
                                setMessage(message, MessageType.INFO);
                                statusIndicator.setStatus(TaskStatusIndicator.Status.SUCCESS);
                                statusIndicator.setManaged(true);
                                statusIndicator.setVisible(true);
                            });
                        }

                        @Override
                        protected void errorMessage(final String message) {
                            Platform.runLater(() -> {
                                setMessage(message, MessageType.ERROR);
                                statusIndicator.setStatus(TaskStatusIndicator.Status.FAILED);
                                statusIndicator.setManaged(true);
                                statusIndicator.setVisible(true);
                            });
                        }

                        @Override
                        protected void connectivityTestFinished(final boolean success, final boolean abortLogin) {
                            Platform.runLater(() -> {
                                if (!abortLogin) {
                                    if (success)
                                        setMessage("Connectivity OK, logging in.", MessageType.INFO);
                                    else
                                        setMessage("Connectivity test failed. But logging in anyway!", MessageType.WARNING);
                                    statusIndicator.setStatus(TaskStatusIndicator.Status.SUCCESS);
                                    statusIndicator.setManaged(true);
                                    statusIndicator.setVisible(true);

                                    closeLoginAndProceed();
                                } else {
                                    if (success)
                                        setMessage("Connectivity OK, but user aborted login.", MessageType.WARNING);
                                    else
                                        setMessage("Connectivity test failed. Login Aborted.", MessageType.ERROR);
                                    statusIndicator.setStatus(TaskStatusIndicator.Status.FAILED);
                                    statusIndicator.setManaged(true);
                                    statusIndicator.setVisible(true);
                                }
                            });
                        }

                        @Override
                        protected Stage getDialogStage() {
                            return (Stage) usernameLabel.getScene().getWindow();
                        }
                    };
            loginConnectivityTest.start();
        } else {
            LOG.info("Unlocking user private key failed.");
            setMessage("Password isn't correct.\nPlease enter your password again.", MessageType.ERROR);
        }
    }

    @FXML
    private void openCertificateFileChooser() {
        FileChooser fc = new FileChooser();

        //set the default open dir
        if (System.getProperty("user.home") != null)
            fc.setInitialDirectory(new File(System.getProperty("user.home")));

        if (jFedPreferences.containsPreference(GuiPreferenceKey.PREF_RSPECFOLDER_REQUEST)) {
            File dir = new File(jFedPreferences.getString(GuiPreferenceKey.PREF_RSPECFOLDER_REQUEST));
            if (dir.exists()) {
                fc.setInitialDirectory(dir);
            }
        }

        File file = fc.showOpenDialog(null);

        if (file == null)
            return;

        loadPemFile(file);

    }

    private void loadPemFile(File file) {
        assert (file != null);

        String fileContent;
        try {
            fileContent = IOUtils.fileToString(file);
        } catch (IOException e) {
            LOG.error("Error while reading certificate content", e);
            fileContent = "";
        }

        try {
            if (fileContent.startsWith("fed4fire.")) {
                userLoginModelManager.setUserLoginModelType(UserLoginModelManager.UserLoginModelType.PLANETLAB);
                PlanetlabUserLoginModel m = userLoginModelManager.getPlanetlabUserLoginModel();
                m.setSshPrivateKeyFile(file);
            } else {
                userLoginModelManager.setUserLoginModelType(UserLoginModelManager.UserLoginModelType.KEY_CERT_INTERNAL_INFO);
                KeyCertUserLoginModel m = userLoginModelManager.getKeyCertUserLoginModel();
                m.setKeyCertPemFile(file);
            }

            modelToGui();
        } catch (InvalidLoginException ex) {
            LOG.error("Error while loading key-certificate-pair from file", ex);
            JFDialogs.create()
                    .owner(fileBrowseButton.getScene().getWindow())
                    .message("The provided login certificate could not be properly processed. " +
                            "Please check if the file format is supported, " +
                            "or download the certificate again from your authority.")
                    .masthead("Invalid login certificate")
                    .title("Invalid login certificate")
                    .showException(ex);

            clear();
        }
    }

    private void loadPemUrl(URL url) {
        assert (url != null);

        //we assume that the url will contain an key-certificate-pair

        userLoginModelManager.setUserLoginModelType(UserLoginModelManager.UserLoginModelType.KEY_CERT_INTERNAL_INFO);
        KeyCertUserLoginModel m = userLoginModelManager.getKeyCertUserLoginModel();
        try {
            m.setKeyCertPemUrl(url);

            modelToGui();
        } catch (InvalidLoginException e) {
            LOG.error("Error while loading key-certificate-pair from URL", e);
            JFDialogs.create()
                    .owner(null)
                    .message("An error occurred when using the provided URL to fetch the login-information")
                    .masthead("Invalid URL")
                    .title("Invalid URL")
                    .showException(e);

            clear();
        }

    }

    /**
     * This function gets all applicable information from the userLoginModel and puts it in the right fields of the GUI
     */
    private void modelToGui() {
        switch (userLoginModelManager.getUserLoginModelType()) {
            case KEY_CERT_INTERNAL_INFO: {
                KeyCertUserLoginModel m = userLoginModelManager.getKeyCertUserLoginModel();
                assert m != null;

                if (m.isPemGiven()) {
                    if (m.isFromFile()) {
                        fileInput.setText(m.getKeyCertFile().getAbsolutePath());
                    } else if (m.isFromURL()) {
                        fileInput.setText(m.getKeyCertURL().toExternalForm());
                    } else {
                        fileInput.setText(CUSTOM_INPUT);
                    }
                    readCertificate();
                } else {
                    usernameLabel.setText("");
                    authorityLabel.setText("");
                    expireLabel.setText("");
                }


                break;
            }
            case PLANETLAB: {
                PlanetlabUserLoginModel m = userLoginModelManager.getPlanetlabUserLoginModel();
                assert m != null;

                if (m.isReadyToFetchCertificate()) {
                    if (m.getPrivateKeyFile() != null && m.getPrivateKeyFile().exists()) {
                        fileInput.setText(m.getPrivateKeyFile().getAbsolutePath());
                    } else if (m.getPrivateKeyUrl() != null) {
                        fileInput.setText(m.getPrivateKeyUrl().toExternalForm());
                    }else {
                        fileInput.setText(CUSTOM_INPUT);
                    }
                    readCertificate();
                } else {
                    usernameLabel.setText("");
                    authorityLabel.setText("");
                    expireLabel.setText("");
                }
                break;
            }
            default:
                throw new RuntimeException("Unsupported type: " + userLoginModelManager.getUserLoginModelType());
        }
    }

    @FXML
    public void showAdvancedLogin() {
        advancedButton.getScene().getWindow().hide();

        advancedUserLoginFactory.showUserLogin(false, advancedButton.getScene().getWindow(), successfulLogin -> {
            if (!successfulLogin) {
                JFDialogs.create()
                        .owner(null)
                        .message("You did not successfully log in. Quiting...")
                        .showError();
            }

            ((Stage) advancedButton.getScene().getWindow()).close();
            fireUserLoginControllerListeners(successfulLogin);
        });
    }

    @FXML
    private void resetJFedAction(ActionEvent actionEvent) {
        JFedCommonDialogs.requestEraseJFedConfiguration(advancedButton.getScene().getWindow());
    }

    public boolean isSuccess() {
        return success;
    }

    private void closeLoginAndProceed() {
        //forget password in GUI
        passwordField.setText("");

        success = true;

        ((Stage) loginButton.getScene().getWindow()).close();
        fireUserLoginControllerListeners(success);
    }


    private void showConnectivityFailed(String shortMessage, String fullMessage) {
        statusIndicator.setStatus(TaskStatusIndicator.Status.FAILED);
        setMessage(shortMessage, MessageType.ERROR);

        JFDialogs.create()
                .owner(loginButton.getScene().getWindow())
                .message(fullMessage)
                .masthead("Could not contact the server")
                .showError();
    }

    @FXML
    private void showConnectivityTesterAction() {
        connectivityTesterGUIProvider.get().showDialog();
    }

    @FXML
    /**
     * This method is called when the user presses enter on the file input
     */
    private void onFileInputAction() {
        if (fileInput.getText().equals(CUSTOM_INPUT))
            return;

        File file = new File(fileInput.getText());
        if (file.exists()) {
            loadPemFile(file);
        } else {
            //check if it is an url
            try {
                URL url = new URL(fileInput.getText());
                loadPemUrl(url);
            } catch (MalformedURLException e) {
                LOG.debug("Given input is not a file, nor an URL");

            }
        }


    }

    private final List<UserLoginControllerListener> loginControllerListenerList = new ArrayList<>();

    public void addListener(UserLoginControllerListener l) {
        loginControllerListenerList.add(l);
    }

    public void removeListener(UserLoginControllerListener l) {
        loginControllerListenerList.remove(l);
    }

    private void fireUserLoginControllerListeners(boolean successfulLogin) {
        for (UserLoginControllerListener l : loginControllerListenerList)
            l.onDone(successfulLogin);
    }

    public static interface UserLoginControllerListener {
        public void onDone(boolean successfulLogin);
    }

    private static enum MessageType {
        INFO("label-info"),
        WARNING("label-warning"),
        ERROR("label-error");
        private final String styleName;

        MessageType(String styleName) {
            this.styleName = styleName;
        }

        private String getStyleName() {
            return styleName;
        }
    }

    public void setupStage(Stage stage) {
        stage.setOnCloseRequest(value -> {
            fireUserLoginControllerListeners(false);
        });
    }
}
