package be.iminds.ilabt.jfed.experimenter_gui.slice.tasks;

import be.iminds.ilabt.jfed.connectivity_tester.ConnectivityTest;
import be.iminds.ilabt.jfed.connectivity_tester.HostAndPortTest;
import be.iminds.ilabt.jfed.experimenter_gui.slice.Experiment;
import be.iminds.ilabt.jfed.experimenter_gui.ui.status.TaskStatusIndicator;
import be.iminds.ilabt.jfed.highlevel.model.Slice;
import be.iminds.ilabt.jfed.highlevel.model.Sliver;
import be.iminds.ilabt.jfed.lowlevel.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.StatusDetails;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnection;
import be.iminds.ilabt.jfed.preferences.GuiPreferenceKey;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.rspec.model.LoginService;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXModelRspec;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecNode;
import be.iminds.ilabt.jfed.ssh_terminal_tool.ssh_key_info.UserSshKeyInfo;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import be.iminds.ilabt.jfed.util.ProxyPreferencesManager;
import javafx.event.Event;
import javafx.event.EventHandler;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.*;
import java.util.stream.Collectors;

/**
 * User: twalcari
 * Date: 2/2/2015
 * Time: 15:54
 */
public class AuthorityConnectivityTestTask extends ExperimentTask implements CustomDetailsView {
    private static final Logger LOG = LoggerFactory.getLogger(AuthorityConnectivityTestTask.class);
    private static final int WAIT_BETWEEN_TESTS = 10; //in seconds
    private static final int MAX_TESTS = 12;

    private final SfaAuthority authority;
    private final Collection<FXRspecNode> nodesToTest;

    private final GeniUserProvider geniUserProvider;
    private final JFedPreferences jFedPreferences;
    private final ProxyPreferencesManager proxyPreferencesManager;

    private int currentTestCount = 0;
    private final Map<FXRspecNode, ConnectivityTest.Status> nodesConnectivity = new HashMap<>();
    private boolean finished = false;

    AuthorityConnectivityTestTask(Experiment experiment, SfaAuthority authority, @Nonnull Collection<FXRspecNode> nodesToTest,
                                         GeniUserProvider geniUserProvider, JFedPreferences jFedPreferences, ProxyPreferencesManager proxyPreferencesManager) {
        super(experiment);
        this.authority = authority;
        this.nodesToTest = nodesToTest;
        this.geniUserProvider = geniUserProvider;
        this.jFedPreferences = jFedPreferences;
        this.proxyPreferencesManager = proxyPreferencesManager;

        updateTitle(String.format("Testing connectivity to nodes from %s.", authority.getName()));
    }

    @Override
    protected TaskStatusIndicator.Status executeTask() {
        assert (nodesToTest != null && !nodesToTest.isEmpty());
        for (Sliver sliver : experiment.getSlice().findSlivers(authority)) {
            assert (sliver.getStatus().getGlobalStatus() == StatusDetails.SliverStatus.READY);
            assert (sliver.getManifestRspec() != null);
        }


        JFedConnection.ProxyInfo proxyInfo = null;
        if (jFedPreferences.getString(GuiPreferenceKey.PREF_SSHPROXY_USE_FOR_SSH).equals(ProxyPreferencesManager.ALWAYS_USE_PROXY)) {
            GeniUser geniUser = geniUserProvider.getLoggedInGeniUser();
            UserSshKeyInfo userSshKeyInfo = new UserSshKeyInfo(geniUser);
            List<SfaAuthority.ProxyInfo> userAuthProxyInfos = geniUser.getUserAuthority().getProxies();
            proxyInfo =
                    proxyPreferencesManager.getSshProxySettings(userAuthProxyInfos, geniUser, userSshKeyInfo);

            if (proxyInfo == null) {
                LOG.warn("Connectivity tests to nodes on authority {} are skipped as SSH-proxy is active but proxy info was not found." +
                        " (this is a jFed bug)", authority.getHrn());


                updateMessage("Skipped because proxy for SSH-connections is active but proxy info was not found. (this is a jFed bug)");
                return TaskStatusIndicator.Status.FAILED;
            }
        }


        LOG.debug("Starting nodes connectivity test for {} nodes on authority {}", nodesToTest.size(), authority.getHrn());

        Iterator<FXRspecNode> it = nodesToTest.iterator();
        while (it.hasNext()) {
            FXRspecNode node = it.next();

            if (node.getLoginServices().isEmpty()) {
                LOG.debug("Removing node {} from test because it doesn't have any loginServices available", node.getUniqueId());
                it.remove();
            }
        }

        int successCount = 0;

        while (currentTestCount < MAX_TESTS) {
            LOG.trace("Starting connectivity test round {} to {}", currentTestCount + 1, authority.getHrn());

            updateMessage(String.format("Could connect to %d out of %d nodes. Testing... ",
                    successCount, nodesToTest.size()));
            for (FXRspecNode node : nodesToTest) {
                if (!nodesConnectivity.containsKey(node)
                        || nodesConnectivity.get(node) != ConnectivityTest.Status.SUCCEEDED) {

                    ConnectivityTest.Status result = testConnectToNode(node, proxyInfo);
                    if (result == ConnectivityTest.Status.SUCCEEDED) {
                        successCount++;

                        updateMessage(String.format("Could connect to %d out of %d nodes. Testing... ",
                                successCount, nodesToTest.size()));

                        LOG.trace("Could successfully connect to node {}@{} in test round {}",
                                node.getClientId(), authority.getHrn(), currentTestCount + 1);
                    } else{

                        LOG.trace("Got result {} when connecting to node {}@{} in test round {}",
                                result,node.getClientId(), authority.getHrn(), currentTestCount + 1);
                    }

                    nodesConnectivity.put(node, result);
                }
            }

            updateProgress(successCount, nodesToTest.size());

            if (successCount == nodesToTest.size()) {
                updateMessage("");
                return TaskStatusIndicator.Status.SUCCESS;
            } else {
                LOG.trace("Could connect to {} of {} nodes in round {}", successCount, nodesToTest.size(), currentTestCount + 1);
                updateMessage("Could connect to " + successCount + " of " + nodesToTest.size() + " requested nodes.");
            }

            //wait a few seconds before continuing test
            currentTestCount++;
            int waitLeft = WAIT_BETWEEN_TESTS;

            while (currentTestCount < MAX_TESTS && waitLeft > 0) {
                updateMessage(String.format("Could connect to %d out of %d nodes. Trying again in %d seconds... ",
                        successCount, nodesToTest.size(), waitLeft));
                try {
                    Thread.sleep(1000);
                } catch (InterruptedException ignored) {
                }
                waitLeft--;
            }
        }

        finished = true;

        return TaskStatusIndicator.Status.FAILED;
    }

    private ConnectivityTest.Status testConnectToNode(FXRspecNode node, JFedConnection.ProxyInfo proxyInfo) {
        LoginService loginService = node.getLoginServices().get(0);

        LOG.trace("Going to try to connect to {} by using loginService {}:{}",
                node.getUniqueId(), loginService.getHostname(), loginService.getPort());

        //try to setup a connection to the node
        HostAndPortTest hostAndPortTest = new HostAndPortTest(loginService.getHostname(),
                Integer.parseInt(loginService.getPort()),
                null, null, proxyInfo);


        ConnectivityTest.ConnectivityTestResult result = hostAndPortTest.call();
        if (result.getStatus() != ConnectivityTest.Status.SUCCEEDED) {
            return result.getStatus();
        }

        return ConnectivityTest.Status.SUCCEEDED;
    }

    public Map<FXRspecNode, ConnectivityTest.Status> getNodesConnectivity() {
        return nodesConnectivity;
    }

    @Nullable
    @Override
    public EventHandler<Event> getCustomDetailsView() {
        if (finished) {
            List<String> failedNodes = nodesConnectivity.entrySet()
                    .stream()
                    .filter(entry -> entry.getValue() == ConnectivityTest.Status.FAILED)
                    .map(entry -> entry.getKey().getClientId())
                    .collect(Collectors.toList());

            if (failedNodes.size() > 0) {
                return event -> {
                    final String message = String.format("jFed detected that node(s) %s cannot be reached. " +
                            "You can take the following steps to resolve the problem:\n\n" +
                            "1. Try enabling the proxy\n" +
                            "2. Check your network settings. Are you behind a firewall? Do you have IPv6-connectivity?\n" +
                            "3. Check the result of the API-calls\n" +
                            "4. Check if there is an error in the requested experiment\n" +
                            "5. Check if the requested resources are available\n" +
                            "6. Contact First Level Support", failedNodes.stream().collect(Collectors.joining(", ")));
                    JFDialogs.create()
                            .owner(null)
                            .message(message)
                            .masthead("Experiment connectivity issues")
                            .title("Experiment connectivity issues").showError();
                };
            } else
                return null;
        } else
            return null;
    }

    /**
     * The connectivity-test can begin if, for each sliver of this authority:
     * - the status of the sliver is READY
     * - the manifest of the sliver is available
     * <p/>
     * If the authority has one or more nodes, it must have at least one loginService available
     * (we assume that if one loginService is available, it will be available for all loginServices in that sliver)
     *
     * @param authority
     * @return
     */
    public static Collection<FXRspecNode> getNodesToTest(@Nonnull Slice slice, @Nonnull SfaAuthority authority) {
        //use a map to make sure that only one rspecnode per id is collected
        Map<String, FXRspecNode> nodesOfAuthorityWithLoginService = new HashMap<>();
        Set<String> nodeIdsWithoutLoginService = new HashSet<>();

        for (Sliver sliver : slice.findSlivers(authority)) {
            //check if sliver is ready
            if (sliver.getStatus().getGlobalStatus() != StatusDetails.SliverStatus.READY) {
                if (LOG.isTraceEnabled())
                    LOG.trace("Cannot start connectivity test to {} as sliver {} is not READY", authority.getHrn(),
                            sliver.getUrn());
                return null;
            }

            //check if manifest is available
            if (sliver.getManifestRspec() == null) {
                if (LOG.isTraceEnabled())
                    LOG.trace("Cannot start connectivity test to {} as manifest is not available for sliver {}",
                            authority.getHrn(), sliver.getUrn());
                return null;
            } else {

                //only add node if it has a loginService, and hasn't been added yet to the map
                FXModelRspec modelRspec = (FXModelRspec) sliver.getManifestRspec().getModelRspec(ModelRspecType.FX);
                assert modelRspec != null;
                Collection<FXRspecNode> nodes = modelRspec.getNodesByAuthority(authority);
                assert nodes != null;
                for (FXRspecNode node : nodes) {
                    if (!node.getLoginServices().isEmpty()) {
                        if (!nodesOfAuthorityWithLoginService.containsKey(node.getUniqueId())) {
                            nodesOfAuthorityWithLoginService.put(node.getUniqueId(), node);
                            nodeIdsWithoutLoginService.remove(node.getUniqueId()); //remove it from the nodeIdsWithoutLoginService
                        }
                    } else {
                        //check if we haven't found a loginService for this node before, and add it to nodeIdsWithoutLoginService otherwise
                        if (!nodesOfAuthorityWithLoginService.containsKey(node.getUniqueId()))
                            nodeIdsWithoutLoginService.add(node.getUniqueId());
                    }
                }
            }
        }

        if (!nodesOfAuthorityWithLoginService.isEmpty()) {
            return nodesOfAuthorityWithLoginService.values();
        } else if (!nodeIdsWithoutLoginService.isEmpty()) {
            if (LOG.isTraceEnabled())
                LOG.trace("Cannot start connectivity test for authority {} as no nodes with loginServices are available",
                        authority.getHrn());
            return null; //we didn't find a loginService
        } else {
            //start the connectivity test
            if (LOG.isTraceEnabled())
                LOG.trace("No connectivity test will be performed for authority {}, as it isn't responsible for any nodes",
                        authority.getHrn());
            return null;   //the authority isn't responsible for a node
        }
    }

    @Override
    public String toString() {
        return "AuthorityConnectivityTestTask{" +
                "authority=" + authority +
                ", nodesToTest=" + nodesToTest +
                '}';
    }
}
