package be.iminds.ilabt.jfed.experimenter_gui.slice.tasks;

import be.iminds.ilabt.jfed.experimenter_gui.slice.Experiment;
import be.iminds.ilabt.jfed.experimenter_gui.slice.ExperimentController;
import be.iminds.ilabt.jfed.experimenter_gui.slice.SliceState;
import be.iminds.ilabt.jfed.experimenter_gui.ui.status.TaskStatusIndicator;
import be.iminds.ilabt.jfed.highlevel.controller.HighLevelController;
import be.iminds.ilabt.jfed.highlevel.controller.Task;
import be.iminds.ilabt.jfed.highlevel.controller.TaskExecution;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.List;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.atomic.AtomicBoolean;

/**
 * User: twalcari
 * Date: 2/23/2015
 * Time: 17:48
 */
public class ReserveSliversTaskGenerator extends ExperimentTaskGenerator {
    private static final Logger LOG = LoggerFactory.getLogger(ReserveSliversTaskGenerator.class);
    private final HighLevelController hlc;
    private final ExperimentTasksFactory experimentTasksFactory;

    public ReserveSliversTaskGenerator(ExperimentController experimentController, HighLevelController hlc, ExperimentTasksFactory experimentTasksFactory) {
        super(experimentController);
        this.hlc = hlc;
        this.experimentTasksFactory = experimentTasksFactory;
    }

    @Override
    protected TaskStatusIndicator.Status generateExperimentTasks() {
        //now we can do the reservation
        experiment.getSlice().setRequestRspec(experiment.getNewRequestRspecSource());

        assert experiment.getSlice() != null;
        assert experiment.getSlice().getExpirationDate() != null;

        if (experiment.getSlice().getExpirationDate().isBefore(experiment.getRequestedEndTime())) {

            LOG.debug("Slice {} currently expires at {}. Renewing until {}", experiment.getName(), experiment.getSlice().getExpirationDate(), experiment.getRequestedEndTime());

            CountDownLatch waitForExpirationRenewal = new CountDownLatch(1);

            final RenewSliceTask renewSliceTask =
                    experimentTasksFactory.createRenewSliceTask(experiment, experiment.getRequestedEndTime());
            renewSliceTask.addTaskFinishedCallback(status -> waitForExpirationRenewal.countDown());

            experimentController.submitExperimentTask(renewSliceTask);
            try {
                waitForExpirationRenewal.await();
            } catch (InterruptedException ignored) {
            }

            if(experiment.getSlice().getExpirationDate().isBefore(experiment.getRequestedEndTime())){
                LOG.debug("Could not update slice end time. Aborting reservation.");
                return TaskStatusIndicator.Status.FAILED;
            }
        }


        List<HighLevelController.CreateSliverReservationTask> createSliverReservationTasks =
                hlc.createSliverReservations(experiment.getSlice(), experiment.getRequestedStartTime(), experiment.getRequestedEndTime());

        LOG.debug("Received {} sliver reservation tasks for {}", createSliverReservationTasks.size(), experiment.getName());

        CountDownLatch reservationsMadeCountDownLatch = new CountDownLatch(createSliverReservationTasks.size());
        AtomicBoolean success = new AtomicBoolean(true);
        for (HighLevelController.CreateSliverReservationTask createSliverReservationTask : createSliverReservationTasks) {

            CreateSliverReservationExperimentTask createSliverReservationExperimentTask =
                    new CreateSliverReservationExperimentTask(experiment, hlc, createSliverReservationTask);

            createSliverReservationExperimentTask.addTaskFinishedCallback(status -> {
                if (!createSliverReservationExperimentTask.isSuccess()) {
                    success.set(false);
                }
                experiment.getFailedReservationResources().addAll(createSliverReservationTask.getFailedResources());
                reservationsMadeCountDownLatch.countDown();
            });

            experimentController.submitExperimentTask(createSliverReservationExperimentTask);
        }

        try {
            reservationsMadeCountDownLatch.await();

            experiment.setReservationMade(true);

            if (success.get()) {
                experiment.setSliceState(SliceState.FUTURE_RESERVATION);
            } else {
                experiment.setSliceState(SliceState.RESERVATION_FAILED);
            }


        } catch (InterruptedException e) {
            LOG.error("Interrupted while making reservations", e);
        }

        return TaskStatusIndicator.Status.SUCCESS;
    }


    private final class CreateSliverReservationExperimentTask extends SingleCallExperimentTask {


        private final HighLevelController.CreateSliverReservationTask createSliverReservationTask;

        public CreateSliverReservationExperimentTask(Experiment experiment, HighLevelController hlc, HighLevelController.CreateSliverReservationTask createSliverReservationTask) {
            super(experiment, hlc);
            this.createSliverReservationTask = createSliverReservationTask;

            updateTitle("Making resource reservation at " + createSliverReservationTask.getAuthority().getName());
        }

        @Override
        public Task createTask() {
            return createSliverReservationTask;
        }

        @Override
        public void onTaskFinished(TaskExecution taskExecution) {
            updateMessage(createSliverReservationTask.getStatus());
        }

        public boolean isSuccess() {
            return createSliverReservationTask.getManifestRspec() != null;
        }
    }
}
