package be.iminds.ilabt.jfed.experimenter_gui.slice.tasks;

import be.iminds.ilabt.jfed.connectivity_tester.ConnectivityTest;
import be.iminds.ilabt.jfed.experimenter_gui.slice.ExperimentController;
import be.iminds.ilabt.jfed.experimenter_gui.slice.SliceState;
import be.iminds.ilabt.jfed.experimenter_gui.ui.status.TaskStatusIndicator;
import be.iminds.ilabt.jfed.highlevel.model.Sliver;
import be.iminds.ilabt.jfed.preferences.GuiPreferenceKey;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecNode;
import be.iminds.ilabt.jfed.ui.javafx.GlyphUtils;
import javafx.application.Platform;
import javafx.scene.control.*;
import javafx.scene.layout.VBox;
import javafx.scene.paint.Color;
import org.controlsfx.glyphfont.FontAwesome;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.concurrent.CountDownLatch;
import java.util.stream.Collectors;

/**
 * User: twalcari
 * Date: 2/3/2015
 * Time: 17:58
 */
public class TestConnectivityTaskGenerator extends ExperimentTaskGenerator {
    private static final Logger LOG = LoggerFactory.getLogger(TestConnectivityTaskGenerator.class);

    private final ExperimentTasksFactory experimentTasksFactory;
    private final JFedPreferences jFedPreferences;

    TestConnectivityTaskGenerator(ExperimentController experimentController, ExperimentTasksFactory experimentTasksFactory, JFedPreferences jFedPreferences) {
        super(experimentController);
        this.experimentTasksFactory = experimentTasksFactory;
        this.jFedPreferences = jFedPreferences;
    }

    @Override
    protected TaskStatusIndicator.Status generateExperimentTasks() {
        LOG.trace("Starting connectivity tests to existing resources in experiment {}", experiment.getName());
        List<AuthorityConnectivityTestTask> authorityConnectivityTestTasks = new ArrayList<>();

        experiment.getSlice().getSlivers().stream()
                .map(Sliver::getAuthority)
                .distinct().forEach(authority -> {

            Collection<FXRspecNode>
                    nodesToTest = AuthorityConnectivityTestTask.getNodesToTest(experiment.getSlice(), authority);
            if (nodesToTest != null) {
                LOG.debug("Testing connectivity to {} nodes for authority {}", nodesToTest.size(), authority.getHrn());
                authorityConnectivityTestTasks.add(experimentTasksFactory.createAuthorityConnectivityTestTask(experiment, authority, nodesToTest));
            } else {
                LOG.info("Skipping connectivity tests to {} as there are no login-services available", authority);
            }
        });

        CountDownLatch authoritiesReadyCountdownLatch = new CountDownLatch(authorityConnectivityTestTasks.size());

        for (AuthorityConnectivityTestTask task : authorityConnectivityTestTasks) {
            task.addTaskFinishedCallback(status -> {
                authoritiesReadyCountdownLatch.countDown();
            });

            experimentController.submitExperimentTask(task);
        }

        try {
            authoritiesReadyCountdownLatch.await();

            //check if we should show a warning about connectivity issues
            final List<String> failedNodeNames =
                    authorityConnectivityTestTasks.stream().flatMap(actt -> actt.getNodesConnectivity().entrySet().stream())
                            .filter(entry -> entry.getValue() == ConnectivityTest.Status.FAILED)
                            .map(entry -> entry.getKey().getClientId())
                            .collect(Collectors.toList());


            Platform.runLater(() -> {
                if (!failedNodeNames.isEmpty()) {
                    //show a (mutable) notification to mac-users that the terminal opens in the background
                    if (jFedPreferences.getBoolean(GuiPreferenceKey.PREF_SHOW_CONNECTIVITY_WARNING_DIALOG, true)) {
                        LOG.debug("Showing connectivity warning dialog");


                        VBox pane = new VBox();
                        pane.setPrefWidth(400);
                        pane.setPrefHeight(250);
                        pane.setSpacing(5);

                        Label textLabel = new Label(String.format("jFed detected that node(s) %s cannot be reached. " +
                                "You can take the following steps to resolve the problem:\n\n" +
                                "1. Try enabling the proxy\n" +
                                "2. Check your network settings. Are you behind a firewall? Do you have IPv6-connectivity?\n" +
                                "3. Check the result of the API-calls\n" +
                                "4. Check if there is an error in the requested experiment\n" +
                                "5. Check if the requested resources are available\n" +
                                "6. Contact First Level Support", failedNodeNames.stream().collect(Collectors.joining(", "))));
                        textLabel.setPrefHeight(180); //work around bug FEDIBBTDEV-1008
                        textLabel.setWrapText(true);
                        pane.getChildren().add(textLabel);
                        final CheckBox checkBox = new CheckBox("Do not show this dialog again");
                        pane.getChildren().add(checkBox);

                        DialogPane dialogPane = new DialogPane();
                        dialogPane.setHeaderText("Experiment connectivity issues");
                        dialogPane.setGraphic(GlyphUtils.createDialogGlyph(FontAwesome.Glyph.PLAY, Color.GREEN));
                        dialogPane.setContent(pane);
                        dialogPane.getButtonTypes().setAll(ButtonType.OK);


                        Dialog<ButtonType> dialog = new Dialog<>();
                        dialog.setTitle("Experiment connectivity issues");
                        dialog.setDialogPane(dialogPane);

                        dialog.showAndWait();

                        jFedPreferences.setBoolean(GuiPreferenceKey.PREF_SHOW_CONNECTIVITY_WARNING_DIALOG,
                                !checkBox.isSelected());
                    } else {
                        LOG.debug("Skipping connectivity issues dialog because user muted it before");
                    }
                }
            });


            if (experiment.getSliceState() == SliceState.TESTING) {
                experiment.setSliceState(SliceState.READY);
            } else {
                LOG.warn("Cannot transfer to SliceState READY as current state is {} instead of TESTING", experiment.getSliceState());
            }
        } catch (InterruptedException ignored) {
        }


        return TaskStatusIndicator.Status.SUCCESS;
    }
}
