package be.iminds.ilabt.jfed.highlevel.model;

import be.iminds.ilabt.jfed.log.ApiCallDetails;
import be.iminds.ilabt.jfed.lowlevel.api.AggregateManager3;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.ErrorDetails;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.StatusDetails;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl.AMv3Wrapper;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.rspec_source.ManifestRspecSource;
import be.iminds.ilabt.jfed.util.GeniUrn;
import javafx.application.Platform;
import org.slf4j.LoggerFactory;

import java.text.ParseException;
import java.util.ArrayList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * EasyModelSliceAuthorityListener: this watches for SliceAuthority calls and fills in EasyModel using the info in them
 */
public class EasyModelAggregateManager3Listener extends EasyModelAbstractListener {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(EasyModelAggregateManager3Listener.class);

    private final AuthorityList authorityList;
    public EasyModelAggregateManager3Listener(EasyModel easyModel, AuthorityList authorityList) {
        super(easyModel);
        this.authorityList = authorityList;
    }

    private List<String> getSliceUrns(List<String> urns) {
        List<String> res = new ArrayList<>();
        for (String u : urns) {
            GeniUrn geniUrn = GeniUrn.parse(u);
            if (geniUrn != null && geniUrn.getEncodedResourceType().equals("slice"))
                res.add(u);
        }
        return res;
    }

    private List<String> getSliverUrns(List<String> urn) {
        List<String> res = new ArrayList<>();
        for (String u : urn) {
            GeniUrn geniUrn = GeniUrn.parse(u);
            if (geniUrn != null && geniUrn.getEncodedResourceType().equals("sliver"))
                res.add(u);
        }
        return res;
    }

    private List<String> getSliverUrnsFromSliverInfo(List<AggregateManager3.SliverInfo> sliverinfos) {
        List<String> res = new ArrayList<>();
        for (AggregateManager3.SliverInfo si : sliverinfos)
            res.add(si.getSliverUrn());
        return res;
    }


    private void onGetVersionResult(ApiCallDetails result) {
        if (result.getReply().getGeniResponseCode().isSuccess()) {
            //TODO
        }
    }

    public void onStatusResult(AggregateManager3.StatusInfo statusInfo) {
        getModel().logExistSlice(statusInfo.getSliceUrn());

        onListOfSliverInfoResult(statusInfo.getSliceUrn(), statusInfo.getSliverInfo(), false);
    }

    /**
     * @param sliceUrn              urn of the slice the sliver belongs to, may be null if unknown
     * @param ignoreOperationStatus
     */
    public void onListOfSliverInfoResult(String sliceUrn, List<AggregateManager3.SliverInfo> sliverinfos, boolean ignoreOperationStatus) {
        assert sliceUrn != null;
        if (sliverinfos == null) return;

        for (AggregateManager3.SliverInfo si : sliverinfos) {
            LOG.debug("");
            if (si.getAllocationStatus().equals("geni_unallocated")) {
                //sliver is deleted
                getModel().logNotExistSliver(sliceUrn, si.getSliverUrn());
            } else {
                Sliver sliver =
                        getModel().logExistSliver(sliceUrn, si.getSliverUrn());

                StatusDetails statusDetails = sliverInfoToSliverStatus(si);

                if (!ignoreOperationStatus)
                    sliver.setAllocationAndOperationalStatus(si.getAllocationStatus(), si.getOperationalStatus(), statusDetails);
                try {
                    sliver.setExpirationDate(si.getExpiresDate().toInstant());
                } catch (ParseException e) {
                    LOG.warn("in EasyModelAggregateManager3Listener in onListOfSliverInfoResult(): Invalid RFC3339 date in sliver expires: " + si.getExpires(), e);
                }
            }
        }
    }

    public static StatusDetails sliverInfoToSliverStatus(AggregateManager3.SliverInfo si) {
        StatusDetails.SliverStatus s = AMv3Wrapper.getGlobalSliverStatus(si);
        StatusDetails statusDetails = new StatusDetails(s);
        statusDetails.addSliverStatus(si.getSliverUrn(), s);
        if (si.getError() != null && !si.getError().isEmpty())
            statusDetails.addError(new ErrorDetails(si.getError(), si.getSliverUrn()));
        return statusDetails;
    }

    public void onAllocateAndProvisionInfoResult(String sliceUrn, AggregateManager3.AllocateAndProvisionInfo info, boolean ignoreOperationalStatus) {
        onListOfSliverInfoResult(sliceUrn, info.getSliverInfo(), ignoreOperationalStatus);
    }

    public void onManifestInfo(SfaAuthority auth, AggregateManager3.ManifestInfo manifestInfo) {
        assert manifestInfo != null;
        assert manifestInfo.getSliceUrn() != null;
        assert manifestInfo.getSliverInfos() != null;
        assert manifestInfo.getManifestRspec() != null;
        assert !manifestInfo.getSliverInfos().isEmpty();

        List<String> sliverUrns = getSliverUrnsFromSliverInfo(manifestInfo.getSliverInfos());
        seeManifestRspec(auth, manifestInfo.getSliceUrn(), sliverUrns, manifestInfo.getManifestRspec());

        onListOfSliverInfoResult(manifestInfo.getSliceUrn(), manifestInfo.getSliverInfos(), false);
    }

    public void seeAdvertisementRspec(SfaAuthority auth, boolean available, String rspec) {
        AuthorityInfo ai = authorityList.get(auth);
        ai.setAdvertisementRspec(available, rspec);
    }

//    public void seeRequestRspec(SfaAuthority auth, String sliceUrn, List<String> sliverUrns, String rspec) {
//        assert sliverUrns != null;
//        assert getModel().getRSpecList() != null;
//        getModel().getRSpecList().seeRequestRspec(auth, sliceUrn, sliverUrns, rspec);
//    }
//
//    public void seeEchoRequestRspec(SfaAuthority auth, String sliceUrn, List<String> sliverUrns, String rspec) {
//        assert sliverUrns != null;
//        assert getModel().getRSpecList() != null;
//        getModel().getRSpecList().seeEchoRequestRspec(auth, sliceUrn, sliverUrns, rspec);
//    }

    public void seeManifestRspec(SfaAuthority auth, String sliceUrn, List<String> sliverUrns, String manifestRspecString) {
        assert auth != null;
        assert sliceUrn != null;
        assert sliverUrns != null;

        //when empty rspec, there will be no slivers.
        if (sliverUrns.isEmpty()) {
            getModel().logNotExistSliverInSliceOnAuth(sliceUrn, auth);
            return;
        }

        assert manifestRspecString != null;
        assert !sliverUrns.isEmpty();

        Slice slice = getModel().getSlice(sliceUrn);
        if (slice == null) {
            LOG.warn("Could not find slice with URN {}, while seeing a Manifest from authority {}", sliceUrn,
                    auth.getUrnString());
            return;
        }

        //This will find the same list, if sliver_id is set in manifest.
        //This list is not used, but it is still done, here, for the side effect: it maps componet_id and client_id to sliver_id
        List<Sliver> sliverListManifest = findSliversInManifest(auth, sliceUrn, new ManifestRspecSource(manifestRspecString, ModelRspecType.FX));

        List<Sliver> sliverList = new ArrayList<>();
        for (String sliverUrn : sliverUrns) {
            Sliver sliver =
                    getModel().logExistSliver(sliceUrn, sliverUrn, auth); //since success is returned, the sliver(s) must exist
            sliverList.add(sliver);
        }

        //////// HACK for PLE ////////
        //hack for ple: handle the fact that ple requires a component_manager_id of
        //   urn:publicid:IDN+ple:ibbtple+authority+cm in the request RSpec but it returns
        //   urn:publicid:IDN+ple+authority+cm
        if (manifestRspecString.contains("component_manager_id=\"urn:publicid:IDN+ple+authority+cm\"")) {
            manifestRspecString = manifestRspecString.replaceAll(
                    Pattern.quote("urn:publicid:IDN+ple+authority+cm"),
                    Matcher.quoteReplacement("urn:publicid:IDN+ple:ibbtple+authority+cm"));
            LOG.warn("Using PLE hack in EasyModelAggregateManager2Listener#seeManifestRspec. succesfully applied: " +
                    (!manifestRspecString.contains("component_manager_id=\"urn:publicid:IDN+ple+authority+cm\"")));
            assert !manifestRspecString.contains("component_manager_id=\"urn:publicid:IDN+ple+authority+cm\"") :
                    "HACK applied unsuccessfully to:\n" + manifestRspecString;
        }
        //////// HACK for PLE ////////

        ManifestRspecSource manifestRspec = new ManifestRspecSource(manifestRspecString, ModelRspecType.FX);
        for (Sliver sliver : sliverList) {
            LOG.debug("EasyModelAggregateManager2Listener seeManifestRspec is setting manifest for " + slice + " -> " + sliceUrn + "  " + sliver + " -> " + sliver.getUrn());
            sliver.setManifestRspec(manifestRspec);
        }

//        //sliver exists on all cm's mentioned in manifest. (even if unallocated on others)
//        try {
//            for (String cmUrn : rSpecInfo.getRSpec().getAllComponentManagerUrns()) {
//                SfaAuthority rspecAuth = getModel().getAuthorityList().getAuthorityListModel().getByUrn(cmUrn);
//                if (rspecAuth == null) {
//                    LOG.error("Error in manifest: contained an unknown authority urn: \"" + cmUrn + "\"");
//                    continue;
//                }
//                getModel().logExistSliverGeniSingle(sliceUrn, rspecAuth); //it is in manifest, so the sliver must exist (even if it is unallocated)
//            }
//        } catch (RspecParseException ex) {
//            LOG.error("Error while processing Rspec XML", ex);
//        }

    }

    @Override
    public void onResult(final ApiCallDetails details) {
//        //Since this triggers updates to the GUI (because it changes Properties) this has to be executed on the JavaFX thread.
//
//        final Lock lock = new ReentrantLock();
//        final Condition waiter = lock.newCondition();
//
//        Platform.runLater(new Runnable() { @Override public void run() {
//            onResultInJavaFXThread(details);
//            lock.lock();
//            waiter.signalAll();
//            lock.unlock();
//        } });
//
//        lock.lock();
//        try {
//            waiter.await();
//        } catch (InterruptedException e) {
//            //TODO: handle?
//            e.printStackTrace();
//        }
//        lock.unlock();
        assert Platform.isFxApplicationThread();
        onResultInJavaFXThread(details);
    }

    public void onResultInJavaFXThread(ApiCallDetails details) {
        //ignore errors here
        if (details.getReply() == null || details.getJavaMethodName() == null)
            return;

        if (!details.getApiName().equals(AggregateManager3.getApiName()))
            return;

        LOG.debug("EasyModelAggregateManager3Listener onResultInJavaFXThread API=" + details.getApiName() + " SfaCommand=\"" + details.getGeniMethodName() + "\" javaCommand=\"" + details.getJavaMethodName() + "\"");

        try {
            if (details.getJavaMethodName().equals("getVersion")) //returns VersionInfo
                onGetVersionResult(details);

            if (details.getJavaMethodName().equals("listResources")) { //returns String (rspec)
                if (details.getReply().getGeniResponseCode().isSuccess()) {
                    String rspec = (String) details.getReply().getValue();
                    boolean available = false;
                    if (details.getMethodParameters().get("available") != null)
                        available = (Boolean) details.getMethodParameters().get("available");
                    seeAdvertisementRspec(details.getAuthority(), available, rspec);
                }
            }

            if (details.getJavaMethodName().equals("describe")) { //returns ManifestInfo
                onManifestInfo(details.getAuthority(), (AggregateManager3.ManifestInfo) details.getReply().getValue());
            }

            if (details.getJavaMethodName().equals("allocate")) { //returns AllocateAndProvisionInfo
                if (details.getReply().getGeniResponseCode().isSuccess()) {
                    AggregateManager3.AllocateAndProvisionInfo info =
                            (AggregateManager3.AllocateAndProvisionInfo) details.getReply().getValue();
                    String sliceUrn = (String) details.getMethodParameters().get("sliceUrn");
                    getModel().logExistSlice(sliceUrn);
                    List<String> sliverUrns = getSliverUrnsFromSliverInfo(info.getSliverInfo());
//                    seeRequestRspec(details.getAuthority(), sliceUrn, sliverUrns, (String) details.getMethodParameters().get("rspec"));
//                    seeEchoRequestRspec(details.getAuthority(), sliceUrn, sliverUrns, info.getRspec());
                    onAllocateAndProvisionInfoResult(sliceUrn, info, false);
                }
            }

            if (details.getJavaMethodName().equals("provision")) { //returns AllocateAndProvisionInfo
                if (details.getReply().getGeniResponseCode().isSuccess()) {
                    AggregateManager3.AllocateAndProvisionInfo info =
                            (AggregateManager3.AllocateAndProvisionInfo) details.getReply().getValue();
                    List<String> urns = (List<String>) details.getMethodParameters().get("urns");
                    List<String> sliceUrns = getSliceUrns(urns);
                    String sliceUrn = sliceUrns.size() != 1 ? null : sliceUrns.get(0);
                    if (sliceUrn != null)
                        getModel().logExistSlice(sliceUrn);

                    //  alternative, get from result instead of arguments:
                    List<String> sliverUrns = getSliverUrnsFromSliverInfo(info.getSliverInfo());
                    //  alternative: get from arguments instead of result:
                    //List<String> sliverUrns = getSliverUrns(urns);

                    seeManifestRspec(details.getAuthority(), sliceUrn, sliverUrns, info.getRspec());
                    boolean ignoreOperationalStatus =
                            details.getAuthority().hasSpecialCase(SfaAuthority.SpecialCase.IGNORE_PROVISION_OPERATIONAL_STATUS);
                    onAllocateAndProvisionInfoResult(sliceUrn, info, ignoreOperationalStatus);
                }
            }

            if (details.getJavaMethodName().equals("status")) // returns StatusInfo
            {
                if (details.getReply().getGeniResponseCode().isSuccess()) {
                    //not much to know from urn argument?   List urns = (List) details.getMethodParameters().get("urns");
                    //TODO if single slice URN, and success and at least one sliverinfo, that slice exists...
                    onStatusResult((AggregateManager3.StatusInfo) details.getReply().getValue());
                    onStatusResult((AggregateManager3.StatusInfo) details.getReply().getValue());
                }
            }

            if (details.getJavaMethodName().equals("performOperationalAction")) { //returns List<SliverInfo>
                if (details.getReply().getGeniResponseCode().isSuccess()) {
                    List<AggregateManager3.SliverInfo> si =
                            (List<AggregateManager3.SliverInfo>) details.getReply().getValue();
                    List<String> urns = (List<String>) details.getMethodParameters().get("urns");
                    List<String> sliceUrns = getSliceUrns(urns);
                    String sliceUrn = sliceUrns.size() != 1 ? null : sliceUrns.get(0);
                    onListOfSliverInfoResult(sliceUrn, si, false);
                }
            }

            if (details.getJavaMethodName().equals("delete")) { //returns List<SliverInfo>
                if (details.getReply().getGeniResponseCode().isSuccess()) {
                    List<AggregateManager3.SliverInfo> si =
                            (List<AggregateManager3.SliverInfo>) details.getReply().getValue();
                    List<String> urns = (List<String>) details.getMethodParameters().get("urns");
                    List<String> sliceUrns = getSliceUrns(urns);
                    String sliceUrn = sliceUrns.size() != 1 ? null : sliceUrns.get(0);
                    onListOfSliverInfoResult(sliceUrn, si, false);
                }
            }

            if (details.getJavaMethodName().equals("renew")) { //returns List<SliverInfo>
                if (details.getReply().getGeniResponseCode().isSuccess()) {
                    List<AggregateManager3.SliverInfo> si =
                            (List<AggregateManager3.SliverInfo>) details.getReply().getValue();
                    List<String> urns = (List<String>) details.getMethodParameters().get("urns");
                    List<String> sliceUrns = getSliceUrns(urns);
                    String sliceUrn = sliceUrns.size() != 1 ? null : sliceUrns.get(0);
                    onListOfSliverInfoResult(sliceUrn, si, false);
                }
            }

            if (details.getJavaMethodName().equals("shutdown")) { //returns Boolean
                //ignored
            }
        } catch (Exception e) {
            LOG.warn("WARNING: Exception when processing AggregateManager3 reply for EasyModel. This will be ignored, but it is most likely a bug. " + e.getMessage(), e);
        }
    }
}
