package be.iminds.ilabt.jfed.highlevel.model;

import be.iminds.ilabt.jfed.log.ApiCallDetails;
import be.iminds.ilabt.jfed.lowlevel.api.OCCI;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.StatusDetails;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl.OcciToAggregateManagerWrapper;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.rspec_source.ManifestRspecSource;
import be.iminds.ilabt.jfed.util.GeniUrn;
import javafx.application.Platform;
import javafx.util.Pair;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * EasyModelSliceAuthorityListener: this watches for OCCI calls and fills in EasyModel using the info in them
 * <p/>
 * WARNING: this assumes a special mapping between OCCI concepts and SFA concepts, see also OcciToAggregateManagerWrapper
 *
 * @see be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl.OcciToAggregateManagerWrapper
 */
public class EasyModelOCCIListener extends EasyModelAbstractListener {
    public static final String bonfireUrnPart = OcciToAggregateManagerWrapper.bonfireUrnPart;
    public static final GeniUrn bonfireUrn = GeniUrn.createGeniUrnFromEncodedParts(bonfireUrnPart, "authority", "cm");

    private final static org.slf4j.Logger LOG = LoggerFactory.getLogger(EasyModelOCCIListener.class);

    private static boolean debug = false;
    //these 2 mappings are kept in sync by receiving calls
    private Map<String, List<OcciToAggregateManagerWrapper.ComputeInfo>> experimentIdToComputeInfoList =
            new HashMap<>();
    private Map<Pair<String, String>, OcciToAggregateManagerWrapper.ComputeInfo> locationAndComputeIdToComputeInfo =
            new HashMap<>();
    private Map<Pair<String, String>, GeniUrn> computeInfoToSliceUrn = new HashMap<>();

//    public void seeRequestRspec(SfaAuthority auth, String sliceUrn, List<String> sliverUrns, String rspec) {
//        model.getRSpecList().seeRequestRspec(auth, sliceUrn, sliverUrns, rspec);
//    }

    private final AuthorityList authorityList;
    private final EasyOcciModel easyOcciModel;

    public EasyModelOCCIListener(EasyModel easyModel, AuthorityList authorityList, EasyOcciModel easyOcciModel) {
        super(easyModel);
        this.authorityList = authorityList;
        this.easyOcciModel = easyOcciModel;
    }

    /*
    *  NOTE: onExperimentComputes and onLocationCompute are needed to process status information
    *
    *  A big hack is used here to process this date.
    *  It relies some assumptions:
    *   - one single EasyModelOCCIListener exists and sees all calls.
    *   - first onExperimentComputes is called, and then onLocationCompute for each compute in it.
    * */

    private EasyOcciModel getEasyOcciModel() {
        return easyOcciModel;
    }

    public void seeAdvertisementRspec(SfaAuthority auth, boolean available, String rspec) {
        LOG.trace("EasyModelOCCIListener.seeAdvertisementRspec(" + auth.getName() + ", " + available + ", rspec.length=" + rspec.length() + ")");
        AuthorityInfo ai = authorityList.get(auth);
        ai.setAdvertisementRspec(available, rspec);
    }

    private void onLocationsResult(ApiCallDetails result) {
        if (result.getReply().getGeniResponseCode().isSuccess()) {
            List<OcciToAggregateManagerWrapper.LocationInfo> locationInfos =
                    OcciToAggregateManagerWrapper.parseLocationsXml(result.getHttpResponse());
            LOG.debug("EasyModelOCCIListener.onLocationsResult: Received " + locationInfos.size() + " locations from server: " + locationInfos);
            final String adRspec = OcciToAggregateManagerWrapper.fakeAdvertisementRspec(locationInfos);
            LOG.debug("EasyModelOCCIListener.onLocationsResult: Created fake advertisement Rspec: " + adRspec);

            boolean available = true; /* experiment GUI looks only at available */
            seeAdvertisementRspec(result.getAuthority(), available, adRspec);

            for (OcciToAggregateManagerWrapper.LocationInfo li : locationInfos)
                getEasyOcciModel().noteLocationExists(li.getName(), li.getHref());
        }
    }

    private void onExperimentComputes(ApiCallDetails result) {
        if (result.getReply().getGeniResponseCode().isSuccess()) {
            String experimentId = (String) result.getMethodParameters().get("experimentId");
            assert experimentId != null; //must be in parameters of getExperimentComputes
            GeniUrn sliceUrn = OcciToAggregateManagerWrapper.mapExperimentIdToSliceUrn(experimentId);
            assert sliceUrn != null;
            List<OcciToAggregateManagerWrapper.ComputeInfo> infoRecv =
                    OcciToAggregateManagerWrapper.parseExperimentComputes(experimentId, result.getHttpResponse());
            //infoRecv contains only locationName, computeId and experimentId, no other fields.
            for (OcciToAggregateManagerWrapper.ComputeInfo computeInfoRecv : infoRecv)
                if (computeInfoRecv.experimentId != null && computeInfoRecv.locationName != null && computeInfoRecv.computeId != null) {
                    computeInfoToSliceUrn.put(new Pair<>(computeInfoRecv.locationName, computeInfoRecv.computeId), sliceUrn);

                    List<OcciToAggregateManagerWrapper.ComputeInfo> storedExpInfoList =
                            experimentIdToComputeInfoList.get(computeInfoRecv.experimentId);
                    if (storedExpInfoList == null) {
                        storedExpInfoList = new ArrayList<>();
                        experimentIdToComputeInfoList.put(computeInfoRecv.experimentId, storedExpInfoList);
                    }
                    //warning, this is an inefficient and slow algorithm! But this list should not get long, so we ignore it for now
                    //TODO: make this more efficient! (now quadratic complexity!)
                    boolean hasThisCompute = false;
                    for (OcciToAggregateManagerWrapper.ComputeInfo storeExpInf : storedExpInfoList) {
                        if (storeExpInf.computeId.equals(computeInfoRecv.computeId) && storeExpInf.locationName.equals(computeInfoRecv.locationName)) {
                            hasThisCompute = true;
                            computeInfoRecv =
                                    storeExpInf; //use stored version from now. (used if not yet stored in locationAndComputeIdToComputeInfo below)
                            break;
                        }
                    }
                    if (!hasThisCompute)
                        storedExpInfoList.add(computeInfoRecv);

                    Pair<String, String> key = new Pair<>(computeInfoRecv.locationName, computeInfoRecv.computeId);
                    OcciToAggregateManagerWrapper.ComputeInfo storedComputeInfo2 =
                            locationAndComputeIdToComputeInfo.get(key);
                    if (storedComputeInfo2 == null) {
                        locationAndComputeIdToComputeInfo.put(key, computeInfoRecv);
                        assert !hasThisCompute :
                                "should always be added to both lists at once: was in experimentIdToComputeInfoList but not in locationAndComputeIdToComputeInfo";
                    } else {
                        assert hasThisCompute :
                                "should always be added to both lists at once: was not in experimentIdToComputeInfoList but was in locationAndComputeIdToComputeInfo";
                    }
                } else {
                    LOG.warn("Parsed info received from /experiments/<ID>/computes missed some field!" +
                            " experimentId=" + computeInfoRecv.experimentId +
                            " locationName=" + computeInfoRecv.locationName +
                            " computeId=" + computeInfoRecv.computeId);
                }
            refreshStatus(sliceUrn, result.getAuthority());
        }
    }

    private void onPostExperiment(ApiCallDetails result) {
        if (result.getReply().getGeniResponseCode().isSuccess()) {
            OcciToAggregateManagerWrapper.ExperimentInfo expInf =
                    OcciToAggregateManagerWrapper.parseExperiment(result.getAuthority(), null/*sliceUrn*/, result.getHttpResponse());
            assert expInf.experimentId != null;
            assert expInf.experimentName != null;
            assert expInf.sliverUrn != null;

            assert expInf.sliceUrn == null;

            //can't get sliceUrn yet, might not be registered yet: is done after postExperiment call is done
//            expInf.sliceUrn = OcciToAggregateManagerWrapper.mapExperimentIdToSliceUrn(expInf.experimentId);
//            assert expInf.sliceUrn != null;
//            assert expInf.sliverUrn.equals(OcciToAggregateManagerWrapper.mapSliceUrnToSliverUrn(expInf.sliceUrn));

            //logExistSliverGeniSingle also logs slice existence
//            getModel().logExistSliverGeniSingle(expInf.sliceUrn, expInf.sliverUrn, result.getAuthority());
        }
    }

    private void onPostExperimentCompute(ApiCallDetails result) {
        if (result.getReply().getGeniResponseCode().isSuccess()) {
            OcciToAggregateManagerWrapper.ComputeInfo expInf =
                    OcciToAggregateManagerWrapper.parseCompute(result.getHttpResponse());
            //TODO: do something with this?
        }
    }

    private void onLocationNetworks(ApiCallDetails result) {
        if (result.getReply().getGeniResponseCode().isSuccess()) {
            List<OcciToAggregateManagerWrapper.NetworkInfo> networkInfos =
                    OcciToAggregateManagerWrapper.parseNetworkCollection(result.getHttpResponse());

            String locationId = (String) result.getMethodParameters().get("locationName");
            String locationHref = "/locations/" + locationId + "/networks";

            Pattern patternHref = Pattern.compile("/locations/([a-zA-Z0-9-]*)/networks/(.*)");
            for (OcciToAggregateManagerWrapper.NetworkInfo ni : networkInfos) {
                Matcher matcherHref = patternHref.matcher(ni.getHref());
                String networkId;
                if (matcherHref.find())
                    networkId = matcherHref.group(2);
                else
                    networkId = null;
                getEasyOcciModel().noteNetworkExists(locationId, locationHref, ni.getName(), networkId, ni.getHref());
            }
        }
    }

    private void onLocationStorage(ApiCallDetails result) {
        if (result.getReply().getGeniResponseCode().isSuccess()) {
            List<OcciToAggregateManagerWrapper.StorageInfo> storageInfos =
                    OcciToAggregateManagerWrapper.parseStorageCollection(result.getHttpResponse());

            String locationId = (String) result.getMethodParameters().get("locationName");
            String locationHref = "/locations/" + locationId + "/storages";

            Pattern patternHref = Pattern.compile("/locations/([a-zA-Z0-9-]*)/storages/(.*)");
            for (OcciToAggregateManagerWrapper.StorageInfo si : storageInfos) {
                Matcher matcherHref = patternHref.matcher(si.getHref());
                String storageId;
                if (matcherHref.find())
                    storageId = matcherHref.group(2);
                else
                    storageId = null;
                getEasyOcciModel().noteStorageExists(locationId, locationHref, si.getName(), storageId, si.getHref());
            }
        }
    }

    private void onLocationCompute(ApiCallDetails result) {
        if (result.getReply().getGeniResponseCode().isSuccess()) {
            OcciToAggregateManagerWrapper.ComputeInfo infoRecv =
                    OcciToAggregateManagerWrapper.parseCompute(result.getHttpResponse());
            assert infoRecv.locationName != null;
            assert infoRecv.computeId != null;
            OcciToAggregateManagerWrapper.ComputeInfo infoToUpdate =
                    locationAndComputeIdToComputeInfo.get(new Pair<>(infoRecv.locationName, infoRecv.computeId));
            infoToUpdate.state = infoRecv.state;

            refreshStatus(
                    computeInfoToSliceUrn.get(new Pair<>(infoRecv.locationName, infoRecv.computeId)),
                    result.getAuthority());
        }
    }

    private void refreshStatus(GeniUrn sliceUrn, SfaAuthority auth) {
        boolean allReady = true;


        for (String experimentId : experimentIdToComputeInfoList.keySet()) {
            GeniUrn sliverUrn = OcciToAggregateManagerWrapper.mapSliceUrnToSliverUrn(sliceUrn);

            for (OcciToAggregateManagerWrapper.ComputeInfo computeInfo : experimentIdToComputeInfoList.get(experimentId)) {
                LOG.debug("EasyModelOCCIListener.refreshStatus computeId=" + computeInfo.computeId + " state=" + computeInfo.state);
                if (computeInfo.state != null) {
                    LOG.debug("                                    ready=" + computeInfo.ready());
                    if (!computeInfo.ready()) allReady = false;
                } else
                    allReady = false;

            }
            if (experimentIdToComputeInfoList.isEmpty()) return;

            LOG.debug("EasyModelOCCIListener.refreshStatus allReady=" + allReady);

            //TODO register state in already
            Slice slice = getModel().logExistSlice(sliceUrn);
            assert slice != null;

            Sliver sliver = getModel().logExistSliver(sliceUrn.getValue(), sliverUrn.getValue(), auth);
            assert sliver != null;

            sliver.setStatusString(allReady ? "ready" : "changing");
            sliver.setStatus(
                    new StatusDetails(allReady ?
                            StatusDetails.SliverStatus.READY :
                            StatusDetails.SliverStatus.CHANGING, sliver.getUrn())
            );

            LOG.debug("EasyModelOCCIListener.refreshStatus sliver=" + sliver.getUrn() + " status=" + sliver.getStatus());

            //to log it doesn't exist: getModel().logNotExistSliverGeniSingle(sliceUrn, details.getAuthority());
        }
    }


    /**
     * process the result, in the JavaFX thread. And wait for it in this thread.
     */
    @Override
    public void onResult(final ApiCallDetails details) {
        assert Platform.isFxApplicationThread();
        onResultInJavaFXThread(details);
    }

    public void onResultInJavaFXThread(ApiCallDetails details) {
        if (LOG.isDebugEnabled())
            LOG.trace("onResultInJavaFXThread SfaCommand=\"" + details.getGeniMethodName() + "\" javaCommand=\"" + details.getJavaMethodName() + "\"");

        if (!details.getApiName().equals(OCCI.getApiName()))
            return;

        if (details.getJavaMethodName().equals("fakeManifest")) {
            //HACK
            SfaAuthority auth = details.getAuthority();
            GeniUrn sliceUrn = (GeniUrn) details.getMethodParameters().get("sliceUrn");
            GeniUrn sliverUrn = (GeniUrn) details.getMethodParameters().get("sliverUrn");
            String manifest = (String) details.getMethodParameters().get("fakeManifest");
            Slice slice = getModel().logExistSlice(sliceUrn);
            Sliver sliver = getModel().logExistSliver(sliceUrn.getValue(), sliverUrn.getValue(), auth);
            List<Sliver> slivers = new ArrayList<>();
            slivers.add(sliver);
            AuthorityInfo authInf = authorityList.get(auth);
            LOG.debug("EasyModelOCCIListener fakeManifest HACK sets rspec manifest on " + auth.getUrn() + " sliceUrn=" + sliceUrn + " sliverUrn=" + sliverUrn + " manifest.size()=" + manifest.length());
            sliver.setManifestRspec(new ManifestRspecSource(manifest, ModelRspecType.FX));//new RSpecInfo(manifest, RSpecInfo.RspecType.MANIFEST, slice, slivers, authInf));
            return;
        }

        //ignore errors here
        if (details.getReply() == null || details.getJavaMethodName() == null)
            return;

        try {
            if (details.getJavaMethodName().equals("postExperimentWizard"))
                onPostExperiment(details);
            if (details.getJavaMethodName().equals("postExperiment"))
                onPostExperiment(details);
            if (details.getJavaMethodName().equals("postExperimentCompute"))
                onPostExperimentCompute(details);
            if (details.getJavaMethodName().equals("getLocations"))
                onLocationsResult(details);
            if (details.getJavaMethodName().equals("getExperimentComputes"))
                onExperimentComputes(details);
            if (details.getJavaMethodName().equals("getLocationCompute"))
                onLocationCompute(details);

            if (details.getJavaMethodName().equals("getLocationNetworks"))
                onLocationNetworks(details);
            if (details.getJavaMethodName().equals("getLocationStorages"))
                onLocationStorage(details);
        } catch (Exception e) {
            LOG.warn("WARNING: Exception when processing AggregateManager2 reply for EasyModel. This will be ignored, but it is most likely a bug. ", e);

        }
    }
}
