package be.iminds.ilabt.jfed.highlevel.model;

import be.iminds.ilabt.jfed.log.ApiCallDetails;
import be.iminds.ilabt.jfed.lowlevel.AnyCredential;
import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.lowlevel.SfaCredential;
import be.iminds.ilabt.jfed.lowlevel.api.*;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityListModel;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.lowlevel.resourceid.ResourceId;
import be.iminds.ilabt.jfed.util.GeniUrn;
import be.iminds.ilabt.jfed.util.RFC3339Util;
import javafx.application.Platform;
import org.slf4j.LoggerFactory;

import java.text.ParseException;
import java.util.ArrayList;
import java.util.Date;
import java.util.Hashtable;
import java.util.List;

/**
 * EasyModelSliceAuthorityListener: this watches for SliceAuthority calls and fills in EasyModel using the info in them
 * <p/>
 * listens to both federation API 1 and 2!
 */
public class EasyUniformFederationSliceAuthorityApiListener extends EasyModelAbstractListener {
    private static final org.slf4j.Logger LOG =
            LoggerFactory.getLogger(EasyUniformFederationSliceAuthorityApiListener.class);

    private final AuthorityList authorityList;

    public EasyUniformFederationSliceAuthorityApiListener(EasyModel easyModel, AuthorityList authorityList) {
        super(easyModel);
        this.authorityList = authorityList;
    }

    private void onGetSliceCredentials(ApiCallDetails result) {
        LOG.debug("EasyUniformFederationSliceAuthorityApiListener#onGetSliceCredentials() parameters.size(): " + result.getMethodParameters().size());

        assert result.getReply() != null;

        ResourceId sliceId = noteSliceUrnInParameters(result);
        assert sliceId != null;
        noteUserCredentialInParameters(result);

        if (result.getReply().getGeniResponseCode().isSuccess()) {
            String sliceUrn = sliceId.getValue();
            Slice slice = getModel().getSlice(sliceUrn);
            assert slice != null;
            @SuppressWarnings("unchecked")
            List<AnyCredential> credentialList = (List<AnyCredential>) result.getReply().getValue();

            if (credentialList != null) {
                if (credentialList.size() > 0)
                    slice.setCredentials(credentialList);

                credentialList.stream()
                        .filter(credential -> credential instanceof SfaCredential)
                        .filter(credential -> slice.getExpirationDate() == null
                                || credential.getExpiresDate().toInstant().isAfter(slice.getExpirationDate()))
                        .forEach(credential -> slice.setExpirationDate(credential.getExpiresDate().toInstant()));
            } else
                LOG.warn("Received an empty credential list for {}", sliceUrn);
        }
    }

    private void onLookupSliceInfoResult(ApiCallDetails result) {
        assert result.getReply() != null;

        if(result.getReply().getGeniResponseCode().isSuccess()){
            @SuppressWarnings("unchecked")
            AbstractFederationApi2.FederationApiReply<AbstractFederationApi.LookupResult> reply =
                    (AbstractFederationApi2.FederationApiReply<AbstractFederationApi.LookupResult>) result.getReply();
            if (reply.getValue() == null)
                return;

            AbstractFederationApi.LookupResult lookupRes = reply.getValue();

            for(int i = 0; i< lookupRes.size();i++){
                @SuppressWarnings("unchecked")
                Hashtable<String,Object> lookupObj = lookupRes.get(i);

                String lookupResSliceUrn = (String)lookupObj.get("SLICE_URN");
                GeniUrn sliceUrn = GeniUrn.parse(lookupResSliceUrn);

                String lookupResSliceExpiration = (String)lookupObj.get("SLICE_EXPIRATION");

                if(sliceUrn == null || lookupResSliceExpiration == null){
                    LOG.warn("lookupSliceInfo processing failed in EasyUniformFederationSliceAuthorityApiListener. Either the reply was incorrect, or there is a bug.\n" +
                            "SLICE_URN={}\n" +
                            "SLICE_EXPIRATION={}", lookupResSliceUrn, lookupResSliceExpiration);
                    continue;
                }

                Slice slice = getModel().logExistSlice(sliceUrn);

                try {
                    Date expireDate = RFC3339Util.rfc3339StringToDate(lookupResSliceExpiration.trim());
                    slice.setExpirationDate(expireDate.toInstant());
                } catch (ParseException e) {
                    //ignore
                    LOG.warn("SLIVER_INFO had invalid date: \"" + lookupResSliceExpiration + "\" (not RFC3339 compliant date format)");
                }
            }
        }
    }

    private void onLookupSliverInfoResult(ApiCallDetails result) throws JFedException {
        ResourceId sliceId = noteSliceUrnInParameters(result);
        noteUserCredentialInParameters(result);

        if (sliceId == null)
            throw new RuntimeException("Error, no slice urn found in parameters: " + result.getMethodParameters());

        assert result.getReply() != null;
        if (result.getReply().getGeniResponseCode().isSuccess()) {
            @SuppressWarnings("unchecked")
            AbstractFederationApi2.FederationApiReply<AbstractFederationApi.LookupResult> reply =
                    (AbstractFederationApi2.FederationApiReply<AbstractFederationApi.LookupResult>) result.getReply();
            if (reply.getValue() == null)
                return;
            AbstractFederationApi.LookupResult lookupRes = reply.getValue();
            //SLIVER_INFO_SLICE_URN 	URN 	URN of slice for registered sliver 	Yes 	Required 	No
            //SLIVER_INFO_URN 	URN 	URN of registered sliver 	Yes 	Required 	No
            //SLIVER_INFO_AGGREGATE_URN

            String sliceUrn = sliceId.getValue();
            Slice slice = getModel().getSlice(sliceUrn);
            assert slice != null;


            //log existence and non existence of slivers
            for (int i = 0; i < lookupRes.size(); i++) {
                @SuppressWarnings("unchecked")
                Hashtable<String, Object> lookupObj = lookupRes.get(i);
                String lookupResSliceUrn = (String) lookupObj.get("SLIVER_INFO_SLICE_URN");
                String lookupResSliverUrnString = (String) lookupObj.get("SLIVER_INFO_URN");
                String amUrnStr = (String) lookupObj.get("SLIVER_INFO_AGGREGATE_URN");

                if (lookupResSliceUrn == null || amUrnStr == null || lookupResSliverUrnString == null) {
                    LOG.warn("lookupSliverInfo processing failed in EasyUniformFederationSliceAuthorityApiListener. Either te reply was incorrect, or there is a bug." + "\n" +
                            " amUrnStr=" + amUrnStr + "\n" +
                            " lookupResSliceUrn=" + lookupResSliceUrn + "\n" +
                            " lookupResSliverUrn=" + lookupResSliverUrnString + "\n" +
                            " lookupRes=" + lookupRes);
                    continue;
                }

                AuthorityInfo ai =
                        authorityList.getByUrn(amUrnStr, AuthorityListModel.SubAuthMatchAllowed.ALLOW_TOPLEVEL, AuthorityListModel.SubAuthMatchPreference.PREFER_EXACT_SUBAUTHORITY);
                if (ai == null) continue;
                SfaAuthority auth = ai.getSfaAuthority();
                GeniUrn amUrn = auth.getUrn();
                assert amUrn.getValue().equals(amUrnStr);

                GeniUrn lookupResSliverUrn = GeniUrn.parse(lookupResSliverUrnString);
                if (lookupResSliverUrn == null) {
                    LOG.warn("lookupSliverInfo processing failed in EasyUniformFederationSliceAuthorityApiListener. Either te reply was incorrect, or there is a bug." + "\n" +
                            " amUrnStr=" + amUrnStr + "\n" +
                            " lookupResSliceUrn=" + lookupResSliceUrn + "\n" +
                            " lookupResSliverUrn=" + lookupResSliverUrnString + "   <= this is the problem\n" +
                            " lookupRes=" + lookupRes);
                    continue;
                }

                if (!lookupResSliverUrn.getEncodedTopLevelAuthority_withoutSubAuth().equals(auth.getUrn().getEncodedTopLevelAuthority_withoutSubAuth())) {
                    LOG.warn("lookupSliverInfo processing failed in EasyUniformFederationSliceAuthorityApiListener. Either te reply was incorrect, or there is a bug." + "\n" +
                            " amUrnStr=" + amUrnStr + "\n" +
                            " lookupResSliceUrn=" + lookupResSliceUrn + "\n" +
                            " lookupResSliverUrn=" + lookupResSliverUrnString + "   <= this is the problem, it doesn't match the authority.\n" +
                            " lookupRes=" + lookupRes);
                    continue;
                }

//                //log existence of sliver
//                //note that can we know anything for sure? we don't know if sliver is on AM with geni_single true...
//                getModel().logExistAtLeastOneSliver(sliceUrn, auth);
                //actually, we know the sliver URN!

                Sliver sliver = getModel().logExistSliver(sliceUrn, lookupResSliverUrnString, auth);
                //if expires unknown, set from this. (thus prefer other sources, and use this as fallback)
                if (sliver != null && sliver.getExpirationDate() != null) {
                    String lookupResSliverExpiration = (String) lookupObj.get("SLIVER_INFO_EXPIRATION ");
                    if (lookupResSliverExpiration != null)
                        try {
                            Date expireDate = RFC3339Util.rfc3339StringToDate(lookupResSliverExpiration.trim());
                            sliver.setExpirationDate(expireDate.toInstant());
                        } catch (ParseException e) {
                            //ignore
                            LOG.warn("SLIVER_INFO had invalid date: \"" + lookupResSliverExpiration + "\" (not RFC3339 compliant date format)");
                        }
                }

            }
            //this assumption goes to far! We can't be sure that it doesn't exist if it is not listed here.
//            List<Sliver> slivers = slice.getSlivers();
//            for (Sliver sliver : slivers) {
//                if (!amUrns.contains(sliver.getAuthority().getUrnString()))
//                    getModel().logNotExistSliverInSliceOnAuth(slice.getUrnString(), sliver.getAuthority());
//            }
        }
    }

    private void onGetSlicesForMember1(ApiCallDetails result) throws JFedException {
        noteUserCredentialInParameters(result);

        assert result.getReply() != null;
        if (result.getReply().getGeniResponseCode().isSuccess()) {
            @SuppressWarnings("unchecked")
            AbstractFederationApi1.FederationApiReply<List<FederationSliceAuthorityApi1.UrnRoleTuple>> reply =
                    (AbstractFederationApi1.FederationApiReply<List<FederationSliceAuthorityApi1.UrnRoleTuple>>) result.getReply();

            if (reply.getGeniResponseCode() == null || !reply.getGeniResponseCode().isSuccess() || reply.getValue() == null)
                return;

            List<FederationSliceAuthorityApi1.UrnRoleTuple> urnRoleTuples = reply.getValue();
            for (FederationSliceAuthorityApi1.UrnRoleTuple urnRoleTuple : urnRoleTuples) {
                String sliceUrnString = urnRoleTuple.getUrn().getValue();
                getModel().logExistSlice(sliceUrnString);
            }
        }
    }

    private void onGetSlicesForMember2(ApiCallDetails result) throws JFedException {
        noteUserCredentialInParameters(result);

        assert result.getReply() != null;
        if (result.getReply().getGeniResponseCode().isSuccess()) {
            @SuppressWarnings("unchecked")
            AbstractFederationApi2.FederationApiReply<List<FederationSliceAuthorityApi2.UrnRoleTuple>> reply =
                    (AbstractFederationApi2.FederationApiReply<List<FederationSliceAuthorityApi2.UrnRoleTuple>>) result.getReply();

            if (reply.getGeniResponseCode() == null || !reply.getGeniResponseCode().isSuccess() || reply.getValue() == null)
                return;

            List<FederationSliceAuthorityApi2.UrnRoleTuple> urnRoleTuples = reply.getValue();
            for (FederationSliceAuthorityApi2.UrnRoleTuple urnRoleTuple : urnRoleTuples) {
                String sliceUrnString = urnRoleTuple.getUrn().getValue();
                if (!urnRoleTuple.isExpired()) {
                    getModel().logExistSlice(sliceUrnString);
                }
            }
        }
    }

    private void onGetProjectsForMember2(ApiCallDetails result) throws JFedException {
        noteUserCredentialInParameters(result);

        assert result.getReply() != null;
        if (result.getReply().getGeniResponseCode().isSuccess()) {
            @SuppressWarnings("unchecked")
            AbstractFederationApi2.FederationApiReply<List<FederationSliceAuthorityApi2.UrnRoleTuple>> reply =
                    (AbstractFederationApi2.FederationApiReply<List<FederationSliceAuthorityApi2.UrnRoleTuple>>) result.getReply();

            if (reply.getGeniResponseCode() == null || !reply.getGeniResponseCode().isSuccess() || reply.getValue() == null)
                return;

            List<FederationSliceAuthorityApi2.UrnRoleTuple> urnRoleTuples = reply.getValue();
            urnRoleTuples.stream()
                    .filter(urnRoleTuple -> !urnRoleTuple.isExpired())
                    .forEach(urnRoleTuple -> {
                        String subAuthName = urnRoleTuple.getUrn().getEncodedResourceName();
                        getModel().addSubAuthorityName(subAuthName);
                    });
        }
    }

    /**
     * create_slice probably won't return a credential (not in default return, but could be in extension)! a call to get_credentials is needed
     */
    private void onCreateSlice(ApiCallDetails result) {
        LOG.debug("onCreateSlice -> called");
        noteUserCredentialInParameters(result);
//        ResourceId sliceId = noteSliceUrnInParameters(result); //only on success we can assume slice exists
        //we could also check other failures, for example, when calling register when a slice already exists.
        //but it's not really needed

        assert result.getReply() != null;
        if (result.getReply().getGeniResponseCode().isSuccess()) {
            @SuppressWarnings("unchecked")
            Hashtable<String, String> sliceInfo = (Hashtable<String, String>) result.getReply().getValue();
            if (sliceInfo.containsKey("SLICE_URN")) {
                LOG.debug("onCreateSlice -> logExistSlice " + sliceInfo.get("SLICE_URN"));
                getModel().logExistSlice(sliceInfo.get("SLICE_URN"));
            } else
                LOG.warn("onCreateSlice -> no URN returned for created slice!");
//            //not in standard apiv1!
//            if (sliceInfo.contains("SLICE_CREDENTIAL")) {
//                ResourceId sliceId = noteSliceUrnInParameters(result);
//                String sliceUrn = sliceId.getValue();
//                Slice slice = getModel().getSlice(sliceUrn);
//                assert slice != null;
//                slice.setCredential(AnyCredential.create("create_slice "+sliceUrn, sliceInfo.get("SLICE_CREDENTIAL"), "geni_sfa", "3")); //what type??
//            }
        }
    }

    private void onGetSliceMembers(ApiCallDetails result) {
        GeniUrn sliceUrn = (GeniUrn) result.getMethodParameters().get("urn");
        assert sliceUrn != null;
        assert result.getReply() != null;

        deriveSliceExistence(sliceUrn, result.getReply().getGeniResponseCode(), result.getAuthority());
        //user or slice credential here? noteUserCredentialInParameters(result);

        if (result.getReply().getGeniResponseCode().isSuccess() &&
                result.getReply().getValue() instanceof List) {
            Slice slice = getModel().getSlice(sliceUrn);
            assert slice != null;

            @SuppressWarnings("unchecked")
            List<FederationSliceAuthorityApi2.UrnRoleTuple> reply =
                    (List<FederationSliceAuthorityApi2.UrnRoleTuple>) result.getReply().getValue();
            List<GeniUrn> usersUrns = new ArrayList<>();
            for (FederationSliceAuthorityApi2.UrnRoleTuple roleTuple : reply) {
                GeniUrn userUrn = roleTuple.getUrn();
                usersUrns.add(userUrn);
            }
            slice.addUsers(usersUrns);
        }
    }

    private void onGetProjectMembers(ApiCallDetails result) {
        GeniUrn projectUrn = (GeniUrn) result.getMethodParameters().get("urn");
        assert projectUrn != null;
        noteUserCredentialInParameters(result);

        assert result.getReply() != null;
        if (result.getReply().getGeniResponseCode().isSuccess() &&
                result.getReply().getValue() instanceof List) {

            @SuppressWarnings("unchecked")
            List<FederationSliceAuthorityApi2.UrnRoleTuple> reply =
                    (List<FederationSliceAuthorityApi2.UrnRoleTuple>) result.getReply().getValue();
            List<GeniUrn> usersUrns = new ArrayList<>();
            for (FederationSliceAuthorityApi2.UrnRoleTuple roleTuple : reply) {
                GeniUrn userUrn = roleTuple.getUrn();
                usersUrns.add(userUrn);
            }
            getModel().setUsersOfSubAuthority(projectUrn.getEncodedResourceName(), usersUrns);
        }
    }

    @Override
    public void onResult(final ApiCallDetails details) {
        assert Platform.isFxApplicationThread();
        onResultInJavaFXThread(details);
    }

    public void onResultInJavaFXThread(ApiCallDetails result) {
        //ignore errors here
        if (result.getReply() == null || result.getJavaMethodName() == null)
            return;

        if (!result.getApiName().equals(FederationSliceAuthorityApi1.getApiName()) &&
                !result.getApiName().equals(FederationSliceAuthorityApi2.getApiName()))
            return;

        try {
            LOG.debug("EasyUniformFederationSliceAuthorityApiListener#onResult() methodName=" + result.getJavaMethodName() + " parameters.size(): " + result.getMethodParameters().size());

            if (result.getApiName().equals(FederationSliceAuthorityApi1.getApiName())) {
                if (result.getJavaMethodName().equals("getSliceCredentials"))
                    onGetSliceCredentials(result);

                if (result.getJavaMethodName().equals("createSlice"))
                    onCreateSlice(result);

                if (result.getJavaMethodName().equals("lookupSlicesForMember"))
                    onGetSlicesForMember1(result);
            }
            if (result.getApiName().equals(FederationSliceAuthorityApi2.getApiName()) &&
                    result.getMethodParameters().containsKey("type")) {
                LOG.debug("EasyUniformFederationSliceAuthorityApiListener#onResult() v2 type=" + result.getMethodParameters().get("type") + "  methodName=" + result.getGeniMethodName() + " javaMethodname=" + result.getJavaMethodName());

                if (result.getGeniMethodName().equals("get_credentials") &&
                        result.getMethodParameters().get("type").equals("SLICE"))
                    onGetSliceCredentials(result);

                if (result.getGeniMethodName().equals("lookup_members") &&
                        result.getMethodParameters().get("type").equals("SLICE"))
                    onGetSliceMembers(result);

                if (result.getGeniMethodName().equals("lookup_members") &&
                        result.getMethodParameters().get("type").equals("PROJECT"))
                    onGetProjectMembers(result);

                if (result.getJavaMethodName().equals("create") &&
                        result.getMethodParameters().get("type").equals("SLICE"))
                    onCreateSlice(result);

                if (result.getJavaMethodName().equals("lookup") &&
                        result.getMethodParameters().get("type").equals("SLIVER_INFO"))
                    onLookupSliverInfoResult(result);

                if (result.getJavaMethodName().equals("lookupForMember") &&
                        result.getMethodParameters().get("type").equals("SLICE"))
                    onGetSlicesForMember2(result);

                if (result.getJavaMethodName().equals("lookupForMember") &&
                        result.getMethodParameters().get("type").equals("PROJECT"))
                    onGetProjectsForMember2(result);
            }

        } catch (Exception e) {
            LOG.warn("Exception when processing SliceAuthority reply for EasyModel. This will be ignored, but it is most likely a bug. ", e);
        }
    }
}
