package be.iminds.ilabt.jfed.call_log_output;

import be.iminds.ilabt.jfed.highlevel.HighLevelModule;
import be.iminds.ilabt.jfed.log.cache.ApiCallDetailsCache;
import be.iminds.ilabt.jfed.util.IOUtils;
import be.iminds.ilabt.jfed.util.XmlUtil;
import org.testng.annotations.Guice;
import org.testng.annotations.Test;
import org.w3c.dom.Document;

import javax.inject.Inject;
import java.io.File;
import java.io.StringReader;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.*;
import static org.mockito.Mockito.mock;

@Test
@Guice(modules = HighLevelModule.class)
public class CallReportTest {

    @Inject
    private CallReportWriter callReportWriter;

    @Inject
    private CallReportReaderFactory callReportReaderFactory;

    @Inject
    private ApiCallDetailsCache cache;

    private CallReport makeReport() {
        CallReport callReport = new CallReport("State1", "repTitle", "repObject", cache);
        callReport.getCurState().debug("Debug1");
        callReport.getCurState().error("Error2");
        callReport.changeState("State2");
        callReport.getCurState().info("Info3");
        callReport.getCurState().warn("Warn4");
        return callReport;
    }

    @Test
    public void testLogLines() throws Exception {
        CallReport callReport = makeReport();
        assertThat(callReport.getLastState("State1").getLogLines(), hasSize(2));
        assertThat(callReport.getLastState("State2").getLogLines(), hasSize(2));

        assertThat(callReport.getLastState("State1").getLogLines().get(0).getType(), equalTo(LogOutput.LogLineType.DEBUG));
        assertThat(callReport.getLastState("State1").getLogLines().get(1).getType(), equalTo(LogOutput.LogLineType.ERROR));
        assertThat(callReport.getLastState("State2").getLogLines().get(0).getType(), equalTo(LogOutput.LogLineType.NOTE));
        assertThat(callReport.getLastState("State2").getLogLines().get(1).getType(), equalTo(LogOutput.LogLineType.WARN));

        assertThat(callReport.getLastState("State1").getLogLines().get(0).getText(), equalTo("Debug1"));
        assertThat(callReport.getLastState("State1").getLogLines().get(1).getText(), equalTo("Error2"));
        assertThat(callReport.getLastState("State2").getLogLines().get(0).getText(), equalTo("Info3"));
        assertThat(callReport.getLastState("State2").getLogLines().get(1).getText(), equalTo("Warn4"));
    }

    @Test
    public void testStateAccessors() throws Exception {
        CallReport callReport = new CallReport("State1", "repTitle", "repObject", cache);
        assertThat(callReport.getCurState().getName(), equalTo("State1"));

        callReport.changeState("State2");
        assertThat(callReport.getCurState().getName(), equalTo("State2"));

        assertThat(callReport.getLastState("State1").getName(), equalTo("State1"));
        assertThat(callReport.getLastState("State2").getName(), equalTo("State2"));
    }

    @Test
    public void testXml() throws Exception {
        //make a mock report, and write it to a file
        CallReport callReport = makeReport();

        File tmpFile = File.createTempFile("testToXmlOverview", ".xml");
        tmpFile.deleteOnExit();

        callReportWriter.writeCallReport(callReport, tmpFile, false);
        String content = IOUtils.fileToString(tmpFile);

        //first some small tests on the contents
        System.out.println("Xml version of report: \n\n" + content + "\n\n");
        assertThat(content.length(), greaterThan(30));
        Document xml = XmlUtil.parseXmlFromStringMakeExceptionsRuntime(content);
        assertThat(xml, hasXPath("/CallReport/states"));
        assertThat(xml, hasXPath("/CallReport/states/state[1]/name", equalTo("State1")));
        assertThat(xml, hasXPath("/CallReport/states/state[2]/name", equalTo("State2")));
        assertThat(xml, hasXPath("/CallReport/states/state[2]/logentries/logentry[2]/text", equalTo("Warn4")));
        assertThat(xml, hasXPath("/CallReport/states/state[2]/logentries/logentry[2]/type", equalTo("WARN")));

        tmpFile.delete();

        //now read the complete content of the file again

        CallRestoreContext mockedCallRestoreContext = mock(CallRestoreContext.class);

        CallReportReader callReportReader = callReportReaderFactory.createCallReportReader(mockedCallRestoreContext);

        CallReport reconstructedCallReport = callReportReader.readCallReport(new StringReader(content));
        assertThat(reconstructedCallReport.getCalls().size(), equalTo(callReport.getCalls().size()));
        assertThat(reconstructedCallReport.getLastState("State1").getName(), equalTo(callReport.getLastState("State1").getName()));
        assertThat(reconstructedCallReport.getLastState("State1").getLogLines(), equalTo(callReport.getLastState("State1").getLogLines()));
        assertThat(reconstructedCallReport.getLastState("State1").getDurationMs(), equalTo(callReport.getLastState("State1").getDurationMs()));
        assertThat(reconstructedCallReport.getLastState("State1").getEnd(), equalTo(callReport.getLastState("State1").getEnd()));
        assertThat(reconstructedCallReport.getLastState("State1").getStateStart(), equalTo(callReport.getLastState("State1").getStateStart()));
        assertThat(reconstructedCallReport.getLastState("State2").getName(), equalTo(callReport.getLastState("State2").getName()));
        assertThat(reconstructedCallReport.getLastState("State2").getLogLines(), equalTo(callReport.getLastState("State2").getLogLines()));
        assertThat(reconstructedCallReport.getLastState("State2").getDurationMs(), equalTo(callReport.getLastState("State2").getDurationMs()));
        assertThat(reconstructedCallReport.getLastState("State2").getEnd(), equalTo(callReport.getLastState("State2").getEnd()));
        assertThat(reconstructedCallReport.getLastState("State2").getStateStart(), equalTo(callReport.getLastState("State2").getStateStart()));
    }

    @Test
    public void testToHtml() throws Exception {
        CallReport callReport = makeReport();

        File tmpFile = File.createTempFile("testToHtml", ".html");
        tmpFile.deleteOnExit();

        callReport.toHtml(tmpFile);
        String content = IOUtils.fileToString(tmpFile);
        assertThat(content.length(), greaterThan(30));
        //only some basic, very simple checks
        assertThat(content, containsString("Debug1"));
        assertThat(content, containsString("Error2"));
        assertThat(content, containsString("Info3"));
        assertThat(content, containsString("Warn4"));
        assertThat(content, containsString("repTitle"));
        assertThat(content, containsString("repObject"));

        tmpFile.delete();
    }

//    @Test
//    public void unmarshalTest() throws Exception {
//        File tmp = new File("/home/wim/bugreport-201502181231.xml");
//        JAXBContext jc = JAXBContext.newInstance(
//                SerializableApiCallDetails.class,
//                LogOutput.LogLine.class,
//                JFedConnection.ProxyInfo.class,
//                CallReport.class);
//
//        Unmarshaller u = jc.createUnmarshaller();
//        CallReport callReport = (CallReport) u.unmarshal(new FileReader(tmp));
//        assertThat(callReport.getStates(), hasSize(greaterThan(1)));
//    }
}
