package be.iminds.ilabt.jfed.lowlevel.connection;

import org.apache.http.Header;
import org.apache.http.HttpEntity;
import org.apache.http.HttpResponse;
import org.apache.http.StatusLine;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.entity.ContentType;
import org.apache.http.entity.StringEntity;
import org.apache.http.impl.client.DefaultHttpClient;
import org.apache.http.util.EntityUtils;
import org.apache.xmlrpc.XmlRpcClientException;
import org.apache.xmlrpc.XmlRpcTransport;
import org.apache.xmlrpc.XmlRpcTransportFactory;
import org.slf4j.LoggerFactory;

import javax.net.ssl.SSLException;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Scanner;

/**
 * CommonsHttpClientXmlRpcTransportFactory is an implementation of the apache commons XML RPC library {@link XmlRpcTransportFactory} interface.
 * which makes use of the apache commons {@link DefaultHttpClient}.
 * <p/>
 * It also captures the data sent and received, for debugging purposes.
 */
public class CommonsHttpClientXmlRpcTransportFactory implements HttpInfoXmlRpcTransportFactory {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(CommonsHttpClientXmlRpcTransportFactory.class);

    private final String serverUrlStr;
    private final LogDebugInfoCallback logDebugInfoCallback;
    StringBuilder httpResult = new StringBuilder(1000);
    private DefaultHttpClient httpClient;
    /* debug will print out a info during the call */
    private boolean debugMode;
    private String httpSent = "";
    private StringBuilder httpRequestHeaders = null;
    private StringBuilder httpResponseHeaders = null;
    private List<Header> rawHttpResponseHeaders = null;
    private StatusLine httpResultStatusLine = null;
    private String httpRequestLine = null;
    private MyXmlRpcTransport transport = null;

    /**
     * Note: the "Http History" automatically resets on new XmlRpc call
     */
    public CommonsHttpClientXmlRpcTransportFactory(String serverUrlStr, DefaultHttpClient httpClient, boolean debugMode, LogDebugInfoCallback logDebugInfoCallback) {
        this.serverUrlStr = serverUrlStr;
        this.httpClient = httpClient;
        this.debugMode = debugMode;
        this.logDebugInfoCallback = logDebugInfoCallback;
    }

    /**
     * debug mode will print out a info during the call
     */
    @Override
    public boolean isDebugMode() {
        return debugMode;
    }

    /**
     * debug mode will print out a info during the call
     */
    @Override
    public void setDebugMode(boolean debug) {
        this.debugMode = debug;
    }

    @Override
    public String getHttpReceivedHistory() {
        return httpResult.toString();
    }

    @Override
    public String getHttpSentHistory() {
        return httpSent;
    }

    @Override
    public StatusLine getHttpStatusLine() {
        return httpResultStatusLine;
    }

    @Override
    public String getHttpRequestLine() {
        return httpRequestLine;
    }

    @Override
    public String getHttpRequestHeaders() {
        if (httpRequestHeaders == null) return null;
        return httpRequestHeaders.toString();
    }

    @Override
    public String getHttpResponseHeaders() {
        if (httpResponseHeaders == null) return null;
        return httpResponseHeaders.toString();
    }
    @Override
    public List<Header> getRawHttpResponseHeaders() {
        if (rawHttpResponseHeaders == null) return null;
        return rawHttpResponseHeaders;
    }

    private void resetHttpHistory() {
        httpResult = new StringBuilder(1000);
        httpRequestHeaders = null;
        httpResultStatusLine = null;
        httpRequestLine = null;
        httpResponseHeaders = null;
        rawHttpResponseHeaders = null;
        httpSent = "";
    }

    @Override
    public void clearLastCallData() {
        resetHttpHistory();
    }

    @Override
    public void close() {
        try {
            transport.endClientRequest();
        } catch (XmlRpcClientException e) {
            //ignore
        }
        httpClient = null;
        transport = null;
        resetHttpHistory();
    }

    @Override
    public XmlRpcTransport createTransport() throws org.apache.xmlrpc.XmlRpcClientException {
        if (transport == null)
            transport = new MyXmlRpcTransport();
        return transport;
    }

    @Override
    public void setProperty(String s, Object o) {
    }

    public interface LogDebugInfoCallback {
        public void logDebugInfo(org.slf4j.Logger log);
    }

    private class SpyingInputStream extends InputStream {
        private InputStream is;

        public SpyingInputStream(InputStream is) {
            this.is = is;
        }

        @Override
        public int read() throws IOException {
            int r = is.read();
            if (r >= 0)
                httpResult.append((char) r);
            return r;
        }
    }

    class MyXmlRpcTransport implements XmlRpcTransport {
        HttpEntity entity = null;

        public InputStream sendXmlRpc(byte[] bytes) throws IOException, org.apache.xmlrpc.XmlRpcClientException {

            //note: we cannot assume the xmlrpc data of one command is sent in one call to this method
            if (entity == null)
                resetHttpHistory();

//            HttpPut httpput = new HttpPut(serverUrlStr);
            HttpPost httpput = new HttpPost(serverUrlStr);
            httpSent += httpput;
            StringEntity myEntity = new StringEntity(new String(bytes), ContentType.create("text/xml", "UTF-8"));
            httpput.setEntity(myEntity);
            HttpResponse response = null;

            try {
                httpSent += new String(bytes);

                httpRequestLine = httpput.getRequestLine().toString();

                httpRequestHeaders = new StringBuilder(200);
                for (Header header : httpput.getAllHeaders())
                    httpRequestHeaders.append(header.toString() + "\n");
                httpRequestHeaders.append("\n");

                if (debugMode)
                    LOG.debug("MyXmlRpcTransport.sendXmlRpc is sending " + bytes.length + " bytes");
                response = httpClient.execute(httpput);

                //not a good way to do this
                if (Thread.currentThread().isInterrupted())
                    throw new XmlRpcClientException("cancelled", new InterruptedException());

                if (debugMode)
                    LOG.debug("MyXmlRpcTransport.sendXmlRpc has sent. response.getStatusLine()=" + response.getStatusLine());

                if (response != null && response.getStatusLine() != null)
                    httpResultStatusLine = response.getStatusLine();

                if (response != null && response.getAllHeaders() != null) {
                    httpResponseHeaders = new StringBuilder(100);
                    for (Header header : response.getAllHeaders())
                        httpResponseHeaders.append(header.toString() + "\n");
                    httpResponseHeaders.append("\n");

                    rawHttpResponseHeaders = new ArrayList<Header>(Arrays.asList(response.getAllHeaders()));
                }

                //TODO: The following check is probably wrong! Exception handling might be better if it is fixed?
                if (response.getStatusLine() != null &&
                        (response.getStatusLine().getStatusCode() == 500  //500 Internal Server Error
                                ||
                                response.getStatusLine().getStatusCode() == 503) //503 Service Unavailable
                        )
                    throw new HttpServerErrorException(response.getStatusLine().getStatusCode(), response.getStatusLine().getReasonPhrase());
            } catch (javax.net.ssl.SSLPeerUnverifiedException ex) {
                LOG.error("SSLPeerUnverifiedException exception.\nWill also call LogDebugInfoCallback to retrieve connection debug info.", ex);
                if (logDebugInfoCallback != null) logDebugInfoCallback.logDebugInfo(LOG);

                throw new org.apache.xmlrpc.XmlRpcClientException("The server certificate for '" + httpput.getURI() + "' could not be verified. " +
                        "Error message \"" + ex.getMessage() + "\". " +
                        "Possible causes:\n" +
                        "  - The server's self-signed certificate is not in our trust store. \n" +
                        "  - The server's certificate \"CN\" field is not the server hostname or a known alias. \n" +
                        "  - The server is not accepting our client login certificate+key pair (perhaps is is not federated with our \"login provider\"). \n" +
                        "  - If the server's certificate is not self signed, we might not have the root certificate of the trust chain in our trust store.", ex);
            } catch (SSLException ex) {
                throw new org.apache.xmlrpc.XmlRpcClientException("SSLException: " + ex.getMessage(), ex);
            } catch (HttpServerErrorException ex) {
                /*first read received entity for logs*/
                entity = response.getEntity();
                if (entity != null) {
                    InputStream is = new SpyingInputStream(entity.getContent());
                    new Scanner(is, "UTF-8").useDelimiter("\\A").next(); //this reads entire stream. We discard the result.
                }
                throw new org.apache.xmlrpc.XmlRpcClientException("HTTP Server error", ex);
            }
            entity = response.getEntity();

            return new SpyingInputStream(entity.getContent());
            //return entity.getContent();
        }

        public void endClientRequest() throws org.apache.xmlrpc.XmlRpcClientException {
            if (entity != null) try {
                EntityUtils.consume(entity);
            } catch (IOException e) {
                e.printStackTrace();
            }
            entity = null;
        }
    }
}
