package be.iminds.ilabt.jfed.lowlevel.connection;

import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.util.KeyUtil;
import be.iminds.ilabt.jfed.util.SSLCertificateDownloader;
import be.iminds.ilabt.jfed.util.SocksProxyHelper;
import be.iminds.ilabt.jfed.util.SshServerProxyHelper;
import org.apache.http.conn.scheme.Scheme;
import org.apache.http.conn.ssl.BrowserCompatHostnameVerifier;
import org.apache.http.conn.ssl.SSLSocketFactory;
import org.apache.http.impl.client.DefaultHttpClient;
import org.apache.http.params.BasicHttpParams;
import org.apache.http.params.CoreConnectionPNames;
import org.apache.http.params.HttpParams;
import org.slf4j.LoggerFactory;

import java.net.InetSocketAddress;
import java.net.MalformedURLException;
import java.net.URL;
import java.security.KeyStore;
import java.security.PrivateKey;
import java.security.SecureRandom;
import java.security.cert.Certificate;
import java.security.cert.CertificateExpiredException;
import java.security.cert.CertificateNotYetValidException;
import java.security.cert.X509Certificate;
import java.util.*;

/**
 * HttpsClientWithUserAuthenticationFactory
 */
public class HttpsClientWithUserAuthenticationFactory {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(HttpsClientWithUserAuthenticationFactory.class);
    private static int extraTrustCount = 7000;

    public static DefaultHttpClient getHttpClient(JFedConnection.ProxyInfo proxyInfo,
                                                  List<X509Certificate> clientCertificateChain,
                                                  PrivateKey privateKey,
                                                  String serverUrlStr,
                                                  KeyStore trustStore,
                                                  final Collection<String> allowedCertificateHostnameAliasesOrig,
                                                  boolean debugMode, final HandleUntrustedCallback handleUntrustedCallback) throws JFedException {
        if (clientCertificateChain == null) throw new RuntimeException("clientCertificateChain == null");
        if (clientCertificateChain.size() == 0) throw new RuntimeException("clientCertificateChain is empty");
        if (privateKey == null) throw new RuntimeException("privateKey == null");
        if (serverUrlStr == null) throw new RuntimeException("serverUrlStr == null");

        final List<String> allowedCertificateHostnameAliases =
                new ArrayList<String>(allowedCertificateHostnameAliasesOrig);

        final URL serverUrl;
        try {
            serverUrl = new URL(serverUrlStr);
        } catch (MalformedURLException e) {
            LOG.error("ERROR: MalformedURLException url=\"" + serverUrlStr + "\"", e);
            return null;
        }

        try {
            String key_store_pass =
                    "somepass"; //password for internal use. We don;t really need one as we'll never save this keystore.
            String privatekey_store_pass =
                    "someotherpass"; //password for internal use. We don't really need one as we'll never save this keystore or share it with other objects

            KeyStore keyStore = KeyStore.getInstance("JKS", "SUN");
            keyStore.load(null, key_store_pass.toCharArray());

            // Create CertificateChain
            Certificate[] certs = new Certificate[clientCertificateChain.size()];
            for (int i = 0; i < clientCertificateChain.size(); i++) {
                X509Certificate c = clientCertificateChain.get(i);
                certs[i] = c;
                try {
                    //FEDIBBTDEV-878: check certificate validity
                    c.checkValidity(new Date());
                } catch (CertificateExpiredException e) {
                    throw new JFedException("Certificate "+i+" (of "+clientCertificateChain.size()+") in the user " +
                            "certificate chain has expired. NotAfter="+c.getNotAfter()+" now="+new Date());
                } catch (CertificateNotYetValidException e) {
                    throw new JFedException("Certificate "+i+" (of "+clientCertificateChain.size()+") in the user " +
                            "certificate chain is not yet valid. NotBefore="+c.getNotBefore()+" now="+new Date());
                }
            }

            // storing private key and certificate to keystore
            keyStore.setKeyEntry("authority", privateKey, privatekey_store_pass.toCharArray()/*password is used to encrypt the key inside the keystore*/, certs);

            ////////// DEBUG //////////////
            //add last client certificate to trust store as well as keystore!
            //            trustStore.setCertificateEntry("clientCertificate", clientCertificateChain.get(clientCertificateChain.size()-1));
            ////////// DEBUG //////////////

            //add all client certificates to trust store as well as keystore!
            for (int i = 0; i < clientCertificateChain.size(); i++)
                trustStore.setCertificateEntry("clientCertificate" + i, clientCertificateChain.get(i));

            /************ Init Trust Store *************/
            HttpParams params = new BasicHttpParams();
            params.setParameter(CoreConnectionPNames.SO_TIMEOUT, 240000); //read timeout in ms (default: 0 = infinite)   (long timeout, as some calls take time to finish)
            params.setParameter(CoreConnectionPNames.CONNECTION_TIMEOUT, 10000); //connection timeout in ms (default: 0 = infinite)

            DefaultHttpClient httpclient = null;
            if (System.getProperty("proxySet") != null &&
                    System.getProperty("proxySet").equals("true") &&
                    System.getProperty("socksProxyHost") != null) {
                LOG.info("Using SOCKS proxy: " + System.getProperty("socksProxyHost"));
                httpclient = SocksProxyHelper.getHttpClientOverSocksProxy(params);
            } else {
                httpclient = new DefaultHttpClient(params);
            }

            String algo = SSLSocketFactory.TLS;
            java.security.SecureRandom random = new SecureRandom();

            SSLSocketFactory socketFactory;
            if (handleUntrustedCallback != null) {
                //                socketFactory = new SSLSocketFactory(TRUSTALL, new AllowAllHostnameVerifier()); //doesn't work as we need client authentication

                //didn't work for some reason. Probably it's missing client certificate (need to check apache source code of SSLSocketFactory to correctly write this)
                //                socketFactory = new MyInsecureSSLAcceptAllSocketFactory(algo, keyStore, privatekey_store_pass/*password is used to encrypt the key inside the keystore*/, trustStore, random, hostnameVerifier);

                //Solved by connecting 2 times. Not best solution, but it works.
                SSLCertificateDownloader.SSLCertificateJFedInfo certInfo =
                        SSLCertificateDownloader.getCertificateInfo(serverUrl);
                if (certInfo != null && certInfo.isSelfSigned() != null && certInfo.isSelfSigned()) {
                    if (handleUntrustedCallback.trust(certInfo)) {
                        trustStore.setCertificateEntry("allTrustCert" + extraTrustCount, certInfo.getCert());
                        if (!certInfo.getSubjectMatchesHostname())
                            allowedCertificateHostnameAliases.add(certInfo.getSubject());
                    } else {
                        LOG.info("User does not trust certificate -> Not adding anything to trust store.");
                    }
                }
                org.apache.http.conn.ssl.X509HostnameVerifier hostnameVerifier = makeHostNameVerifier(allowedCertificateHostnameAliases);

                socketFactory =
                        new SSLSocketFactory(algo, keyStore, privatekey_store_pass/*password is used to encrypt the key inside the keystore*/, trustStore, random, hostnameVerifier);
            } else {

                if (proxyInfo != null && proxyInfo instanceof JFedConnection.SshProxyInfo) {
                    JFedConnection.SshProxyInfo sshProxyInfo = (JFedConnection.SshProxyInfo) proxyInfo;

                    LOG.debug("Using SSH Proxy for connection: " + sshProxyInfo.getHostname());

                    org.apache.http.conn.ssl.X509HostnameVerifier hostnameVerifier = makeHostNameVerifier(allowedCertificateHostnameAliases);

                    SSLSocketFactory sslSocketFactory = new SshServerProxyHelper.SslOverSshProxySocketFactory(
                            new SshServerProxyHelper.SshProxyInfo(
                                    new InetSocketAddress(sshProxyInfo.getHostname(), sshProxyInfo.getPort()),
                                    sshProxyInfo.getUsername(),
                                    sshProxyInfo.getHostKey(),
                                    new String(KeyUtil.privateKeyToAnyPem(sshProxyInfo.getSshKeyInfo().getPrivateKey()))
                            ),
                            algo,
                            keyStore,
                            privatekey_store_pass /* password is used to encrypt the key inside the keystore */,
                            trustStore,
                            random,
                            hostnameVerifier
                    );
                    socketFactory = sslSocketFactory;
                } else {
                    org.apache.http.conn.ssl.X509HostnameVerifier hostnameVerifier = makeHostNameVerifier(allowedCertificateHostnameAliases);

                    SSLSocketFactory sslSocketFactory = new SSLSocketFactory(
                            algo,
                            keyStore,
                            privatekey_store_pass/*password is used to encrypt the key inside the keystore*/,
                            trustStore,
                            random,
                            hostnameVerifier);
                    socketFactory = sslSocketFactory;
                }
            }

            Scheme sch1 = new Scheme("https", 443, socketFactory); //port is the default port for the given protocol
            httpclient.getConnectionManager().getSchemeRegistry().register(sch1);


            return httpclient;
        } catch (Exception e) {
            throw new JFedException("Error creating SSL connection to " + serverUrlStr, e);
        }
    }

    public static class INSECURE_TRUSTALL_HandleUntrustedCallback implements HandleUntrustedCallback {
        public INSECURE_TRUSTALL_HandleUntrustedCallback() {
            LOG.warn("SECURITY WARNING: constructing INSECURE_TRUSTALL_HandleUntrustedCallback");
        }

        @Override
        public boolean trust(SSLCertificateDownloader.SSLCertificateJFedInfo sSLCertificateJFedInfo) {
            return true;
        }
    }


    static org.apache.http.conn.ssl.X509HostnameVerifier makeHostNameVerifier(final List<String> allowedCertificateHostnameAliases) {
        return new org.apache.http.conn.ssl.AbstractVerifier() {
            private BrowserCompatHostnameVerifier base = new BrowserCompatHostnameVerifier();

            /**
             * @param host the hostname of the url to check.
             * @param cns the CN's of the certificate
             * @param subjectAlts the aliases fot the CN's that the certificate allows
             * @throws javax.net.ssl.SSLException
             */
            @Override
            public final void verify(java.lang.String host, String[] cns, String[] subjectAlts) throws javax.net.ssl.SSLException {
                List<String> modifiedSubjectAlts = new ArrayList<String>();
                List<String> modifiedCns = new ArrayList<String>();


                if (subjectAlts != null)
                    Collections.addAll(modifiedSubjectAlts, subjectAlts);
                if (cns != null)
                    Collections.addAll(modifiedCns, cns);

                //force add of host itself, this makes any certificate we have stored match
                for (String allowedAlias : allowedCertificateHostnameAliases)
                    if (modifiedCns.contains(allowedAlias) || modifiedSubjectAlts.contains(allowedAlias) || allowedAlias.equals("*")) {
                        modifiedSubjectAlts.clear();
                        modifiedCns.clear();

                        modifiedSubjectAlts.add(host);
                        modifiedCns.add(host);
                        break;
                    }

                String[] newSubjectAlts = new String[modifiedSubjectAlts.size()];
                for (int i = 0; i < modifiedSubjectAlts.size(); i++)
                    newSubjectAlts[i] = modifiedSubjectAlts.get(i);

                String[] newCns = new String[modifiedCns.size()];
                for (int i = 0; i < modifiedCns.size(); i++)
                    newCns[i] = modifiedCns.get(i);

                try {
                    base.verify(host, newCns, newSubjectAlts);
                } catch (javax.net.ssl.SSLException e) {
                    LOG.warn("HttpsClientWithUserAuthenticationFactory hostnameVerifier host name verification failed:" + e.getMessage() + "\n" +
                            "host=" + host + "\n" +
                            "cns=" + (cns == null ? "null" : Arrays.toString(cns)) + "\n" +
                            "subjectAlts=" + Arrays.toString(subjectAlts) + "\n" +
                            "newCns=" + modifiedCns + "\n" +
                            "newSubjectAlts=" + modifiedSubjectAlts+"\n"+
                            "allowedCertificateHostnameAliases=" + allowedCertificateHostnameAliases
                            , e);
                    throw e;
                }
            }
        };
    }
}
