package be.iminds.ilabt.jfed.lowlevel.connection;

import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.util.KeyUtil;
import be.iminds.ilabt.jfed.util.SSLCertificateDownloader;
import be.iminds.ilabt.jfed.util.SocksProxyHelper;
import be.iminds.ilabt.jfed.util.SshServerProxyHelper;
import org.apache.http.conn.scheme.Scheme;
import org.apache.http.conn.ssl.SSLSocketFactory;
import org.apache.http.impl.client.DefaultHttpClient;
import org.apache.http.params.BasicHttpParams;
import org.apache.http.params.CoreConnectionPNames;
import org.apache.http.params.HttpParams;
import org.slf4j.LoggerFactory;

import java.net.InetSocketAddress;
import java.net.MalformedURLException;
import java.net.URL;
import java.security.KeyStore;
import java.security.SecureRandom;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

/**
 * HttpsClientWithUserAuthenticationFactory
 */
public class HttpsClientWithoutUserAuthenticationFactory {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(HttpsClientWithoutUserAuthenticationFactory.class);
    private static int extraTrustCount = 7000;

    public static DefaultHttpClient getHttpClient(JFedConnection.ProxyInfo proxyInfo,
                                                  String serverUrlStr,
                                                  KeyStore trustStore,
                                                  final Collection<String> allowedCertificateHostnameAliasesOrig,
                                                  boolean debugMode, HandleUntrustedCallback handleUntrustedCallback) throws JFedException {
        if (serverUrlStr == null) throw new RuntimeException("serverUrlStr == null");

        final List<String> allowedCertificateHostnameAliases =
                new ArrayList<String>(allowedCertificateHostnameAliasesOrig);

        final URL serverUrl;
        try {
            serverUrl = new URL(serverUrlStr);
        } catch (MalformedURLException e) {
            LOG.error("ERROR: MalformedURLException url=\"" + serverUrlStr + "\"", e);
            return null;
        }

        try {
            /************ Init Trust Store *************/
            HttpParams params = new BasicHttpParams();
            params.setParameter(CoreConnectionPNames.SO_TIMEOUT, 120000); //read timeout in ms (default: 0 = infinite)   (long timeout, as some calls take time to finish)
            params.setParameter(CoreConnectionPNames.CONNECTION_TIMEOUT, 10000); //connection timeout in ms (default: 0 = infinite)

            DefaultHttpClient httpclient = null;
            if (System.getProperty("proxySet") != null &&
                    System.getProperty("proxySet").equals("true") &&
                    System.getProperty("socksProxyHost") != null) {
                LOG.info("Using SOCKS proxy: " + System.getProperty("socksProxyHost"));
                httpclient = SocksProxyHelper.getHttpClientOverSocksProxy(params);
            } else {
                httpclient = new DefaultHttpClient(params);
            }

            String algo = SSLSocketFactory.TLS;
            SecureRandom random = new SecureRandom();

            SSLSocketFactory socketFactory;
            if (handleUntrustedCallback != null) {
                //                socketFactory = new SSLSocketFactory(TRUSTALL, new AllowAllHostnameVerifier()); //doesn't work as we need client authentication

                //didn't work for some reason. Probably it's missing client certificate (need to check apache source code of SSLSocketFactory to correctly write this)
                //                socketFactory = new MyInsecureSSLAcceptAllSocketFactory(algo, keyStore, privatekey_store_pass/*password is used to encrypt the key inside the keystore*/, trustStore, random, hostnameVerifier);

                //Solved by connecting 2 times. Not best solution, but it works.
                SSLCertificateDownloader.SSLCertificateJFedInfo certInfo =
                        SSLCertificateDownloader.getCertificateInfo(serverUrl);
                if (certInfo != null && certInfo.isSelfSigned() != null && certInfo.isSelfSigned()) {
                    if (handleUntrustedCallback.trust(certInfo)) {
                        trustStore.setCertificateEntry("allTrustCert" + extraTrustCount, certInfo.getCert());
                        if (!certInfo.getSubjectMatchesHostname())
                            allowedCertificateHostnameAliases.add(certInfo.getSubject());
                    } else {
                        LOG.info("User does not trust certificate -> Not adding anything to trust store.");
                    }
                }
            }

            org.apache.http.conn.ssl.X509HostnameVerifier hostnameVerifier = HttpsClientWithUserAuthenticationFactory.makeHostNameVerifier(allowedCertificateHostnameAliases);

            if (proxyInfo != null && proxyInfo instanceof JFedConnection.SshProxyInfo) {
                JFedConnection.SshProxyInfo sshProxyInfo = (JFedConnection.SshProxyInfo) proxyInfo;

                LOG.debug("Using SSH Proxy for connection: " + sshProxyInfo.getHostname());

                SSLSocketFactory sslSocketFactory = new SshServerProxyHelper.SslOverSshProxySocketFactory(
                        new SshServerProxyHelper.SshProxyInfo(
                                new InetSocketAddress(sshProxyInfo.getHostname(), sshProxyInfo.getPort()),
                                sshProxyInfo.getUsername(),
                                sshProxyInfo.getHostKey(),
                                new String(KeyUtil.privateKeyToAnyPem(sshProxyInfo.getSshKeyInfo().getPrivateKey()))
                        ),
                        algo,
                        null /*keyStore*/,
                        null /*privatekey_store_pass*/,
                        trustStore,
                        random,
                        hostnameVerifier
                );
                socketFactory = sslSocketFactory;
            } else {
                SSLSocketFactory sslSocketFactory = new SSLSocketFactory(
                        algo,
                        null /*keyStore*/,
                        null /*privatekey_store_pass*/,
                        trustStore,
                        random,
                        hostnameVerifier);
                socketFactory = sslSocketFactory;
            }

            Scheme sch1 = new Scheme("https", 443, socketFactory); //port is the default port for the given protocol
            httpclient.getConnectionManager().getSchemeRegistry().register(sch1);


            return httpclient;
        } catch (Exception e) {
            throw new JFedException("Error creating SSL connection to " + serverUrlStr, e);
        }
    }

    public static class INSECURE_TRUSTALL_HandleUntrustedCallback implements HandleUntrustedCallback {
        public INSECURE_TRUSTALL_HandleUntrustedCallback() {
            LOG.warn("SECURITY WARNING: constructing INSECURE_TRUSTALL_HandleUntrustedCallback");
        }

        @Override
        public boolean trust(SSLCertificateDownloader.SSLCertificateJFedInfo sSLCertificateJFedInfo) {
            return true;
        }
    }
}
