package be.iminds.ilabt.jfed.lowlevel.connection;

import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.util.JFedTrustStore;
import org.apache.xmlrpc.XmlRpcClient;
import org.slf4j.Logger;

import java.net.URL;
import java.security.KeyStore;
import java.security.KeyStoreException;

/**
 * SfaSslWithClientAuthConnection: XmlRpc over HTTPs, which doesn't use SSL with client authentication
 * THIS IS NOT A NORMAL SFA CONNECTION. This is only used for GetVersion etc...
 */
public class SfaSslWithoutClientAuthConnection extends SfaConnection {
    private CommonsHttpClientXmlRpcTransportFactory xmlRpcTransportFactory;
    private XmlRpcClient xmlRpcClient;


    /**
     * Create a connection. This constructor assumes the url is to the specified sfaAuthority and the client authentication uses the specified user
     *
     * @param serverUrl url to connect to (may not be null)
     * @param sfaAuthority the authority connected to this URL. Will be used to get trusted certificate info, and to fill in debug info. (may not be null)
     * @param proxyInfo the proxy to use (may be null if no proxy)
     * @param handleUntrustedCallback the callback to use when the servers certificate is not trusted. This is typically used to communicate with the user about untrusted self-signed certificates. (this may be null, which will cause connection failure for untrusted certificates)
     * @param debugMode if true, additional logging is enabled
     *
     * @throws be.iminds.ilabt.jfed.lowlevel.JFedException
     */
    public SfaSslWithoutClientAuthConnection(String serverUrl,
                                             SfaAuthority sfaAuthority,
                                             ProxyInfo proxyInfo,
                                             HandleUntrustedCallback handleUntrustedCallback,
                                             boolean debugMode) throws JFedException {
        this(serverUrl, new JFedTrustStore(sfaAuthority), proxyInfo, handleUntrustedCallback, new DebugInfo(debugMode, sfaAuthority, null));
    }
    /**
     * Create a connection by specifying all low level details needed.
     *
     * @param serverUrl url to connect to (may not be null)
     * @param jFedTrustStore The trust store to use to verify the server's identity. (may not be null, you can always use the {@link be.iminds.ilabt.jfed.util.JFedTrustStore#JFedTrustStore()} constructor to use the system trust store)
     * @param proxyInfo the proxy to use (may be null if no proxy)
     * @param handleUntrustedCallback the callback to use when the servers certificate is not trusted. This is typically used to communicate with the user about untrusted self-signed certificates. (this may be null, which will cause connection failure for untrusted certificates)
     * @param debugInfo debug information, such as authority and user, and whether debug mode is enabled. (may be null)
     * @throws JFedException
     */
    public SfaSslWithoutClientAuthConnection(final String serverUrl,
                                             final JFedTrustStore jFedTrustStore,
                                             final ProxyInfo proxyInfo,
                                             final HandleUntrustedCallback handleUntrustedCallback,
                                             DebugInfo debugInfo) throws JFedException {
        super(serverUrl, jFedTrustStore, proxyInfo, debugInfo);

        error = true;

        if (serverUrl == null) throw new IllegalArgumentException("Illegal argument: serverURL == null");
        if (jFedTrustStore == null) throw new IllegalArgumentException("Illegal argument: jFedTrustStore == null");

        final KeyStore trustStore = jFedTrustStore.getTrustStore();

        final SfaAuthority debugAuth = this.debugInfo.getAuth();

        xmlRpcTransportFactory = new CommonsHttpClientXmlRpcTransportFactory(
                serverUrl,
                HttpsClientWithoutUserAuthenticationFactory.getHttpClient(
                        proxyInfo,
                        serverUrl,
                        trustStore,
                        jFedTrustStore.getAllowedServerCertificateHostnameAliases(),
                        this.debugInfo.isDebugMode(),
                        handleUntrustedCallback
                ),
                this.debugInfo.isDebugMode(),
                new CommonsHttpClientXmlRpcTransportFactory.LogDebugInfoCallback() {
                    @Override
                    public void logDebugInfo(Logger log) {
                        String trustStoreAliases = null;
                        try {
                            if (trustStore != null && trustStore.aliases() != null)
                                trustStoreAliases = trustStore.aliases().toString();
                        } catch (KeyStoreException e) {
                            trustStoreAliases = "Exception: " + e.getMessage();
                        }

                        log.info("SfaSslWithoutClientAuthConnection debug info:\n" +
                                "proxyInfo=" + proxyInfo + "\n" +
                                "geniAuthority=" + (debugAuth == null ? null
                                : debugAuth.getUrnString()) + "\n" +
                                "geniAuthority allowed host name aliases=" + (debugAuth == null ? null
                                : debugAuth.getAllowedCertificateHostnameAliases()) + "\n" +
                                "geniAuthority pem trust certs=" + (debugAuth == null ? null
                                : debugAuth.getPemSslTrustCerts()) + "\n" +
                                "serverUrl=" + serverUrl + "\n" +
                                "trustStore.aliases=" + trustStoreAliases + "\n"
                        );
                    }
                }
        );
        try {
            xmlRpcClient = new XmlRpcClient(new URL(serverUrl), xmlRpcTransportFactory);
            error = false;
        } catch (Exception e) {
            throw new JFedException("Error creating XmlRpcClient in GeniConnection constructor: " + e.getMessage(), e);
        }
    }

    @Override
    public CommonsHttpClientXmlRpcTransportFactory getXmlRpcTransportFactory() {
        assert !debugInfo.isFakeForDebugging() :
                "Error: This connection is a fake debugging connection, yet the CommonsHttpClientXmlRpcTransportFactory was requested";
        return xmlRpcTransportFactory;
    }

    @Override
    public XmlRpcClient getXmlRpcClient() {
        assert !debugInfo.isFakeForDebugging() :
                "Error: This connection is a fake debugging connection, yet the XmlRpcClient was requested";
        return xmlRpcClient;
    }

    public void clearLastCallData() {
        xmlRpcTransportFactory.clearLastCallData();
    }

    public void close() {
        error = true;
        xmlRpcTransportFactory.clearLastCallData();
        xmlRpcTransportFactory.close();
        xmlRpcTransportFactory = null;
        xmlRpcClient = null;
    }
}