package be.iminds.ilabt.jfed.scanner;

import be.iminds.ilabt.jfed.log.Logger;
import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.lowlevel.ServerType;
import be.iminds.ilabt.jfed.lowlevel.api.AggregateManager2;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.lowlevel.connection.BasicConnectionFactory;
import be.iminds.ilabt.jfed.lowlevel.connection.ConnectionFactory;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnection;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.util.*;
import org.slf4j.LoggerFactory;

import javax.annotation.Nullable;
import java.net.MalformedURLException;
import java.net.URL;
import java.security.cert.X509Certificate;
import java.util.*;

/**
 * ServerScanner: scans for all info about an authority, based on input data.
 */
public class AuthorityScanner {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(AuthorityScanner.class);
    private static Set<ServerPort> failedServerPorts = new HashSet<ServerPort>();
    private final Logger logger;
    private final JFedPreferences jFedPreferences;
    private final AuthorityScannerInput input;
    private final ScanFeedback feedback;
    private final Set<X509Certificate> trustedCerts;     //all certs, including input
    private final Set<X509Certificate> trustedAuthCerts; //all certs used by servers of this authority
    private final Set<String> trustedAliases;
    public Map<ServerType, ScannedInformation<String>> urlsByServerType =
            new HashMap<ServerType, ScannedInformation<String>>();
    protected List<AuthorityScannerLogListener> resultListeners = new ArrayList<AuthorityScannerLogListener>();
    SfaAuthority result = null;
    private ScannedInformation<AuthorityType> types = new ScannedInformation<AuthorityType>();
    private List<String> logs = new ArrayList<String>();
    private ScannedInformation<String> hrns = new ScannedInformation<String>();
    private ScannedInformation<String> urns = new ScannedInformation<String>();
    private Set<ServerBase> serverBases = new HashSet<ServerBase>();


    public AuthorityScanner(Logger logger, AuthorityScannerInput input, ScanFeedback feedback, JFedPreferences jFedPreferences) {
        this.logger = logger;
        this.jFedPreferences = jFedPreferences;
        this.input = input;
        this.feedback = feedback;
        trustedCerts = new HashSet<X509Certificate>(input.trustedCerts);
        trustedAuthCerts = new HashSet<X509Certificate>();
        trustedAliases = new HashSet<String>();
    }

    public static SfaAuthority createResult(ScannedInformation<AuthorityType> types,
                                            Map<ServerType, ScannedInformation<String>> urlsByServerType,
                                            ScannedInformation<String> hrns,
                                            ScannedInformation<String> urns,
                                            Set<X509Certificate> trustedAuthCerts,
                                            Set<String> trustedAliases) throws JFedException {
        Map<ServerType, URL> urls = new HashMap<ServerType, URL>();
        for (Map.Entry<ServerType, ScannedInformation<String>> e : urlsByServerType.entrySet()) {
            ServerType st = e.getKey();
            ScannedInformation<String> urlInfo = e.getValue();

            if (!urlInfo.isEmpty())
                try {
                    urls.put(st, new URL(urlInfo.getMostTrusted()));
                } catch (MalformedURLException ex) {
                    throw new RuntimeException("Invalid URL: " + urlInfo.getMostTrusted() + " -> " + ex.getMessage(), ex);
                }
        }

        String urnString = urns.getMostTrusted();

        String hrn = hrns.getMostTrusted();

        SfaAuthority result = new SfaAuthority(urnString, hrn, null, urls, null/*proxies*/, null/*gid*/,
                types.getMostTrusted() == null ? null : types.getMostTrusted().toString());
        for (X509Certificate cert : trustedAuthCerts)
            result.addPemSslTrustCert(cert);
        for (String trustedAlias : trustedAliases)
            result.addAllowedCertificateHostnameAlias(trustedAlias);
        return result;
    }

    public void addToUrlsByServerType(ScanCertainty certainty, ServerType serverType, String url) {
        if (urlsByServerType.containsKey(serverType)) {
            ScannedInformation<String> set = urlsByServerType.get(serverType);
            set.add(certainty, url);
        } else {
            ScannedInformation<String> set = new ScannedInformation<String>();
            set.add(certainty, url);
            urlsByServerType.put(serverType, set);
        }
    }

    public Logger getLogger() {
        return logger;
    }

    private void log(String line) {
        logs.add(line);
        LOG.debug(line);
        fireLogLine(line);
    }

    private void log(String line, Throwable t) {
        logs.add(line);
        LOG.debug(line, t);
        fireLogLine(line);
    }

    public void findServerBases() {
        for (String urn : input.urns) {
            GeniUrn u = GeniUrn.parse(urn);
            if (u != null)
                serverBases.add(new ServerBase(u.getEncodedTopLevelAuthority_withoutSubAuth(), 0, null));
        }
        for (String urlStr : input.urls) {
            try {
                URL url = new URL(urlStr);
                int port = url.getPort();
                if (port <= 0) port = url.getDefaultPort();
                String path = url.getPath();
                if (path.endsWith("/")) path = path.substring(0, path.length() - 1);
                if (path.startsWith("/")) path = path.substring(1);
                serverBases.add(new ServerBase(url.getHost(), port, path));
            } catch (MalformedURLException e) {
                log("Invalid URL: " + urlStr);
            }
        }
    }

    public void processInputUrl(String url) {
        //TODO put inputURL in toTestUrlsByServerType somehow
        addToUrlsByServerType(ScanCertainty.USER_INPUT, new ServerType(ServerType.GeniServerRole.AM, 2), url);
    }

    public void addUrlsForType(AuthorityType authType) {
        addUrlsForType(authType, ScanCertainty.GUESS);
    }

    public void addUrlsForType(AuthorityType authType, ScanCertainty scanCertainty) {
        for (ServerBase serverBase : serverBases)
            switch (authType) {
                case EMULAB: {
                    ServerBase normal = serverBase;
                    ServerBase knownPort = new ServerBase(serverBase.hostname, 12369, serverBase.path);
                    List<ServerBase> bases = new ArrayList<ServerBase>();
                    bases.add(normal);
                    if (!normal.equals(knownPort))
                        bases.add(knownPort);
                    for (ServerBase base : bases) {
                        String serverBaseStr =
                                base.toString().replaceAll("^www.", "").replaceAll("protogeni/xmlrpc.*", "");
                        addToUrlsByServerType(scanCertainty, new ServerType(ServerType.GeniServerRole.PROTOGENI_SA, 2), "https://www." + serverBaseStr + "protogeni/xmlrpc/sa");
                        addToUrlsByServerType(scanCertainty, new ServerType(ServerType.GeniServerRole.GENI_CH_SA, 1), "https://www." + serverBaseStr + "protogeni/xmlrpc/geni-sa");
                        addToUrlsByServerType(scanCertainty, new ServerType(ServerType.GeniServerRole.GENI_CH_MA, 1), "https://www." + serverBaseStr + "protogeni/xmlrpc/geni-ma");
                        addToUrlsByServerType(scanCertainty, new ServerType(ServerType.GeniServerRole.AM, 2), "https://www." + serverBaseStr + "protogeni/xmlrpc/am/2.0");
                        addToUrlsByServerType(scanCertainty, new ServerType(ServerType.GeniServerRole.AM, 3), "https://www." + serverBaseStr + "protogeni/xmlrpc/am/3.0");
                    }
                    break;
                }
                case SFAWRAP: {
                    addToUrlsByServerType(scanCertainty, new ServerType(ServerType.GeniServerRole.AM, 3),
                            "https://" + new ServerBase(serverBase.hostname, 12346, serverBase.path) + "");

                    addToUrlsByServerType(scanCertainty, new ServerType(ServerType.GeniServerRole.PlanetLabSliceRegistry, 1),
                            "https://" + new ServerBase(serverBase.hostname, 12345, serverBase.path) + "");
                    break;
                }
                case FOAM: {
                    String serverPath = serverBase.toString().replaceAll("foam/gapi.*", "");
                    addToUrlsByServerType(scanCertainty, new ServerType(ServerType.GeniServerRole.AM, 1),
                            "https://" + new ServerBase(serverBase.hostname, 3626, serverPath) + "foam/gapi/1");
                    addToUrlsByServerType(scanCertainty, new ServerType(ServerType.GeniServerRole.AM, 2),
                            "https://" + new ServerBase(serverBase.hostname, 3626, serverPath) + "foam/gapi/2");
                    break;
                }
                case GCF: {
                    //no specific URLs preconfigured in GCF
                    break;
                }
            }
    }


    /**
     * Check if SSL connection can be set up securely. Does not call anything yet.
     */
    public boolean checkUrl(String urlStr) {
        URL url = null;
        try {
            url = new URL(urlStr);
            SSLCertificateDownloader.SSLCertificateJFedInfo info = SSLCertificateDownloader.getCertificateInfo(url);
            log("Note: got certificate info for \"" + urlStr + "\".");
            assert info != null;

            String urn = info.getUrn();
            if (info.getUrn() == null) {
//                log("The URN of the server could not be determined from the certificate. Using placeholder URN for now.");
                urn = "urn:publicid:IDN+" + url.getHost() + "+authority+cm";
                urns.add(ScanCertainty.GUESS, urn);
            } else {
                urns.add(ScanCertainty.INDIRECT_RECV_FROM_SERVER, info.getUrn());
            }

            if (info.getCert() == null) {
                log("Problem fetching certificate from server for URL: " + urlStr);
                return false;
            }

            if (info.isSelfSigned() && !trustedCerts.contains(info.getCert())) {
                log("Note: server uses self signed certificate:\n" +
                        KeyUtil.x509certificateToPem(info.getCert()));

                List<X509Certificate> certChain = new ArrayList<X509Certificate>();
                certChain.add(info.getCert());
                if (!feedback.trustCert(certChain)) {
                    log("user does not trust certificate for: " + urlStr);
                    return false;
                } else {
                    trustedCerts.addAll(certChain);
                    trustedAuthCerts.addAll(certChain);
                    log("User trusts self signed certificate for " + urlStr);
                }
            }

            if (!info.getSubjectMatchesHostname() && !trustedAliases.contains(info.getSubject())) {
                log("Note: Server certificate subject name is not for \"" + url.getHost() + "\" but for \"" + info.getSubject() + "");

                if (!feedback.trustAlias(url.getHost(), info.getSubject())) {
                    log("user does not trust certificate alias for: " + urlStr);
                    return false;
                } else {
                    trustedAliases.add(info.getSubject());
                    log("User trusts alias in certificate.");
                }
            }

            return true;
        } catch (MalformedURLException e) {
            log("Invalid URL: " + urlStr);
            return false;
        }
    }

    public String fixUrl(URL serverUrl, String getVersionUrl) {
        String serverBase = serverUrl.getHost();
        if (serverUrl.getPort() >= 0)
            serverBase += ":" + serverUrl.getPort();

        String res = getVersionUrl;
        if (serverUrl.toString().startsWith("https:") && res.startsWith("http:"))
            res = res.replace("http:", "https:");

        res = res.replace("localhost", serverBase);
        res = res.replace("127.0.0.1", serverBase);
        return res;
    }

    /**
     * returns true if call successful
     */
    public boolean callAMGetVersion(@Nullable ServerType serverType, URL url) {
        log("calling AM GetVersion on: " + url);
        try {
            Map<ServerType, URL> urlMap = new HashMap<ServerType, URL>();
            urlMap.put(new ServerType(ServerType.GeniServerRole.AM, 2), url);
            //Note: already, this will (possibly, depending on errors) be overwritten later
            SfaAuthority wizardCreatedGeniAuthority = null;
            wizardCreatedGeniAuthority =
                    new SfaAuthority("urn:publicid:IDN+dummyforscan+authority+cm", "scannedServer", null, urlMap, null/*proxies*/, null, null);
            wizardCreatedGeniAuthority.setAllowedCertificateHostnameAlias(new ArrayList<String>(trustedAliases));

            //TODO WARNING: only 1 supported! TODO: let SfaAuthority support multiple trusted certs
            for (X509Certificate cert : trustedCerts) {
                wizardCreatedGeniAuthority.addPemSslTrustCert(cert);
            }

            //contact server using AM API(v2 but it doesn't really matter) and use GetVersion to get correct AM urls...
            AggregateManager2 am2 = new AggregateManager2(getLogger(), jFedPreferences);
            ConnectionFactory connectionFactory = new BasicConnectionFactory();
            AggregateManager2.AggregateManagerReply<AggregateManager2.VersionInfo> versionInfoReply =
                    am2.getVersion(
                            connectionFactory.createSfaSslWithClientAuthConnection(
                                    url.toString(),
                                    new JFedTrustStore(wizardCreatedGeniAuthority),
                                    input.userCerts,
                                    input.userPrivateKey,
                                    null /*proxyInfo*/,
                                    null/*handleUntrustedCallback*/,
                                    new JFedConnection.DebugInfo(false, wizardCreatedGeniAuthority, null))
                            //connectionFactory.createSfaSslWithClientAuthConnection(url, new JFedTrustStore(wizardCreatedGeniAuthority), null /*proxyInfo*/, null/*handleUntrustedCallback*/, debugMode, user);
                    );

            log("  Processing GetVersion result. (" + XmlRpcPrintUtil.xmlRpcObjectToString(versionInfoReply.getRawResult()).length() + " bytes in json)");//XmlRpcPrintUtil.xmlRpcObjectToString(versionInfoReply.getRawResult()));
            if (versionInfoReply.getRawResult() == null) {
                log("GetVersion result is null");
                return false;
            }
            if (versionInfoReply.getRawResult().get("value") == null) {
                log("GetVersion result has no value");
                return false;
            }
            if (!(versionInfoReply.getRawResult().get("value") instanceof Hashtable)) {
                log("GetVersion result has non Hashtable value: " + versionInfoReply.getRawResult().get("value").getClass().getName());
                return false;
            }
            Hashtable valueHt = (Hashtable) versionInfoReply.getRawResult().get("value");

            AuthorityScanner.ScannedInformation<AuthorityType> scannedType = AuthorityTypeFinder.findTypeFromGetVersionReply(versionInfoReply.getRawResult());
            if (scannedType != null) {
                types.addAll(scannedType);
                log("Type(s) derived from GetVersion: "+scannedType.getSet());
            }
            else
                log("No type derived from GetVersion reply");

            String newHrn = (String) valueHt.get("hrn");
            if (newHrn != null)
                wizardCreatedGeniAuthority.updateHrn(newHrn);

            String newUrn = (String) valueHt.get("urn");
            if (newUrn != null) {
                wizardCreatedGeniAuthority.updateUrn(newUrn);
                urns.add(ScanCertainty.RECV_FROM_SERVER, newUrn);
            }

            ServerType realServerType = null;
            if (valueHt.containsKey("geni_api")) {
                Object apiVersion = valueHt.get("geni_api");
                if (apiVersion instanceof Integer)
                    realServerType = new ServerType(ServerType.GeniServerRole.AM, (Integer) apiVersion);
                if (apiVersion instanceof String)
                    realServerType =
                            new ServerType(ServerType.GeniServerRole.AM, Integer.parseInt((String) apiVersion));
                if (realServerType != null)
                    addToUrlsByServerType(ScanCertainty.CONFIRMED, realServerType, url.toString());
            }

            Hashtable geniApiVersions = (Hashtable) valueHt.get("geni_api_versions");
            if (geniApiVersions.containsKey("3"))
                addToUrlsByServerType(ScanCertainty.RECV_FROM_SERVER, new ServerType(ServerType.GeniServerRole.AM, 3), fixUrl(url, (String) geniApiVersions.get("3")));
            if (geniApiVersions.containsKey("2"))
                addToUrlsByServerType(ScanCertainty.RECV_FROM_SERVER, new ServerType(ServerType.GeniServerRole.AM, 2), fixUrl(url, (String) geniApiVersions.get("2")));
            if (geniApiVersions.containsKey("1"))
                addToUrlsByServerType(ScanCertainty.RECV_FROM_SERVER, new ServerType(ServerType.GeniServerRole.AM, 1), fixUrl(url, (String) geniApiVersions.get("1")));
            //TODO: check if received URL matches url tested.  (could be problem with ports or hostname)
//            if (confirmedUrlsByServerType.get(realServerType) == null || !confirmedUrlsByServerType.get(realServerType).contains(url.toString()))

            return true;
        } catch (Exception e) {
            log("exception calling GetVersion: " + e.getMessage(), e);
            return false;
        }
    }

    public void scan() {
        //no reason not to cache this for all sessions
//        failedServerPorts.clear();

        for (X509Certificate cert : input.trustedCerts)
            types.addAll(AuthorityTypeFinder.findTypeFromCert(cert));
        for (X509Certificate cert : input.userCerts)
            types.addAll(AuthorityTypeFinder.findTypeFromCert(cert));
        for (String urn : input.urns)
            types.addAll(AuthorityTypeFinder.findTypeFromUrn(urn));
        for (String url : input.urls)
            types.addAll(AuthorityTypeFinder.findTypeFromUrl(url));
        if (input.type != null)
            types.add(ScanCertainty.USER_INPUT, input.type);

        log("Derived Authority type from initial input: " + types);

        findServerBases();
        log("serverBases: " + serverBases);

        for (String url : input.urls)
            processInputUrl(url);
        for (AuthorityType authorityType : types.getSet())
            addUrlsForType(authorityType);

        //make copy of URLs
        Map<ServerType, List<String>> origUrlsByServerType = new HashMap<ServerType, List<String>>();
        for (Map.Entry<ServerType, ScannedInformation<String>> e : urlsByServerType.entrySet()) {
            ServerType serverType = e.getKey();
            ScannedInformation<String> urls = e.getValue();
            origUrlsByServerType.put(serverType, new ArrayList<String>(urls.getSet()));
        }

        for (Map.Entry<ServerType, List<String>> e : origUrlsByServerType.entrySet()) {
            ServerType serverType = e.getKey();
            List<String> urls = e.getValue();
            testUrls(serverType, urls);
        }

        if (types.isEmpty()) {
            log("No types are known, so falling back to trying all types. This could take a while.");
            for (AuthorityType authorityType : AuthorityType.values())
                addUrlsForType(authorityType, ScanCertainty.FALLBACK);
        }

        Map<ServerType, List<String>> newUrlsByServerType = new HashMap<ServerType, List<String>>();
        for (Map.Entry<ServerType, ScannedInformation<String>> e : urlsByServerType.entrySet()) {
            ServerType serverType = e.getKey();
            ScannedInformation<String> urls = e.getValue();
            List<String> orig_urls = origUrlsByServerType.get(serverType);
            List<String> newUrlList = new ArrayList<String>(urls.getSet());
            if (orig_urls != null)
                newUrlList.removeAll(orig_urls);
            if (!newUrlList.isEmpty())
                newUrlsByServerType.put(serverType, newUrlList);
        }

        for (Map.Entry<ServerType, List<String>> e : newUrlsByServerType.entrySet()) {
            ServerType serverType = e.getKey();
            List<String> urls = e.getValue();
            testUrls(serverType, urls);
        }
    }

    private void testUrls(ServerType serverType, List<String> urls) {
        log("URLs to test for " + serverType + ": " + urls);

        for (String urlStr : urls) {
            URL url = null;
            try {
                url = new URL(urlStr);
            } catch (MalformedURLException e1) {
                log("Invalid URL: " + urlStr);
                continue;
            }

            ServerPort serverPort =
                    new ServerPort(url.getHost(), url.getPort() <= 0 ? url.getDefaultPort() : url.getPort());
            boolean ok;
            if (failedServerPorts.contains(serverPort)) {
                log("Server port is already known not to work, will not try again: " + urlStr);
                ok = false;
            } else
                ok = checkUrl(urlStr);
            if (!ok) {
                log("Could not setup secure SSL connection to: " + urlStr);
                failedServerPorts.add(serverPort);
                continue;
            }

            boolean triedGetVersion = false;
            boolean successGetVersion = false;
            switch (serverType.getRole()) {
//                    case PlanetLabSliceRegistry: { triedGetVersion = true; successGetVersion = callPlanetLabSliceRegistryGetVersion(serverType, url); break; }
//                    case PROTOGENI_SA: { triedGetVersion = true; successGetVersion = callProtoGeniSAGetVersion(serverType, url); break; }
                case AM: {
                    triedGetVersion = true;
                    successGetVersion = callAMGetVersion(serverType, url);
                    break;
                }
//                    case GENI_CH_SA: { triedGetVersion = true; successGetVersion = callFederationSAOrMAGetVersion(serverType, url); break; }
//                    case GENI_CH_MA: { triedGetVersion = true; successGetVersion = callFederationSAOrMAGetVersion(serverType, url); break; }
                case PROTOGENI_CH: {
                    break;
                }
                case GENI_CH: {
                    break;
                }
                case SCS: {
                    break;
                }
                case OCCI: {
                    break;
                }
                default: {
                    log("unsupported servertype: " + serverType);
                    break;
                }
            }
        }

    }


    private void createAutoResult() {
        try {
            result = createResult(types, urlsByServerType, hrns, urns, trustedAuthCerts, trustedAliases);
        } catch (JFedException e) {
            log("Error creating result: " + e.getMessage(), e);
        }
    }

    public AuthorityType getType() {
        return types.getMostTrusted();
    }

    public SfaAuthority getAutoResult() {
        if (result == null)
            createAutoResult();
        return result;
    }

    public synchronized void fireLogLine(String logLine) {
        for (AuthorityScannerLogListener l : resultListeners)
            l.onLogLine(logLine);
    }

    public synchronized void addResultListener(AuthorityScannerLogListener l) {
        resultListeners.add(l);
    }

    public synchronized void removeResultListener(AuthorityScannerLogListener l) {
        resultListeners.remove(l);
    }

    public String getLogText() {
        String res = "";
        for (String logline : logs)
            res += logline + "\n";
        return res;
    }

    public List<String> getLogs() {
        return logs;
    }

    public ScannedInformation<String> getHrns() {
        return hrns;
    }

    public ScannedInformation<String> getUrns() {
        return urns;
    }

    public Map<ServerType, ScannedInformation<String>> getUrlsByServerType() {
        return urlsByServerType;
    }

    public Set<String> getTrustedAliases() {
        return trustedAliases;
    }

    public Set<X509Certificate> getTrustedAuthCerts() {
        return trustedAuthCerts;
    }

    public ScannedInformation<AuthorityType> getTypes() {
        return types;
    }

    public static enum ScanCertainty {
        WRONG(-1),                       //Known to be false, but might be close to truth
        FALLBACK(0),                     //wild guess, derived from other data, licely not correct. But possibly better than nothing.
        GUESS(1),                        //derived from other data (may be incorrect)
        GOOD_GUESS(2),                   //derived from other data, likely correct (may still be incorrect)
        USER_INPUT(3),                   //originally inputted by user
        INDIRECT_RECV_FROM_SERVER(4),    //derived from server date, such as certificate (but server could be wrong)
        RECV_FROM_SERVER(5),             //received from server, (but server could be wrong)
        CONFIRMED(10);                   //confirmed to be correct

        public final int value;

        private ScanCertainty(int value) {
            this.value = value;
        }

        public boolean isMoreCertainThan(ScanCertainty o) {
            return value > o.value;
        }
    }

    public static class ScannedSingleFact<T> {
        private final T fact;
        private final ScanCertainty certainty;

        public ScannedSingleFact(ScanCertainty certainty, T fact) {
            this.fact = fact;
            this.certainty = certainty;
        }

        public T getFact() {
            return fact;
        }

        public ScanCertainty getCertainty() {
            return certainty;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            ScannedSingleFact that = (ScannedSingleFact) o;

            if (certainty != that.certainty) return false;
            if (fact != null ? !fact.equals(that.fact) : that.fact != null) return false;
            return true;
        }

        @Override
        public int hashCode() {
            int result = fact != null ? fact.hashCode() : 0;
            result = 31 * result + (certainty != null ? certainty.hashCode() : 0);
            return result;
        }
    }

    public static class ScannedInformation<T> {
        final Set<ScannedSingleFact<T>> infoSet = new HashSet<ScannedSingleFact<T>>();

        public void add(ScanCertainty certainty, T fact) {
            infoSet.add(new ScannedSingleFact<T>(certainty, fact));
        }

        public void add(ScannedSingleFact<T> fact) {
            infoSet.add(fact);
        }

        /**
         * return null if nothing known
         */
        public T getMostTrusted() {
            ScannedSingleFact<T> res = getMostTrustedScannedSingleFact();
            if (res == null) return null;
            return res.getFact();
        }

        /**
         * return null if nothing known
         */
        public ScannedSingleFact<T> getMostTrustedScannedSingleFact() {
            ScannedSingleFact<T> res = null;
            for (ScannedSingleFact<T> fact : infoSet) {
                if (res == null || fact.certainty.isMoreCertainThan(res.certainty))
                    res = fact;
            }
            if (res == null)
                return null;
            return res;
        }

        public Set<T> getSet() {
            Set<T> res = new HashSet<T>();
            for (ScannedSingleFact<T> f : infoSet)
                res.add(f.getFact());
            return res;
        }

        public Collection<ScannedSingleFact<T>> getFacts() {
            return infoSet;
        }

        public boolean isEmpty() {
            return infoSet.isEmpty();
        }

        public void addAll(ScannedInformation<T> infos) {
            if (infos == null) return;
            infoSet.addAll(infos.infoSet);
        }

        @Override
        public String toString() {
            return infoSet.toString();
        }
    }

    //serverBase: hostname:port  or hostname:port/bla/bla/       port == 0 means unknown port   (known default ports (80,443) should be added)
    private class ServerBase {
        final String hostname;
        final int port;
        final String path;

        ServerBase(String hostname, int port, String path) {
            assert hostname != null;
            this.hostname = hostname;
            this.port = port;
            this.path = path;
        }

        @Override
        public String toString() {
            if (path == null) {
                if (port <= 0)
                    return hostname + "/";
                else
                    return hostname + ':' + port + "/";
            } else {
                if (port <= 0)
                    return hostname + "/" + path + '/';
                else
                    return hostname + ':' + port + "/" + path + '/';
            }
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            ServerBase that = (ServerBase) o;

            if (port != that.port) return false;
            if (!hostname.equals(that.hostname)) return false;
            if (path != null ? !path.equals(that.path) : that.path != null) return false;

            return true;
        }

        @Override
        public int hashCode() {
            int result = hostname.hashCode();
            result = 31 * result + port;
            result = 31 * result + (path != null ? path.hashCode() : 0);
            return result;
        }
    }

    private class ServerPort {
        public final String hostname;
        public final int port;

        private ServerPort(String hostname, int port) {
            this.hostname = hostname;
            this.port = port;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            ServerPort that = (ServerPort) o;

            if (port != that.port) return false;
            if (!hostname.equals(that.hostname)) return false;

            return true;
        }

        @Override
        public int hashCode() {
            int result = hostname.hashCode();
            result = 31 * result + port;
            return result;
        }
    }
}
