package be.iminds.ilabt.jfed.util;

import ch.ethz.ssh2.Connection;
import ch.ethz.ssh2.LocalStreamForwarder;
import ch.ethz.ssh2.ServerHostKeyVerifier;
import org.slf4j.LoggerFactory;
import org.xbill.DNS.Address;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.*;

public class GanymedSshSocketImplFactory implements SocketImplFactory {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(GanymedSshSocketImplFactory.class);

    protected Connection sshConnection;

    private final static boolean singleConnectionPerSShConnection = true;
    public synchronized static GanymedSshSocketImplFactory getFactory(SshServerProxyHelper.SshProxyInfo sshProxyInfo) throws IOException {
        //see also SSHSocketImpl.close. since singleConnectionPerSShConnection == true, we can close SSH connection there!
        if (singleConnectionPerSShConnection)
            return new GanymedSshSocketImplFactory(sshProxyInfo);
        else
            throw  new RuntimeException("SSH connection reuse yet implemented");
    }

    private GanymedSshSocketImplFactory(final SshServerProxyHelper.SshProxyInfo sshProxyInfo) throws IOException {
        sshConnection = new Connection(null/*socketImplFactory*/, sshProxyInfo.getHostName(), sshProxyInfo.getPort());

        ServerHostKeyVerifier serverHostKeyVerifier =
                sshProxyInfo.getServerHostKey() == null ? null : new ServerHostKeyVerifier() {
                    @Override
                    public boolean verifyServerHostKey(String hostName, int port, String hostKeyAlgo, byte[] hostKey) throws Exception {
                        //we will ignore hostname, as we don't know if an IP or full name is used. (we could check but it's not that important)
//                if (!sshProxyInfo.getHostName().equals(hostName))
//                    return false;
                        LOG.debug("Checking key for host " + hostName);
                        if (sshProxyInfo.getPort() != port)
                            return false;
                        if (!sshProxyInfo.getServerHostKeyAlgo().equals(hostKeyAlgo))
                            return false;
                        if (sshProxyInfo.getServerHostKey().length != hostKey.length)
                            return false;
                        for (int i = 0; i < hostKey.length; i++) {
                            if (sshProxyInfo.getServerHostKey()[i] != hostKey[i])
                                return false;
                        }
                        LOG.debug("   Key known");
                        return true;
                    }
                };

        //timeouts in milliseconds. Zero means no timeout.
        int connectTimeout_ms = 1000; //TCP connection setup
        int kexTimeout_ms = 8000;     //full SSH connection setup

        LOG.debug("Connecting to " + sshProxyInfo.getHostName() + ":" + sshProxyInfo.getPort() + " with timeout " + connectTimeout_ms);
        sshConnection.connect(serverHostKeyVerifier/*ServerHostKeyVerifier = null means don't check host keys*/, connectTimeout_ms, kexTimeout_ms);

        boolean isAuthenticated =
                sshConnection.authenticateWithPublicKey(sshProxyInfo.getSshUsername(), sshProxyInfo.getPemPrivateKeyChars(), null/*pass*/);
        LOG.debug("Connected to " + sshProxyInfo.getHostName() + ":" + sshProxyInfo.getPort() + " authenticated=" + isAuthenticated);

        if (!isAuthenticated) {
            throw new IOException("Could not authenticate");
        }
    }

    //note: multiple SocketImpl can be needed per factory. example: When using ganymed itself using this factory, it needs multiple
    public SocketImpl createSocketImpl() {
        if (sshConnection == null)
            throw new RuntimeException("The SSH connection is not ready. Cannot create socket over it.");
        assert sshConnection != null : "sshConnection == null";
        SSHSocketImpl res = new SSHSocketImpl(sshConnection);
        return res;

    }


    /**
     * We aren't going to actually create any new connection, we will forward
     * things to SSH.
     */
    private static class SSHSocketImpl extends SocketImpl {
        Connection sshConnection;

        LocalStreamForwarder localStreamForwarder;

        InputStream is;

        OutputStream os;

        SSHSocketImpl(Connection sshConnection) {
            this.sshConnection = sshConnection;
        }

        @Override
        protected void accept(SocketImpl s) throws IOException {
            throw new IOException("SSHSocketImpl does not implement accept");
        }

        @Override
        protected int available() throws IOException {
            if (is == null) {
                throw new ConnectException("Not connected");
            }
            return is.available();
        }

        @Override
        protected void bind(InetAddress host, int port) throws IOException {
            if ((host != null && !host.isAnyLocalAddress()) || port != 0) {
                throw new IOException("SSHSocketImpl does not implement bind");
            }
        }

        @Override
        protected void close() throws IOException {
            if (localStreamForwarder != null) {
                localStreamForwarder.close();
                localStreamForwarder = null;
                is = null;
                os = null;
            }

            //also close SSH connection:
            if (singleConnectionPerSShConnection) {
                //hard close: close by closing socket
                sshConnection.close(new Throwable("Hard close due to user request"), true/*hard*/);

                //soft close: close with disconnect packet
                //sshConnection.close();

                sshConnection = null;
            }
            else
                throw new RuntimeException("connection close not yet implemented for SSH connection reuse");
        }

        @Override
        protected void connect(String host, int port) throws IOException {
            InetAddress addr = Address.getByName(host);
            connect(addr, port);
        }

        @Override
        protected void connect(InetAddress address, int port) throws IOException {
            connect(new InetSocketAddress(address, port), 300000);
        }

        @Override
        protected void connect(SocketAddress address, int timeout)
                throws IOException {
            localStreamForwarder = sshConnection.createLocalStreamForwarder((
                            (InetSocketAddress) address).getHostName(),
                    ((InetSocketAddress) address).getPort()
            );

            is = localStreamForwarder.getInputStream();
            os = localStreamForwarder.getOutputStream();
        }

        @Override
        protected void create(boolean stream) throws IOException {
            if (stream == false) {
                throw new IOException("Cannot handle UDP streams");
            }
        }

        @Override
        protected InputStream getInputStream() throws IOException {
            return is;
        }

        @Override
        protected OutputStream getOutputStream() throws IOException {
            return os;
        }

        @Override
        protected void listen(int backlog) throws IOException {
            throw new IOException("SSHSocketImpl does not implement listen");
        }

        @Override
        protected void sendUrgentData(int data) throws IOException {
            throw new IOException("SSHSocketImpl does not implement sendUrgentData");
        }

        public Object getOption(int optID) throws SocketException {
            //fake answer to some common option requests if possible.
            if (optID == SocketOptions.TCP_NODELAY) return new Boolean(false);
            if (optID == SocketOptions.SO_BINDADDR)
                throw new SocketException("SSHSocketImpl does not implement getOption for SO_BINDADDR");
            if (optID == SocketOptions.SO_REUSEADDR) return new Boolean(false);
            if (optID == SocketOptions.SO_BROADCAST) return new Boolean(false);
            if (optID == SocketOptions.IP_MULTICAST_IF)
                throw new SocketException("SSHSocketImpl does not implement getOption for IP_MULTICAST_IF");
            if (optID == SocketOptions.IP_MULTICAST_IF2)
                throw new SocketException("SSHSocketImpl does not implement getOption for IP_MULTICAST_IF2");
            if (optID == SocketOptions.IP_MULTICAST_LOOP)
                throw new SocketException("SSHSocketImpl does not implement getOption for IP_MULTICAST_LOOP");
            if (optID == SocketOptions.IP_TOS)
                throw new SocketException("SSHSocketImpl does not implement getOption for IP_TOS");
            if (optID == SocketOptions.SO_LINGER) return new Integer(0); //no linger
            if (optID == SocketOptions.SO_TIMEOUT) return new Integer(0); //?
            if (optID == SocketOptions.SO_SNDBUF)
                throw new SocketException("SSHSocketImpl does not implement getOption for SO_SNDBUF"); //actual send buffer size
            if (optID == SocketOptions.SO_RCVBUF)
                throw new SocketException("SSHSocketImpl does not implement getOption for SO_RCVBUF");
            //actual recv buffer size
            if (optID == SocketOptions.SO_KEEPALIVE) return new Boolean(false);
            if (optID == SocketOptions.SO_OOBINLINE) return new Boolean(false);

            throw new SocketException("SSHSocketImpl does not implement getOption for " + optID);
        }

        /**
         * We silently ignore setOptions because they do happen, but there is
         * nothing that we can do about it.
         */
        public void setOption(int optID, Object value) throws SocketException {
        }

    }


    public static class SshSocket extends Socket {
        public SshSocket(GanymedSshSocketImplFactory ganymedSshSocketImplFactory) throws SocketException {
            super(ganymedSshSocketImplFactory.createSocketImpl());
        }
    }
}
