package be.iminds.ilabt.jfed;

import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.lowlevel.TestCertHelper;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl.AutomaticAggregateManagerWrapper;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl.AutomaticUserAndSliceApiWrapper;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityListModel;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.lowlevel.connection.*;
import be.iminds.ilabt.jfed.util.ExtraInfoCallback;
import be.iminds.ilabt.jfed.util.JFedTrustStore;
import com.google.inject.AbstractModule;
import com.google.inject.Provides;

import javax.inject.Singleton;
import java.net.MalformedURLException;
import java.net.URL;
import java.security.PrivateKey;
import java.security.cert.X509Certificate;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import static org.mockito.Mockito.mock;

/**
 * Fake connections for testing, with mock replies to requests.
 *
 * This uses almost every real thing, except for the underlying XmlRpc transport etc.
 */
public class MockHttpRepliesTestCoreModule extends AbstractModule {
    protected final TestCertHelper testCertHelper;
    protected final List<MockSfaConnection.MockXmlRpcReply> mockReplies;
    protected final List<MockSfaConnection.OnMockXmlRpcListener> onNoMock;
    protected final List<MockSfaConnection.MockXmlRpcHistoryListener> historyListeners;

    public MockHttpRepliesTestCoreModule(List<MockSfaConnection.MockXmlRpcReply> mockReplies) {
        this.testCertHelper = new TestCertHelper();
        this.mockReplies = mockReplies;
        onNoMock = new ArrayList<MockSfaConnection.OnMockXmlRpcListener>();
        historyListeners = new ArrayList<MockSfaConnection.MockXmlRpcHistoryListener>();
    }
    public MockHttpRepliesTestCoreModule() {
        this(new TestCertHelper());
    }
    public MockHttpRepliesTestCoreModule(TestCertHelper testCertHelper) {
        this.testCertHelper = testCertHelper;
        this.mockReplies = new ArrayList<MockSfaConnection.MockXmlRpcReply>();
        onNoMock = new ArrayList<MockSfaConnection.OnMockXmlRpcListener>();
        historyListeners = new ArrayList<MockSfaConnection.MockXmlRpcHistoryListener>();
    }

    public void addMockReply(MockSfaConnection.MockXmlRpcReply mr) {
        mockReplies.add(mr);
    }

    public void addOnNoMockReply(MockSfaConnection.OnMockXmlRpcListener l) {
        onNoMock.add(l);
    }

    public void addMockXmlRpcHistoryListener(MockSfaConnection.MockXmlRpcHistoryListener historyListener) {
        historyListeners.add(historyListener);
    }


    @Override
    protected void configure() {
        bind(JFedTrustStore.class).in(Singleton.class);
        bind(TestCertHelper.class).toInstance(testCertHelper);
        bind(JFedConnectionProvider.class).to(SfaConnectionPool.class).in(Singleton.class);

        bind(AutomaticAggregateManagerWrapper.AutomaticAggregateManagerWrapperFactory.class).
                to(AutomaticAggregateManagerWrapper.BasicAutomaticAggregateManagerWrapperFactory.class).in(Singleton.class);

        bind(AutomaticUserAndSliceApiWrapper.AutomaticUserAndSliceApiWrapperFactory.class).
                to(AutomaticUserAndSliceApiWrapper.BasicAutomaticUserAndSliceApiWrapperFactory.class).in(Singleton.class);

//        bind(ConnectionFactory.class).to(BasicConnectionFactory.class).in(Singleton.class);
    }

    private Map<String, MockSfaConnection> conCache = new HashMap<String, MockSfaConnection>();
    public MockSfaConnection createCon(AuthorityListModel authorityListModel, String urlStr) throws JFedException {
        if (conCache.containsKey(urlStr)) return conCache.get(urlStr);

        ConnectionFactory res = mock(ConnectionFactory.class);

        URL url = null;
        try {
            url = new URL(urlStr);
        } catch (MalformedURLException e) {
            throw new RuntimeException(e);
        }
        SfaAuthority auth = null;
        for (SfaAuthority a : authorityListModel.getAuthorities()) {
            if (a.getUrls().containsValue(url))
                auth = a;
        }
        assert auth != null;

        MockSfaConnection con = new MockSfaConnection(urlStr, auth, null);
        con.setMockReplies(mockReplies);
        con.setNoMockListeners(onNoMock);
        con.setHistoryListeners(historyListeners);

        conCache.put(urlStr, con);

        return con;
    }

    @Provides
    @Singleton
    public ConnectionFactory provideConnectionFactory(final AuthorityListModel authorityListModel) throws JFedException {
        assert authorityListModel != null;
        SfaAuthority userAuth = authorityListModel.getFromAnyUrnExact(TestCertHelper.saUrn);
        SfaAuthority amAuth = authorityListModel.getFromAnyUrnExact(TestCertHelper.am1Urn);
        SfaAuthority am2Auth = authorityListModel.getFromAnyUrnExact(TestCertHelper.am2Urn);
        SfaAuthority scsAuth = authorityListModel.getFromAnyUrnExact(TestCertHelper.scsUrn);
        assert userAuth != null;
        assert amAuth != null;
        assert am2Auth != null;
        assert scsAuth != null;

        ConnectionFactory res = new ConnectionFactory() {
            @Override
            public RestSslNoAuthConnection createRestSslNoAuthConnection(String serverUrlString, JFedTrustStore conTrustStore, JFedConnection.ProxyInfo usedProxyInfo, HandleUntrustedCallback handleUntrustedCallback, JFedConnection.DebugInfo debugInfo) throws JFedException {
                throw new RuntimeException("not mocked");
            }

            @Override
            public RestSslPasswordConnection createRestSslPasswordConnection(String serverUrlString, JFedTrustStore conTrustStore, JFedConnection.ProxyInfo usedProxyInfo, ExtraInfoCallback.Login bonfireLogin, HandleUntrustedCallback handleUntrustedCallback, JFedConnection.DebugInfo debugInfo) throws JFedException {
                throw new RuntimeException("not mocked");
            }

            @Override
            public SfaConnection createSfaSslWithClientAuthConnection(String serverUrlStr, JFedTrustStore conTrustStore, List<X509Certificate> clientCertChain, PrivateKey privateKey, JFedConnection.ProxyInfo usedProxyInfo, HandleUntrustedCallback handleUntrustedCallback, JFedConnection.DebugInfo debugInfo) throws JFedException {
                return createCon(authorityListModel, serverUrlStr);
            }

            @Override
            public SfaConnection createSfaSslWithoutClientAuthConnection(String serverUrlStr, JFedTrustStore conTrustStore, JFedConnection.ProxyInfo usedProxyInfo, HandleUntrustedCallback handleUntrustedCallback, JFedConnection.DebugInfo debugInfo) throws JFedException {
                return createCon(authorityListModel, serverUrlStr);
            }

            @Override
            public SfaConnection createSfaPlainHttpConnection(String serverUrlStr, JFedConnection.ProxyInfo usedProxyInfo, JFedConnection.DebugInfo debugInfo) throws JFedException {
                return createCon(authorityListModel, serverUrlStr);
            }
        };

//        ConnectionFactory res = mock(ConnectionFactory.class);
//
//        when(res.createSfaSslWithClientAuthConnection(
//                org.mockito.Matchers.eq(amAuth.getUrl(ServerType.GeniServerRole.AM, 3).toExternalForm()),
//                org.mockito.Matchers.any(JFedTrustStore.class),
//                org.mockito.Matchers.any(List.class),
//                org.mockito.Matchers.any(PrivateKey.class),
//                org.mockito.Matchers.any(JFedConnection.ProxyInfo.class),
//                org.mockito.Matchers.any(HandleUntrustedCallback.class),
//                org.mockito.Matchers.any(JFedConnection.DebugInfo.class)
//        )).thenReturn(amCon);
//
//        when(res.createSfaSslWithClientAuthConnection(
//                org.mockito.Matchers.eq(am2Auth.getUrl(ServerType.GeniServerRole.AM, 2).toExternalForm()),
//                org.mockito.Matchers.any(JFedTrustStore.class),
//                org.mockito.Matchers.any(List.class),
//                org.mockito.Matchers.any(PrivateKey.class),
//                org.mockito.Matchers.any(JFedConnection.ProxyInfo.class),
//                org.mockito.Matchers.any(HandleUntrustedCallback.class),
//                org.mockito.Matchers.any(JFedConnection.DebugInfo.class)
//        )).thenReturn(am2Con);
//
//        when(res.createSfaSslWithClientAuthConnection(
//                org.mockito.Matchers.eq(userAuth.getUrl(ServerType.GeniServerRole.PROTOGENI_SA, 1).toExternalForm()),
//                org.mockito.Matchers.any(JFedTrustStore.class),
//                org.mockito.Matchers.any(List.class),
//                org.mockito.Matchers.any(PrivateKey.class),
//                org.mockito.Matchers.any(JFedConnection.ProxyInfo.class),
//                org.mockito.Matchers.any(HandleUntrustedCallback.class),
//                org.mockito.Matchers.any(JFedConnection.DebugInfo.class)
//        )).thenReturn(saCon);
//
//        when(res.createSfaSslWithClientAuthConnection(
//                org.mockito.Matchers.eq(scsAuth.getUrl(ServerType.GeniServerRole.SCS, 1).toExternalForm()),
//                org.mockito.Matchers.any(JFedTrustStore.class),
//                org.mockito.Matchers.any(List.class),
//                org.mockito.Matchers.any(PrivateKey.class),
//                org.mockito.Matchers.any(JFedConnection.ProxyInfo.class),
//                org.mockito.Matchers.any(HandleUntrustedCallback.class),
//                org.mockito.Matchers.any(JFedConnection.DebugInfo.class)
//        )).thenReturn(scsCon);

        return res;
    }

    public TestCertHelper getTestCertHelper() {
        return testCertHelper;
    }
}
