package be.iminds.ilabt.jfed.rspec.parser;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.xml.namespace.QName;
import javax.xml.stream.Location;
import javax.xml.stream.XMLEventReader;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.events.Attribute;
import javax.xml.stream.events.EndElement;
import javax.xml.stream.events.StartElement;
import javax.xml.stream.events.XMLEvent;
import javax.xml.transform.stream.StreamSource;
import java.io.StringReader;
import java.util.*;

import static be.iminds.ilabt.jfed.rspec.parser.RspecXmlConstants.*;

/**
 * RspecCompare
 *
 * This class can compare 2 rspecs, and determine if they are equal.
 *
 * Restrictions:
 *   - The nodes and links MUST be in the same order.   (this assumption is needed to make this fast enough)
 *   - no support for text mixed with elements
 *
 * Things that are ignored when comparing:
 *   - Where namespaces are declared, and which namespaces are declared.
 *   - Order of elements within nodes and links
 *   - Order of non node and link elements within <rspec>
 *   - comments
 *   - generated and generated_by attribute in <rspec>
 *   - http://jfed.iminds.be/rspec/ext/jfed/1 namespace
 *
 * Things that are certainly cause of inequality:
 *   - extra or missing attributes
 *   - extra or missing elements
 */
public class RspecCompare {
    private final static Logger LOG = LoggerFactory.getLogger(RspecCompare.class);

    private boolean equal = true;
    private List<Difference> differences;

    private final String rspecNameA;
    private final String rspecNameB;

    public RspecCompare(String rspecA, String rspecB) throws XMLStreamException, RspecParseException {
        this(rspecA, rspecB, "RSpec A", "RSpec B");
    }
    public RspecCompare(String rspecA, String rspecB, String rspecNameA, String rspecNameB) throws XMLStreamException, RspecParseException {
        this.differences = new ArrayList<Difference>();
        this.rspecNameA = rspecNameA;
        this.rspecNameB = rspecNameB;
        this.equal = equalRspec(rspecA, rspecB);
    }

    public boolean isEqual() {
        return equal;
    }

    public List<Difference> getDifferences() {
        return differences;
    }


    private final static QName FIXHACK_BAD = Q_JFEDCOMMAND_EXP_COMMAND_OLDERROR;
    private final static QName FIXHACK_GOOD = Q_JFEDCOMMAND_EXP_COMMAND;
    private static QName fixNameHack(QName name) {
        if (name.equals(FIXHACK_BAD))
            return FIXHACK_GOOD;
        return name;
    }

    /**
     * Checks if 2 lists of element events are the same.
     *
     * Assumptions:
     *   - first event is startElement
     *   - last event is matching endElement
     *
     * @param elementEventsA
     * @param elementEventsB
     * @return
     */
    private boolean compareElement(LinkedList<XMLEvent> elementEventsA, LinkedList<XMLEvent> elementEventsB, boolean logDiffs) throws XMLStreamException, RspecParseException {
        if (elementEventsA.isEmpty()) {
            if (!elementEventsB.isEmpty() && logDiffs)
                differences.add(new Difference(rspecNameA+" has empty element, where "+rspecNameB+
                        " has non empty element ("+elementEventsB.getFirst().asStartElement().getName()+")",
                        null, elementEventsB.getFirst().asStartElement().getLocation(), rspecNameA, rspecNameB));
            if (elementEventsB.isEmpty())
                return true;
            else
                return false;
        }
        if (elementEventsB.isEmpty()) {
            if (!elementEventsB.isEmpty() && logDiffs)
                differences.add(new Difference(rspecNameB+" has empty element, where "+rspecNameA+
                        " has non empty element ("+elementEventsA.getFirst().asStartElement().getName()+")",
                        null, elementEventsA.getFirst().asStartElement().getLocation(), rspecNameA, rspecNameB));
            return false;
        }

        assert elementEventsA.size() >= 2 : "sizeA="+elementEventsA.size()+" sizeB="+elementEventsB.size();
        assert elementEventsB.size() >= 2 : "sizeA="+elementEventsA.size()+" sizeB="+elementEventsB.size();

        XMLEvent firstEventA = elementEventsA.getFirst();
        XMLEvent firstEventB = elementEventsB.getFirst();
        XMLEvent lastEventA = elementEventsA.getLast();
        XMLEvent lastEventB = elementEventsB.getLast();

        assert firstEventA.isStartElement();
        assert firstEventB.isStartElement();
        assert lastEventA.isEndElement();
        assert lastEventB.isEndElement();

        StartElement startA = firstEventA.asStartElement();
        StartElement startB = firstEventB.asStartElement();
        EndElement endA = lastEventA.asEndElement();
        EndElement endB = lastEventB.asEndElement();

        assert fixNameHack(startA.getName()).equals(fixNameHack(endA.getName()));
        assert fixNameHack(startB.getName()).equals(fixNameHack(endB.getName()));

        if (! fixNameHack(startA.getName()).equals(fixNameHack(startB.getName()))) {
            if (logDiffs) differences.add(new Difference(""+rspecNameA+" has different name than "+rspecNameB+"." +
                    " "+rspecNameA+" name="+startA.getName()+
                    " "+rspecNameB+" name="+startB.getName(),
                    startA.getLocation(), startB.getLocation(), rspecNameA, rspecNameB));
            return false;
        }

        assert fixNameHack(endA.getName()).equals(fixNameHack(endB.getName()));

        if (!sameAttributes(startA, startB, logDiffs)) {
            if (logDiffs) differences.add(
                    new Difference(""+rspecNameA+" has different attributes than "+rspecNameB+" for element \"" + startA.getName()+"\"."+
                            " "+rspecNameA+" attributes: "+attributesToString(startA)+
                            " "+rspecNameB+" attributes: "+attributesToString(startB)
                            , startA.getLocation(), startB.getLocation(), rspecNameA, rspecNameB));
            return false;
        }

        LinkedList<LinkedList<XMLEvent>> subElementsA = new LinkedList<LinkedList<XMLEvent>>();
        LinkedList<String> charsA = new LinkedList<String>();

        for (Iterator<XMLEvent> it = elementEventsA.iterator(); it.hasNext(); ) {
            XMLEvent e = it.next();

            if (e == firstEventA || e == lastEventA)
                continue;

            if (e.isStartElement()) {
                if (isIgnoredElement(e.asStartElement().getName()))
                    getAllElementEventsFromIt(e.asStartElement(), it);
                else
                    subElementsA.add(getAllElementEventsFromIt(e.asStartElement(), it));
            }
            if (e.isCharacters() && !e.asCharacters().isWhiteSpace())
                charsA.add(e.asCharacters().getData());
        }
        LinkedList<LinkedList<XMLEvent>> subElementsB = new LinkedList<LinkedList<XMLEvent>>();
        LinkedList<String> charsB = new LinkedList<String>();
        for (Iterator<XMLEvent> it = elementEventsB.iterator(); it.hasNext(); ) {
            XMLEvent e = it.next();

            if (e == firstEventB || e == lastEventB)
                continue;

            if (e.isStartElement()) {
                if (isIgnoredElement(e.asStartElement().getName()))
                    getAllElementEventsFromIt(e.asStartElement(), it);
                else
                    subElementsB.add(getAllElementEventsFromIt(e.asStartElement(), it));
            }
            if (e.isCharacters() && !e.asCharacters().isWhiteSpace())
                charsB.add(e.asCharacters().getData());
        }

        if (charsA.size() != charsB.size()) {
            if (logDiffs) differences.add(new Difference(""+rspecNameA+" has "+charsA.size()+" CharacterElements, while "+rspecNameB+" has "+charsB.size()
                    +" "+rspecNameA+"text="+(charsA.size() == 1 ? "\""+charsA.get(0)+"\"" : (charsA.isEmpty() ? "" : charsA))
                    +" "+rspecNameB+"text="+(charsB.size() == 1 ? "\""+charsB.get(0)+"\"" : (charsB.isEmpty() ? "" : charsB)),
                    startA.getLocation(), startB.getLocation(), rspecNameA, rspecNameB));
            return false;
        }
        if (!charsA.isEmpty()) {
            if (!charsA.equals(charsB)) {
                if (logDiffs) differences.add(new Difference("Text in "+rspecNameA+" and "+rspecNameB+" differs:" +
                        " "+rspecNameA+"text="+(charsA.size() == 1 ? "\""+charsA.get(0)+"\"" : charsA)+
                        " "+rspecNameB+"text="+(charsB.size() == 1 ? "\""+charsB.get(0)+"\"" : charsB),
                        startA.getLocation(), startB.getLocation(), rspecNameA, rspecNameB));
                return false;
            }
        }

        for (LinkedList<XMLEvent> subElA : subElementsA) {
            assert !subElA.isEmpty();
            StartElement startElementA = subElA.getFirst().asStartElement();
            assert subElA.size() >= 2 : "sizeA="+subElA.size();

            boolean found = false;
            for (ListIterator<LinkedList<XMLEvent>> itB = subElementsB.listIterator(); itB.hasNext(); ) {
                LinkedList<XMLEvent> subElB = itB.next();
                assert subElA.size() >= 2 : "sizeA="+subElA.size()+" sizeB="+subElB.size();
                assert subElB.size() >= 2 : "sizeA="+subElA.size()+" sizeB="+subElB.size();
                if (compareElement(subElA, subElB, false)) { //todo: solve non recursively
                    itB.remove();
                    found = true;
                    break;
                }
            }
            if (!found) {
                if (logDiffs) differences.add(new Difference(
                        ""+rspecNameA+" has element \""+startElementA.getName()+"\". " +
                                ""+rspecNameB+" does not have a matching element",
                        startElementA.getLocation(), null,
                        rspecNameA, rspecNameB));
                return false;
            }
        }
        if (!subElementsB.isEmpty()) {
            if (logDiffs) {
                for (List<XMLEvent> e : subElementsB) {
                    assert !e.isEmpty();
                    assert e.get(0).isStartElement();
                    StartElement bStart = e.get(0).asStartElement();
                    differences.add(new Difference(
                            ""+rspecNameB+" has element \"" + bStart.getName() + "\""+
                                    " which "+rspecNameA+" does not have",
                            null, bStart.getLocation(),
                            rspecNameA, rspecNameB));
                }
            }
            return false;
        }

        return true;
    }

    @SuppressWarnings("unchecked")
    public static String attributesToString(StartElement startElement) {
        String res = "";
        for (Iterator<Attribute> it = startElement.getAttributes(); it.hasNext(); ) {
            Attribute att = it.next();
            if (!res.isEmpty())
                res += " ";
            res +=""+att.getName()+"="+att.getValue();
        }
        return res;
    }

    private static boolean isIgnoredRspecAttribute(QName name) {
        if (name.equals(Q_ATT_RSPEC_GEN_BY)) return true;
        if (name.equals(Q_ATT_RSPEC_GEN_DATE)) return true;
        if (name.equals(Q_XSI_SCHEMALOCATION)) return true;
        return false;
    }

    private static boolean isIgnoredAttribute(QName name) {
//        if (name.getNamespaceURI().equals(NAMESPACE_JFED)) return true;
        return false;
    }
    private static boolean isIgnoredElement(QName name) {
        return name.getNamespaceURI().equals(NAMESPACE_JFED) && !name.equals(FIXHACK_BAD);
    }

    @SuppressWarnings("unchecked")
    private boolean sameAttributes(StartElement startA, StartElement startB, boolean logDiffs) {
        boolean isRspec = startA.getName().equals(Q_RSPEC);
        Map<QName, String> attributesB = new HashMap<QName, String>();
        Map<QName, String> attributesA = new HashMap<QName, String>();

        for (Iterator<Attribute> itB = startB.getAttributes(); itB.hasNext(); ) {
            Attribute attB = itB.next();
            if (!isIgnoredAttribute(attB.getName()) && (!isRspec || !isIgnoredRspecAttribute(attB.getName())))
                attributesB.put(attB.getName(), attB.getValue());
        }
        for (Iterator<Attribute> itA = startA.getAttributes(); itA.hasNext(); ) {
            Attribute attA = itA.next();
            if (!isIgnoredAttribute(attA.getName()) && (!isRspec || !isIgnoredRspecAttribute(attA.getName())))
                attributesA.put(attA.getName(), attA.getValue());
        }
        //rely on HashMap equals. It might be the most efficient method here.
        return attributesA.equals(attributesB);
    }

    /** assumed that iterator won't return first element */
    public LinkedList<XMLEvent> getAllElementEventsFromIt(StartElement startEvent, Iterator<XMLEvent> it) throws XMLStreamException, RspecParseException {
        QName startElementName = startEvent.getName();
        String name = startElementName.getLocalPart();
        LinkedList<XMLEvent> res = new LinkedList<XMLEvent>();
        res.add(startEvent);
        try {
            int depth = 0; //depth is used for support for recursive extra xml  (ex:  <a><a> </a></a>  )
            while (it.hasNext()) {
                XMLEvent e = it.next();

                res.add(e);
                if (e.isStartElement()) {
                    if (e.asStartElement().getName().getLocalPart().equals(name)) {
                        depth++;
                    }
                }
                if (e.isEndElement()) {
                    if (e.asEndElement().getName().getLocalPart().equals(name)) {
                        if (depth == 0) {
                            assert res.size() >= 2 : "not enough elements ("+res.size()+") for startElementName="+startElementName+" @ "+RspecParseException.locationToString(startEvent.getLocation());
                            return res;
                        }
                        depth--;
                    }
                }
            }
            throw new RspecParseException(startEvent.getLocation(), "<"+startElementName+"> tag was never closed");
        } catch (RspecParseException e) {
            String curList = "";
            for (XMLEvent event : res)
                curList += "      "+StaxHelper.xmlEventToStringWithAttributes(event)+"\n";
            LOG.error("Parse error in getAllElementEvents. \n      startEvent="+StaxHelper.xmlEventToString(startEvent)+"\n    Current list:\n"+curList, e);
            throw e;
        } catch (Exception e) {
            String curList = "";
            for (XMLEvent event : res)
                curList += "      "+StaxHelper.xmlEventToStringWithAttributes(event)+"\n";
            LOG.error("RspecCompare error in getAllElementEvents. \n      startEvent="+StaxHelper.xmlEventToString(startEvent)+"\n    Current list:\n"+curList, e);
            throw new RuntimeException("RspecCompare error in getAllElementEvents. \n      startEvent="+StaxHelper.xmlEventToString(startEvent)+"\n    Current list:\n"+curList, e);
        }
    }
    /** assumed that iterator won't return first element */
    public LinkedList<XMLEvent> getAllElementEventsFromReader(StartElement startEvent, XMLEventReader it) throws XMLStreamException, RspecParseException {
        QName startElementName = startEvent.getName();
        String name = startElementName.getLocalPart();
        LinkedList<XMLEvent> res = new LinkedList<XMLEvent>();
        res.add(startEvent);
        try {
            int depth = 0; //depth is used for support for recursive extra xml  (ex:  <a><a> </a></a>  )
            while (it.hasNext()) {
                XMLEvent e = it.nextEvent();

                res.add(e);
                if (e.isStartElement()) {
                    if (e.asStartElement().getName().getLocalPart().equals(name)) {
                        depth++;
                    }
                }
                if (e.isEndElement()) {
                    if (e.asEndElement().getName().getLocalPart().equals(name)) {
                        if (depth == 0) {
                            assert res.size() >= 2 : "not enough elements ("+res.size()+") for startElementName="+startElementName+" @ "+RspecParseException.locationToString(startEvent.getLocation());
                            return res;
                        }
                        depth--;
                    }
                }
            }
            throw new RspecParseException(startEvent.getLocation(), "<"+startElementName+"> tag was never closed");
        } catch (RspecParseException e) {
            String curList = "";
            for (XMLEvent event : res)
                curList += "      "+StaxHelper.xmlEventToStringWithAttributes(event)+"\n";
            LOG.error("Parse error in getAllElementEvents. \n      startEvent="+StaxHelper.xmlEventToString(startEvent)+"\n    Current list:\n"+curList, e);
            throw e;
        } catch (Exception e) {
            String curList = "";
            for (XMLEvent event : res)
                curList += "      "+StaxHelper.xmlEventToStringWithAttributes(event)+"\n";
            LOG.error("Parse error in getAllElementEvents. \n      startEvent="+StaxHelper.xmlEventToString(startEvent)+"\n    Current list:\n"+curList, e);
            throw new RuntimeException("Parse error in getAllElementEvents. \n      startEvent="+StaxHelper.xmlEventToString(startEvent)+"\n    Current list:\n"+curList, e);
        }
    }
//    /** assumed that eventsReader won't return fist element */
//    public static LinkedList<XMLEvent> getAllElementEventsFromEventReader(StartElement startEvent, XMLEventReader eventsReader) throws XMLStreamException, RspecParseException {
//        return getAllElementEventsFromIterator(startEvent, eventsReader);
//    }

    private XMLEvent readUntilLinkOrNode(XMLEventReader eventReader, LinkedList<XMLEvent> restRspec) throws XMLStreamException, RspecParseException {
        XMLEvent e = null;
        while (eventReader.hasNext()) {
            e = eventReader.nextEvent();
            //                updateProgress(e.getLocation().getCharacterOffset(), inputRspec.length());

            if (e.isStartElement()) {
                if (e.asStartElement().getName().equals(Q_RSPEC)) {
                    restRspec.add(e);
                    continue;
                }
                if (e.asStartElement().getName().equals(Q_NODE)) {
                    return e;
                }
                if (e.asStartElement().getName().equals(Q_LINK)) {
                    return e;
                }
                restRspec.addAll(getAllElementEventsFromReader(e.asStartElement(), eventReader));
            }
            if (e.isEndElement()) {
                if (e.asEndElement().getName().equals(Q_RSPEC)) {
                    restRspec.add(e);
                    continue;
                }
            }
        }
        return e;
    }
    private boolean equalRspec(String rspecA, String rspecB) throws XMLStreamException, RspecParseException {
        try {
            XMLInputFactory xif = XMLInputFactory.newFactory();
            StreamSource xmlA = new StreamSource(new StringReader(rspecA));
            XMLEventReader eventReaderA = xif.createXMLEventReader(xmlA);
            XMLEvent eA = eventReaderA.nextEvent();
            assert eA.isStartDocument();


            StreamSource xmlB = new StreamSource(new StringReader(rspecB));
            XMLEventReader eventReaderB = xif.createXMLEventReader(xmlB);
            XMLEvent eB = eventReaderB.nextEvent();
            assert eB.isStartDocument();

            LinkedList<XMLEvent> restRspecA = new LinkedList<XMLEvent>();
            LinkedList<XMLEvent> restRspecB = new LinkedList<XMLEvent>();

            while (eventReaderA.hasNext() || eventReaderB.hasNext()) {

                eA = readUntilLinkOrNode(eventReaderA, restRspecA);
                boolean nodeAtA = eA != null && eA.isStartElement() && eA.asStartElement().getName().equals(Q_NODE);
                boolean linkAtA = eA != null && eA.isStartElement() && eA.asStartElement().getName().equals(Q_LINK);

                eB = readUntilLinkOrNode(eventReaderB, restRspecB);
                boolean nodeAtB = eB != null && eB.isStartElement() && eB.asStartElement().getName().equals(Q_NODE);
                boolean linkAtB = eB != null && eB.isStartElement() && eB.asStartElement().getName().equals(Q_LINK);

                assert !(nodeAtA && linkAtA);
                assert !(nodeAtB && linkAtB);

                if (nodeAtA != nodeAtB) {
                    assert eA != null && eB != null;
                    differences.add(new Difference(""+rspecNameA+" "+(nodeAtA ? "has a node" : "does not have a node")+
                            " where "+rspecNameB+" "+(nodeAtB ? "has a node" : "does not have a node")+
                            " "+rspecNameA+" e="+StaxHelper.xmlEventToString(eA)+
                            " "+rspecNameB+" e="+StaxHelper.xmlEventToString(eB),
                            eA.getLocation(), eB.getLocation(), rspecNameA, rspecNameB));
                    return false;
                }
                if (linkAtA  != linkAtB) {
                    assert eA != null && eB != null;
                    differences.add(new Difference(""+rspecNameA+" "+(nodeAtA ? "has a link" : "does not have a link")+
                            " where "+rspecNameB+" "+(nodeAtB ? "has a link" : "does not have a link")+
                            " "+rspecNameA+" e="+StaxHelper.xmlEventToString(eA)+
                            " "+rspecNameB+" e="+StaxHelper.xmlEventToString(eB),
                            eA.getLocation(), eB.getLocation(), rspecNameA, rspecNameB));
                    return false;
                }

                LinkedList<XMLEvent> eventsA = null;
                if (nodeAtA || linkAtA) {
                    eventsA = getAllElementEventsFromReader(eA.asStartElement(), eventReaderA);
                }
                LinkedList<XMLEvent> eventsB = null;
                if (nodeAtB || linkAtB) {
                    eventsB = getAllElementEventsFromReader(eB.asStartElement(), eventReaderB);
                }

                if (nodeAtA && nodeAtB) {
                    if (!compareElement(eventsA, eventsB, true)) {
                        differences.add(new Difference(rspecNameA+" and "+rspecNameB+" have a different node", eA.getLocation(), eB.getLocation(), rspecNameA, rspecNameB));
                        return false;
                    }
                }
                if (linkAtA && linkAtB) {
                    if (!compareElement(eventsA, eventsB, true)) {
                        differences.add(new Difference(rspecNameA+" and "+rspecNameB+" have a different link", eA.getLocation(), eB.getLocation(), rspecNameA, rspecNameB));
                        return false;
                    }
                }
            }

            if (!compareElement(restRspecA, restRspecB, true)) {
                differences.add(new Difference(rspecNameA+" and "+rspecNameB+" have the same node and link elements. But the rest of the RSpec differs", null, null, rspecNameA, rspecNameB));
                return false;
            }

            return true;
        } catch (XMLStreamException e) {
            throw new RspecParseException(e.getLocation(), "Problem parsing RSpec", e);
        }
    }

    public static class Difference {
        private final String reason;
        private final Location locationA;
        private final Location locationB;
        private final String rspecNameA;
        private final String rspecNameB;


        private Difference(String reason, Location locationA, Location locationB, String rspecNameA, String rspecNameB) {
            this.reason = reason;
            this.locationA = locationA;
            this.locationB = locationB;
            this.rspecNameA = rspecNameA;
            this.rspecNameB = rspecNameB;
        }

        public String getReason() {
            return reason;
        }

        public Location getLocationA() {
            return locationA;
        }

        public Location getLocationB() {
            return locationB;
        }

        public String getRspecNameB() {
            return rspecNameB;
        }

        public String getRspecNameA() {
            return rspecNameA;
        }

        @Override
        public String toString() {
            return reason
                    + (locationA != null ? " in "+rspecNameA+" @ "+RspecParseException.locationToString(locationA) : "")
                    + (locationB != null ? " in "+rspecNameB+" @ "+RspecParseException.locationToString(locationB) : "");
        }
    }

    public static String differencesListToString(List<Difference> diffs) {
        StringBuilder sb = new StringBuilder();
        for (Difference diff : diffs) {
            sb.append(diff.getReason()).append("\n");
            if (diff.getLocationA() != null)
                sb.append("   " + " in ").append(diff.getRspecNameA()).append(" @ ").append(RspecParseException.locationToString(diff.getLocationA())).append("\n");
            if (diff.getLocationB() != null)
                sb.append("   " + " in ").append(diff.getRspecNameB()).append(" @ ").append(RspecParseException.locationToString(diff.getLocationB())).append("\n");
            sb.append("\n");
        }
        return sb.toString();
    }
}
